//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package eg;

import la.la.Type;
import la.la.Value;
import static la.la.Value.*;
import la.maths.Vector;  // A data-set is a Vector of Values
import la.util.*;        // error(...)
import mml.*;            // For Model, UPModel, Normal, Mixture, Tree, etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.
 *
 * Reference: 
 *   Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.
 *
 * This file is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL) as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this file. If not, see <http://www.gnu.org/licenses/>.
 */
// ----------------------------------------------------------------------------

/** Is the <I>27&nbsp;Club</I> a thing or is it an urban myth?
    Simple analysis of the age at death of musicians.
    Note, please do not ask me for the raw data &mdash;
    that belongs to Professor Dianna Kenny, University of Sydney.
    Also see
    [<A HREF="https://doi.org/10.1007/978-3-319-76433-7" TARGET="BOOKA">the&nbsp;book</A>]
    and the
    [<A HREF="https://www.cantab.net/users/mmlist/MML/A/" TARGET="mmlist">home&nbsp;page</A>].  */
public class Musicians
 {
   /** A double compactly to at least 4 significant figures. */
   public static String x2S(double x) { return Util.x2String(x, 4); }

   /** Each row of the input is: INT2 = int x int, being age & frequency. */
   public static Type.Tuple INT2 =
      new Type.Tuple.GP("IntxInt", new Type [] {Type.INT, Type.INT});

   /** Bounds on mean(s) and standard deviation(s). */
   public static Real
      mLwb = real(10), mUpb = real(102),    // mu in [mLwb, mUpb]
      sLwb = real(1), sUpb = real(50);   // sigma in [sLwb, sUpb]
   // -------------------------------------------------------------------------

   /** Get the data from the named file;
       return a Vector of ages weighted by frequency.  */
   public static Vector getData(String filename)           // getData(filename)
    {
      System.out.println( "getData( \"" + filename + "\" )" );
      java.io.FileInputStream inp = null;
      try{ inp = new java.io.FileInputStream(filename);
         }catch(Exception e){ Util.error("getData(" + filename+  "): "+e); }
      Vector raw  = Vector.csv(INT2, inp);      // (age, freq)^*
      int nElts = raw.nElts();

      Vector col0 = raw.col(0), col1 = raw.col(1);
      int minAge  = 200, maxAge = -1,
          modeAge = -1,  modeF  = -1,
          frqSum = 0;
      double [] ages = new double [nElts];

      for( int i = 0; i < nElts; i ++ )
       { int ai = col0.elt(i).n(),  // i-th age
             fi = col1.elt(i).n();  // i-th frequency
         ages[i] = ai;
         minAge = Math.min(ai, minAge);
         maxAge = Math.max(ai, maxAge);
         frqSum += fi;
         if( fi > modeF ) { modeAge = ai; modeF = fi; }
       }

      System.out.println( "Raw: rows=" + raw.nElts()  // DEBUG
        + " frqSum=" + frqSum
        + " minAge=" + minAge
        + " maxAge=" + maxAge
        + " modeAge=" + modeAge + " (" + modeF + ")" );     // DEBUG

      Vector ds = Vector.doubles(ages, /* AoM = */ 1.0) .weight(col1);

// for(int i = 0; i < nElts; i ++ )  // DEBUG
// System.out.print( ds.elt(i).x() + "x" + ds.wt(i) + ", " );  // DEBUG
// System.out.println();             // DEBUG

      System.out.println("--");
      return ds;
    }//getData(filename)
   // -------------------------------------------------------------------------

   /** Fit a {@link Mixture Mixture} of one or more
       {@link NormalUPM Normals} to data-set 'ds'.  */
   public static void analyse(String title, Vector ds)
    { System.out.println( title + ":" );
      Mixture     mxUPM = new Mixture(MML.Normal);
      Mixture.Est mxEst =
         (Mixture.Est) mxUPM.estimator(Value.quad(mLwb,mUpb, sLwb,sUpb));
      Mixture.M   mx = (Mixture.M) mxEst.ds2Model(ds);
      results(mx, ds);
    }//analyse

   /** Display {@link Model Model} 'm', its two-part message length, and
       the {@link Model#nlLH negative log likelihood} of data-set 'ds'. */
   public static void results(Model m, Vector ds)
    { System.out.println( m + "\n"
       + "msg="
       + x2S(m.msg()) + "=" + x2S(m.msg1()) + "+" + x2S(m.msg2()) + " nits\n"
       + "    "
       + x2S(m.msgBits()) + "=" + x2S(m.msg1bits()) + "+"
                                 + x2S(m.msg2bits()) + " bits\n"
       + "(nlLH=" + x2S(m.nlLH(ds)) +" nits = " + x2S(m.nl2LH(ds)) +" bits)\n"
       + "--" );
    }//results

  /** Starting with mixture model 'mx1', use
      est.{@link mml.Mixture.Est#EM(Mixture.M,la.maths.Vector) EM(mx1,ds)}
      to improve it.  */
   public static Mixture.M improve(Mixture.Est est, Mixture.M mx1, Vector ds)
    { System.out.println( "improve on " + mx1 + "\n"
                        + "as a start, leads to..." );
      Mixture.M mx2 = est.EM(mx1, ds);
      results(mx2, ds);
      return mx2;
    }//improve(...)
   // -------------------------------------------------------------------------

   /** Simple analysis of a data-set, [(ageAtDeath, frequency)],
       from a given file or from file ../data/musicians-age-freq. */
   public static void main(String[] argv)                               // main
    {
      System.out.print("-- run Musicians " );
      for( int i = 0; i < argv.length; i ++ )
         System.out.print( argv[i] + " " );
      System.out.println( "--" );
      String filename = argv.length > 0 ? argv[0]
                                        : "../data/musicians-age-freq";

      Vector ds = getData( filename );
      analyse("analyse data from " + filename, ds);  // *

      Mixture     mxUPM = new Mixture(MML.Normal);
      Mixture.Est mxEst =
         (Mixture.Est) mxUPM.estimator(Value.quad(mLwb,mUpb, sLwb,sUpb));
      // initial guess at a two-component mixture...
      Mixture.M mx2 = mxUPM.sp2Model(0, 0,
        pair( Vector.doubles(new double[]{0.1, 0.9}),
              Vector.values(new Value[]{ pair(real(27), real(4)),
                                         pair(real(55), real(19)) }) ) );
      improve(mxEst, mx2, ds);       // *

      Mixture.M mx3 = mxUPM.sp2Model(0, 0,
        pair( Vector.doubles(new double[]{0.1, 0.1, 0.8}),
              Vector.values(new Value[]{ pair(real(27), real(4)),
                                         pair(real(40), real(6)),
                                         pair(real(55), real(19)) }) ) );
      improve(mxEst, mx3, ds);       // *

   // double []
   // fakeAges = new double []
   //   {22,23,24,25,26, 27, 28,29,30,31,32, 33,34,35,36,37, 38,39,40,41,42},
   // fakeWts = new double []
   //   { 1, 1, 1, 5, 9, 20,  9, 5, 1, 1, 1,  2, 3, 6, 9,20, 30,20, 9, 4, 1};
   // Vector fakeDs = Vector.doubles(fakeAges, 1.0).weight(fakeWts);
   // analyse("fake data-set", fakeDs);

      System.out.println("-- done Musicians --");
    }//main

 }//class Musicians

// (c) Lloyd Allison, Jan 2017, 4 Nov 2017.  Licence to 'Musicians' is Gnu GPL.
