//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package graph;

import la.util.Util;
import la.util.Series;
import la.la.Value;
import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of Directed Graphs.
    Operations on Directed Graphs, and special Directed Graphs.
    Also see {@link Undirected} and {@link Graph}. */
public abstract class Directed extends Graph                      // * Directed
 {
   /** The standard type of Directed (directed, no self-loop) Graphs. */
   public final static Type unlabelled = new Type("directed", true);

   /** Type of Directed, unlabelled, self-loops (O) <I>allowed</I> Graphs. */
   public final static Type unlabelled_O = new Type("directed", true, true);

// from Graph:
// public Type type()                abstract                                 *
// public boolean   isDirected()
// public boolean isUndirected()
// public int vSize()                abstract                                 *
// public int eSize()
// public boolean selfLoops()        ? v -> v allowed ?
// public boolean isEdge(v0,v1)      ? v0 -> v1 ?  abstract                   *
// public boolean adjacent(v0, v1)   ? v0 -> v1 or v1 -> v0 ?
// public boolean vLabelled()
// public boolean eLabelled()
// public Value vLabel(v)
// public Value eLabel(v0, v1)

   /** {@link #inDegree inDegree(v)}+{@link #outDegree outDegree(v)}
       (a self-loop &lang;v,v&rang; is counted twice).  */
   public int degree(int v)
    { return inDegree(v) + outDegree(v); }  // no test, cf. Graph

   // public indegree(v)           and outDegree(v)        defined in Graph
   // public directPredecessors(v) and directSuccessors(v) defined in Graph
   // .........................................................................

   /** Identity function; see {@link Graph#toDirected}. */
   public Directed toDirected() { return this; }

   /** Convenience function. */
   public AsUndirected asUndirected()                           // asUndirected
    { return new AsUndirected(); }

   /** Ignore the directions of the Edges in 'this' Directed Graph to
       create an Undirected version; in general there is a <I>loss</I>
       of information. If this Directed is Edge-labelled, "prefer" a
       label of &lang;v0,&nbsp;v1&rang; where v0&le;v1.
       Also see {@link Undirected#asDirected}. */
   public class AsUndirected extends Undirected        // Directed.AsUndirected
                             implements SubGraph
    { private final Type t = Directed.this.type().clone(/*directed*/false);
      public Directed parent() { return Directed.this; }
      public int v2pv(int v) { return v; }
      public Type type() { return t; }
      public int vSize() { return Directed.this.vSize(); }
      public boolean isEdge(int v0, int v1)
       { return Directed.this.adjacent(v0, v1); } // !
      public Value vLabel(int v) { return Directed.this.vLabel(v); }
      public Value eLabel(int v0, int v1)
       { if( v0 > v1 ) { final int t=v0; v0=v1; v1=t; }  // v0 <= v1
         return Directed.this.isEdge(v0, v1)
                   ? Directed.this.eLabel(v0, v1)
                   : Directed.this.eLabel(v1, v0);
       }//eLabel(v0,v1)
      public String toString()
       { return Directed.this + ".asUndirected()"; }
    }//Directed.AsUndirected class
   // .........................................................................

   /** A Dense, Directed Graph with Type t and adjacency Matrix A.
       It is not necessarily "dense", that's just its intended use.
       Also see {@link Sparse} and {@link Undirected.Dense}.  */
   public static class Dense extends Directed                 // Directed.Dense
                             implements Graph.Dense
    { private final Type   t;
      private final Matrix A;
      public Dense(Type t, Matrix A)
       { this.t = t;  this.A = A;
         if( ! A.isSquare() )
            throw new RuntimeException( ", A not square" );
       }//constructor
      public Type type() { return t; }
      public int vSize() { return A.nElts(); }
      public boolean isEdge(int v0, int v1)
       { return A.elt(v0, v1).n() != 0; }
      public String toString()
       { return "Directed.Dense(" + type() + "," + A() + ")"; }
    }//Directed.Dense class


   /** Return a Dense Directed Graph having Type 't' and Adjacency Matrix 'A'.
       Also see {@link Undirected#dense} and {@link Graph#dense}. */
   public static Dense dense(final Type t, final Matrix A)
    { return new Dense(t, A); }

   /** {@link #dense(Type,Matrix) dense(t,A)}.  'selfLoops'&nbsp;specifies
       if Graphs are <I>allowed</I> to have self-loops, <I>not</I> if
       this Graph actually has any. */
   public static Dense dense(boolean selfLoops, Matrix A)
    { return dense(selfLoops ? unlabelled_O : unlabelled, A); }

   /** {@link #dense(boolean,Matrix) dense(false, A)} -- no self-loops. */
   public static Dense dense(Matrix A) { return dense(false, A); }


   /** {@link #dense(Type,Matrix) dense(t,ints(A))}. */
   public static Dense dense(final Type t, final int [] [] A)
    { return dense(t, Matrix.ints(A)); }

   /** {@link #dense(Type,int[][]) dense(t,A)}. */
   public static Dense dense(boolean selfLoops, int [] [] A)
    { return dense(selfLoops ? unlabelled_O : unlabelled, A); }

   /** {@link #dense(boolean,int[][]) dense(false,A)} -- no self-loops. */
   public static Dense dense(int [] [] A) { return dense(false, A); }
   // .........................................................................

   /** Convenience function. */
    public static Sparse sparse(Type t, int V, int [] [] es)
     { return new Sparse(t, V, es); }

   /** The class of Sparse Directed Graphs.
       Also see {@link Dense} and {@link Undirected.Sparse}.  */
   public static class Sparse extends Directed               // Directed.Sparse
                             implements Graph.Sparse
    { private Type t;
      public Type type() { return t; }

      /** The number of Vertices (V) and Edges (E). */
      private int V, E;

      /** Given an Edge &lang;v0,&nbsp;v1&rang;, v1 is in (ascending) succ[v0],
          and v0 is in (ascending) pred[v1].  */
      public final int [] [] succ, pred;
      // NB. 'pred[v]' is used in directPredecessors(v)

      /** The Edges of 'this' Sparse, as an int[E][2] array. */
      public final int [] [] es;

      /** Assume that the max Vertex mentioned in es is the last Vertex. */
      public Sparse(int [] [] es)
       { this(unlabelled, maxV(es)+1, es); }

      /** Directed Edges <NOBR>es = {{v00, v01}, {v10, v11}, ...}</NOBR>. */
      public Sparse(Type t, int V, int [] [] es)
       { this.t  = t;
         this.E  = es.length;
         this.es = es;
         this.V  = V;
         int [] inDeg = new int[V], outDeg = new int[V];
         for( int v = 0; v < V; v ++ ) { inDeg[v] = 0; outDeg[v] = 0; }
         for( int i = 0; i < es.length; i ++ )  // find Vertex in/out-degrees
          { final int v0 = es[i][0], v1 = es[i][1];
            outDeg[v0] ++ ; inDeg[v1] ++ ;
          }
         succ = new int [V] []; pred = new int [V] [];
         for( int v = 0; v < V; v ++ )
          { succ[v] = new int [outDeg[v]];
            pred[v] = new int [ inDeg[v]];
          }
         for( int i = 0; i < es.length; i ++ )
          { final int v0 = es[i][0], v1 = es[i][1];    // <v0, v1>
            int [] vs = succ[v0];                      // v0 -> v1
            int n = vs.length - outDeg[v0];  // 0, 1, ...
               for( ; n > 0 && v1 < vs[n-1]; n -- ) vs[n] = vs[n-1];
               // n == 0 || v1 >= vs[n-1], i.e., insertion sort
               vs[n] = v1; outDeg[v0] -- ;
               if( n > 0 && v1 == vs[n-1] )
                  error( ", duplicate Edge " + new Edge(v0, v1) );
            vs = pred[v1];                             // v1 <- v0
            n = vs.length -  inDeg[v1];  // 0, 1, ...
               for( ; n > 0 && v0 < vs[n-1]; n -- ) vs[n] = vs[n-1];
               // n == 0 || v0 >= vs[n-1]
               vs[n] = v0;  inDeg[v1] -- ;
               if( n > 0 && v0 == vs[n-1] )
                  error( ", duplicate Edge " + new Edge(v0, v1) );
          }//for i
       }//constructor

      public int          vSize() { return V; }
      public int          eSize() { return E; }
      /** O(1)-time. */
      public int  inDegree(int v) { return pred[v].length; }
      /** O(1)-time. */
      public int outDegree(int v) { return succ[v].length; }

      public boolean isEdge(int v0, int v1)
       { return Util.search(succ[v0], v1); }

      public Series.Int joinedTo(int v)
       { return directPredecessors(v).merge(false, directSuccessors(v)); }

      public Series.Int directPredecessors(int v)
       { return Series.fromInts(pred[v]); }

      public Series.Int   directSuccessors(int v)
       { return Series.fromInts(succ[v]); }

      /** Return the largest Vertex number mentioned in {@link #es es},
          which may or may not be the largest one in the Graph.
          Look at Sparse's constructors. */
      public static int maxV(int [] [] es)
       { int max = 0;
         for( int i = 0; i < es.length; i ++ )
          { if( es[i][0] > max ) max = es[i][0];
            if( es[i][1] > max ) max = es[i][1];
          }
         return max;
       }//max(es)

      public String toString()
       { return "Directed.Sparse(" + t + "," + Matrix.ints(es) + ")"; }
      // ..................................

      /** A renumbered Sparse Directed Graph is Sparse and Directed. */
      public Renumbered renumbered(int [] vs)
       { return new Renumbered(vs); }

      /** A Sparse Directed Graph renumbered accpording to 
          {@link #vs vs} is Sparse and Directed. */
      public class Renumbered extends Sparse      // Directed.Sparse.Renumbered
                              implements SubGraph // Note, G <= G
       { public Sparse parent() { return Sparse.this; }
         /** Vertex 'v' of 'this' is Vertex vs[v] of the parent Graph. */
         public final int [] vs;
         public Renumbered(int [] vs)
          { super(Sparse.this.t, Sparse.this.V /* ! */, renumbered_es(vs));
            this.vs = vs;
            if( vs.length != Sparse.this.vSize() )
               error( ", vs["+vs.length+"] need "+Sparse.this.vSize() );
          }//constructor
         public int     v2pv(int v) { return vs[v]; }
         public Value vLabel(int v) { return Sparse.this.vLabel(vs[v]); }
         public Value eLabel(int v0, int v1)
          { return Sparse.this.eLabel(vs[v0], vs[v1]); }
         public String toString()
          { return parent() + ".renumbered(" + Vector.ints(vs) + ")"; }
       }//Directed.Sparse.Renumbered class

      /** Renumber the Vertices in the Edges of the Renumbered Graph. */
      private int [] [] renumbered_es(int [] vs)
       { int [] inv = new int [vs.length]; // inverse perm. to vs
         for( int i = 0; i < vs.length; i ++ ) inv[vs[i]] = i;
         // v of 'this' SubGraph corr. to  vs[v] of super graph and
         // w of super graph corr. to inv[w] of 'this' SubGraph.
         int [] [] es2 = new int [E][2];
         for( int e = 0; e < E; e ++ )
          { es2[e][0] = inv[es[e][0]]; es2[e][1] = inv[es[e][1]]; }
         return es2;
       }//renumbered_es(vs)
      // ..................

      /** An induced subgraph of a Sparse Directed Graph is
          Sparse and Directed. */
      public Induced induced(int [] vs) { return new Induced(vs); }

      /** An induced subgraph of a Sparse Directed Graph is
          Sparse and Directed. */
      public class Induced extends Sparse            // Directed.Sparse.Induced
                           implements SubGraph
       { public Sparse parent() { return Sparse.this; }
         public final int [] vs;
         public Induced(int [] vs)
          { super(Sparse.this.t, vs.length, induced_es(vs));
            this.vs = vs;
           if( vs.length > Sparse.this.vSize() )
               error( ", vs["+vs.length+"] need <= "+Sparse.this.vSize() );
          }//constructor
         public int     v2pv(int v) { return vs[v]; }
         public Value vLabel(int v) { return Sparse.this.vLabel(vs[v]); }
         public Value eLabel(int v0, int v1)
          { return Sparse.this.eLabel(vs[v0], vs[v1]); }
         public String toString()
          { return parent() + ".induced(" + Vector.ints(vs) + ")"; }
       }//Directed.Sparse.Induced class

      /** Work out which Edges of parent exist in subgraph,
          and renumber the Vertices in them. */
      private int [] [] induced_es(int [] vs)
       { int [] inv = new int [V];  // map v in parent to v in subgraph
         for( int i = 0; i < vs.length; i ++ ) inv[vs[i]] = i;
         int E = 0;
         for( int e = 0; e < es.length; e ++ )  // count Edges to be...
            if( Util.search(vs, es[e][0]) && Util.search(vs, es[e][1]) )
               E ++ ; // <v,succ[v0][v1]> ok
         int [] [] es2 = new int [E] [2]; // subgraph's Edges...
         E = 0;
         for( int e = 0; e < es.length; e ++ )  // adjust Vertices...
            if( Util.search(vs, es[e][0]) && Util.search(vs, es[e][1]) )
             { es2[E][0] = inv[es[e][0]]; es2[E][1] = inv[es[e][1]]; E ++ ; }
         return es2;
       }//induced_es(vs)

    }//Sparse class
   // .........................................................................

   /** Also see {@link Undirected.Vertex}. */
   public class Vertex extends Graph.Vertex                  // Directed.Vertex
    { public Vertex(int v) { super(v); }
      public Directed parent() { return Directed.this; }
    }//Directed.Vertex class

   /** Also see {@link Undirected.Edge}. */
   public class Edge extends Graph.Edge                        // Directed.Edge
    { public Edge(int v0, int v1) { super(v0, v1); }
      public Directed parent() { return Directed.this; }
      public String toString() { return "<" + v0 + "," + v1 + ">"; }
    }//Directed.Edge class
   // .........................................................................

   /** The class of Cyclic Directed Graphs of the form
       <NOBR>v0 &rarr; v1 &rarr; ... &rarr; v(n-1) &rarr; v0</NOBR>, again. */
   public static class C extends Sparse                           // Directed.C
    { /** The number of Vertices in the Graph and in the cycle. */
      public C(int n) { super(C_A(n)); }
      public String toString() { return "C_" + vSize(); }
    }//C class

   private static int [] [] C_A(int n)
    { int [] [] es = new int [n] [2];
      for( int i = 0; i < n; i ++ )
       { es[i][0] = i; es[i][1] = i+1; }
      es[n-1][1] = 0;  // close cycle
      return es;
    }//C_A(n)
   // -------------------------------------------------------------------------

   /** See {@link Test#main}. */
   public static void main(String[] argv)                               // main
   { System.out.println("-- test Directed --");
     Test.main(argv);
     System.out.println("-- done Directed --");
   }//main

 }//Directed class,  https://www.cantab.net/users/mmlist/MML/A/

// (c) L.Allison July 2014, Tue 2 Sept 2014. Licence to 'Directed' is Gnu GPL.
