//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package graph;

import java.util.BitSet;

import la.util.Series;
import la.util.RefInt;
import la.la.Value;
import la.maths.Vector;
import la.maths.Matrix;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Core Graph operations and useful constants etc..
    The minimum requirements to instantiate an <I>unlabelled</I> Graph are
    {@link #type}, {@link #vSize}, and {@link #isEdge}.
    If the Graph is Vertex-labelled, specify {@link #vLabel}.
    If the Graph is Edge-labelled,   specify {@link #eLabel}.<BR>
    Note, testing whether a Graph is Directed or Undirected should be done
    via {@link #type}, {@link #isDirected} or {@link #isUndirected}, not
    by looking whether its class is {@link Directed} or {@link Undirected}.
    This is because there are common operations such as {@link #induced},
    {@link #renumbered} and so on, and it may be more "convenient" to
    define them in (and return a) {@link Graph} rather than duplicate
    them in Directed and Undirected, or even in {@link Directed.Dense},
    {@link Directed.Sparse}, {@link Undirected.Dense}, and
    {@link Undirected.Sparse}.<BR>
    Also see {@link README README}.  */
public abstract class Graph extends Value                              // Graph
 {
   /** Note, type(),           {@link #adjacent},
       {@link #vLabelled},     {@link #vLabel},
       {@link #eLabelled}, and {@link #eLabel} must be consistent.  */
   public abstract Type type();                                     // * type()

   /** Check {@link #type() type()}'s properties hold for 'this' Graph. */
   public boolean checkProperties() { return type().checkProperties(this); }

   public       boolean   isDirected() { return type().isDirected; }

   /** &not;{@link #isDirected() isDirected()}. */
   public final boolean isUndirected() { return ! isDirected(); }

   /** Are self-loops &lang;v,&nbsp;v&rang; <I>allowed</I>?
       Not, does this Graph actually have any self-loops,
       but rather <I>could</I> it have a self-loop? */
   public boolean selfLoops() { return type().selfLoops; }

   /** The number of Vertices, |V|&ge;1,
       <NOBR>V = {v<SUB>0</SUB>, ..., v<SUB>(vSize()-1)</SUB>}</NOBR>,
       of 'this' Graph.  The "null Graph" (|V|=0) is usually considered
       not to be a Graph. Also see {@link #eSize() eSize()}.  */
   public abstract int vSize();                                    // * vSize()

   /** The number of Edges in 'this' Graph, |E|&ge;0. By the way,
       the "empty Graph" (|E|=0), for a given |V|&ge;1, is a
       perfectly good Graph. Also see {@link #vSize() vSize()}. */
   public int eSize()                                                // eSize()
    { int sum = 0;
      if( isDirected() )
       { for( int v = 0; v < vSize(); v ++ ) sum += outDegree(v);
         return sum;               // ok for (v,w) and for (v,v)
       }// else isUndirected ...
      for( int v = 0; v < vSize(); v ++ ) sum += degree(v);
      assert (sum & 1) == 0 : ".eSize(), sum=" + sum;
      return sum / 2;
    }//eSize()

   /** Return Edge statistics, [#non_Edges,&nbsp;#Edges], as doubles.
       That is [{@link #maxEdges}-eSize(),&nbsp;{@link #eSize()}],
       taking Directed or not, and loops or not, into account.
       E.g.,&nbsp;useful for an edge-density model (hence doubles).  */
   public double [] eStats()                                        // eStats()
    { final int Esz = eSize();
      assert Esz <= maxEdges() : ".eStats(), "+Esz+" edges, >"+maxEdges();
      return new double [] {maxEdges() - Esz, Esz};
    }//eStats()

   /** The maximum possible number of Edges in a Graph with 'this'
       Graph's {@link #type() type} and {@link #vSize() vSize}. */
   public int maxEdges()                                          // maxEdges()
    { return type().maxEdges(vSize()); }

   /** Given Vertices v0 and v1 that <I>could</I> form
       a legitimate Edge, return a unique integer that is
       &ge;0 and &lt;{@link #maxEdges() maxEdges()}.
       This can be done in an arbitrary (but consistent) way. For an
       Undirected Graph, <NOBR>vPair2n(v, w) = vPair2n(w, v).</NOBR> */
   public int vPair2n(int v0, int v1)                         // vPair2n(v0,v1)
    { final boolean sl = selfLoops();
      if( ! sl && v0 == v1 ) error( "vPair2n("+v0+","+v1+"), v0=v1" );
      final int diag_out = sl ? 0 : 1;
      final int    width = vSize() - diag_out;
      if( isDirected() )
         return v0*width + v1 - (v1 > v0 ? diag_out : 0);
      // else isUndirected() ...
      if( v0 > v1 ) { final int temp = v0; v0 = v1; v1 = temp; } // v0 <= v1
      return ((v0*(2*width-v0+1)) >>> 1) + v1-v0 - diag_out;
    }//vPair2n(v0,v1)

   /** Return an int that is "likely" to differ for different
       Vertices v and v', and that can be computed "quickly".
       These aims are in conflict, of course.  If H is a
       renumbered G, and if w in H corresponds to v in G,
       then H.localHash(w)&nbsp;=&nbsp;G.localHash(v).
       Also see {@link #hashCode() hashCode()}. */
   public int localHash(int v)                                  // localHash(v)
    { return isDirected()
           ? la.util.Util.interleave16(inDegree(v), outDegree(v))
           : degree(v);
    }//localHash(v)

   /** Also see {@link Directed#degree} and {@link Undirected#degree}.
       A sub-class may have a more efficient implementation,
       e.g., see {@link Undirected.Sparse#degree}. */
   public int degree(int v)                                        // degree(v)
    { if( isDirected() ) return inDegree(v) + outDegree(v);
      // else isUndirected() ...
      Series.Int js = joinedTo(v);
      js.exhaust();  // i.e., iterative
      return js.position() + (isEdge(v,v) ? 1 : 0); // NB. loop?
    }//degree(v)

   /** Only applicable to {@link #isDirected Directed} Graphs.
       The in-degree of Vertex&nbsp;'v'.  Calls {@link #directPredecessors}.
       A sub-class may have a more efficient implementation,
       e.g., see {@link Directed.Sparse#inDegree}. */
   public int inDegree(int v)                                  // ? inDegree(v)
    { Series.Int vp = directPredecessors(v); // only for Directed!
      vp.exhaust();  // i.e., iterative
      return vp.position();
    }//inDegree(v)

   /** Only applicable to {@link #isDirected Directed} Graphs.
       The out-degree of Vertex&nbsp;'v'.  Calls {@link #directSuccessors}.
       A sub-class may have a more efficient implementation,
       e.g., see {@link Directed.Sparse#outDegree}. */
   public int outDegree(int v)                                // ? outDegree(v)
    { Series.Int vs = directSuccessors(v); // only for Directed!
      vs.exhaust();  // i.e., iterative
      return vs.position();
    }//outDegree(v)


   /** Is &lang;v0,&nbsp;v1&rang; an Edge in 'this' Graph?
       Also see {@link #adjacent adjacent(v0,v1)}.  */
   public abstract boolean isEdge(int v0, int v1);           // * isEdge(v0,v1)

   /** v0 and v1 are {@link #adjacent adjacent} if &lang;v0,&nbsp;v1&rang;
       is an Edge ({@link #isEdge isEdge(v0,v1)}) <I>or</I> if
       &lang;v1,&nbsp;v0&rang; is. (Which somewhat clashes with the term
       {@link #A() Adjacency&nbsp;Matrix} for a Directed Graph.) */
   public boolean adjacent(int v0, int v1)                   // adjacent(v0,v1)
    { return isEdge(v0, v1) || isEdge(v1, v0); }

   /** Are the Vertices labelled? Also see {@link #vLabel vLabel(v)}. */
   public boolean vLabelled() { return type().vType != null; }

   /** Are the Edges labelled? Also see {@link #eLabel eLabel(v0,v1)}. */
   public boolean eLabelled() { return type().eType != null; }

   /** UnsupportedOperation, the <I>default</I> assumption is no
       Vertex labels.  Also see {@link #vLabelled vLabelled()} and
       {@link #eLabel eLabel(v0,v1)}.  */
   public Value vLabel(int v)                                    // ? vLabel(v)
    { throw UOE( ".vLabel(" + v + ") not supported, type=" + type()
               + (v < 0 || v >= vSize() ? " (& given v not a Vertex)" : "") );
    }//vLabel(v)

   /** Return all Vertex labels, or null if unlabelled. */
   public Value [] vLabels()                                       // vLabels()
    { if( ! vLabelled() ) return null;
      final int V = vSize();
      final Value [] ls = new Value [V];
      for( int i = 0; i < V; i ++ ) ls[i] = vLabel(i);
      return ls;
    }//vLabels()

   /** UnsupportedOperation, the <I>default</I> assumption is no
       Edge labels.  Also see {@link #eLabelled eLabelled()} and
       {@link #vLabel vLabel(v)}.  */
   public Value eLabel(int v0, int v1)                       // ? eLabel(v0,v1)
    { throw UOE( ".eLabel(" + v0 + "," + v1 + ") not supported, type=" + type()
               + (isEdge(v0,v1) ? "" : " (& given (v0,v1) not an Edge)") );
    }//eLabel(v0,v1)

   /** Return all Edge labels, or null if unlabelled. */
   public Value [] [] eLabels()                                    // eLabels()
    { if( ! eLabelled() ) return null;
      final int V = vSize();
      final Value [] [] ls = new Value [V] [V];
      for( int i = 0; i < V; i ++ )
         for( int j = 0; j < V; j ++ )
            ls[i][j] = isEdge(i, j) ? eLabel(i, j) : null;
      return ls;
    }//eLabels()

   /** Return the Adjacency Matrix (of {0,1}s) of 'this' Graph.
       The Matrix is symmetric for an {@link Undirected Undirected} Graph
       but not in general for a {@link Directed Directed} Graph
       (which somewhat clashes with {@link #adjacent adjacent(u,v)} usage).
       Also see {@link #arrayA() arrayA()}.  */
   public Matrix.Ints A()                                                // A()
    { return new Matrix.Ints(/* nCols= */vSize()) // i.e., square
       { public int nElts() { return vSize(); }
         public Value elt(int i, int j)
          { return new Int(isEdge(i, j) ? 1 : 0); }
    }; }//A()

   /** Return the Adjacency Matrix as a square <TT>int[][]</TT>.
       Also see {@link #A() A()} and {@link Undirected#upperRightA()}. */
   public int [] [] arrayA()                                        // arrayA()
    { final int V = vSize();
      final int a [] [] = new int [V] [V];
      for( int u = 0; u < V; u ++ )
         for( int v = 0; v < V; v ++ )
            a[u][v] = isEdge(u, v) ? 1 : 0;
      return a;
    }//arrayA()

   /** Return the edges, [..., <NOBR>(e[i][0], e[i]][1]),</NOBR> ...] of
      'this' Graph as a E&times;2 array suitable for {@link Directed.Sparse}
      or {@link Undirected.Sparse}, say. The edges will be in increasing
      lexicographical order. Also see {@link #A() A()}.  */
   public int [] [] edges()                                          // edges()
    { int [] [] es = new int [eSize()] [2];  int n = 0;
      if( isDirected() )
         for( int v = 0; v < vSize(); v ++ )
          { Series.Int vs = directSuccessors(v);
            while( vs.hasSome() )
             { es[n][0] = v; es[n][1] = vs.elt_n(); vs.advance(); n ++ ;
             }
          }//for v
      else // Undirected
         for( int v = 0; v < vSize(); v ++ )
          { Series.Int vs = joinedTo(v);
            while( vs.hasSome() )
             { int w = vs.elt_n();
               if( w < v ) continue; // don't double count
               es[n][0] = v; es[n][1] = w; vs.advance(); n ++ ;
             }
          }//for v
      return es;
    }//edges()

   /** Return a randomised Graph with the same (in- and out-) degree
       distribution(s) as 'this' Graph. The result is efficient for
       Sparse Graphs.  */
   public Graph randomised()                                    // randomised()
    { java.util.BitSet eSet = new java.util.BitSet(maxEdges());
      int [] [] es = edges();
      if( es.length <= 0 ) return this;            // no Edges
      for( int i = 0; i < es.length; i ++ )
         eSet.set(vPair2n(es[i][0], es[i][1]));
      for( int i = 1; i < es.length; i ++ )        // shuffle e[.][1] ...
       { final int j = (int) Math.floor((i + 1) * mml.MML.RNG.nextDouble());
         // after Durstenfeld, Alg. 235 Random permutation, 1964. 0 <= j <= i
         final int [] es_i = es[i], es_j = es[j];     // 0 <= j < i
         final int es_i0 = es_i[0], es_i1 = es_i[1],  // ith Edge
                   es_j0 = es_j[0], es_j1 = es_j[1];  // jth Edge
         if( es_i0 == es_i1 || es_j0 == es_j1 ) continue; // loops now
         if( es_i0 == es_j1 || es_j0 == es_i1 ) continue; // would be loops
         final int ij = vPair2n(es_i0, es_j1), ji = vPair2n(es_j0, es_i1);
         if( eSet.get(ij) || eSet.get(ji) ) continue; // Don't duplicate edge
         es_i[1] = es_j1; es_j[1] = es_i1;            // Swap
         eSet.clear(vPair2n(es_i0,es_i1));            // Ex-
         eSet.clear(vPair2n(es_j0,es_j1));            // -Edges
         eSet.set(ij);    eSet.set(ji);               // New Edges
       }//for i
      return isDirected() ? new Directed.Sparse(type(), vSize(), es)
                        : new Undirected.Sparse(type(), vSize(), es);
    }//randomised()

   /** An <I>increasing</I> Series of those Vertices 'w'
       {@link #adjacent adjacent} to v, that is (v,w) is an Edge or
       (w,v) is (includes v, once, if there is a loop (v,v)).
       Note,&nbsp;this default is inefficient for sparse Graphs but
       also see {@link Directed.Sparse#joinedTo} and
       {@link Undirected.Sparse#joinedTo}.  */
   public Series.Int joinedTo(final int v)                       // joinedTo(v)
    { return new Series.Int()
       { private int w = -1, pos = -1;
         { advance(); }//constructor
         public int       elt_n() { return w; }
         public int    position() { return pos; }
         public boolean hasSome() { return w < vSize(); }
         public void    advance()
          { if( hasNone() ) error( ".advance(), hasNone" );
            do{ w ++ ; }while( w < vSize() && ! adjacent(w, v) );
            // w >= vSize || adjacent(w,v)
            pos ++ ;  // 0, 1, 2, ...
          }//advance()
         public String toString() { return "joinedTo(" + v + ")"; }
    }; }//joinedTo(v)

   /** Only applicable to {@link #isDirected Directed} Graphs.
       An <I>increasing</I> Series of those Vertices 'p' such that
       &lang;p,v&rang; is an Edge (includes v if there is a loop
       &lang;v,v&rang;).  Note,&nbsp;this default is inefficient for sparse
       Graphs but see {@link Directed.Sparse#directPredecessors} */
   public Series.Int directPredecessors(final int v)                       // ?
    { if( isUndirected() )
         error( ".directPredecessors(" + v + ") of undirected" );
      return new Series.Int()
       { private int p = -1, pos = -1;
         { advance(); }//constructor
         public int       elt_n() { return p; }
         public int    position() { return pos; }
         public boolean hasSome() { return p < vSize(); }
         public void    advance()
          { if( hasNone() ) error( ".advance(), hasNone" );
            do{ p ++ ; }while( p < vSize() && ! isEdge(p, v) );
            pos ++ ;  // 0, 1, 2, ...
          }//advance()
         public String toString() { return "directPredecessors(" + v + ")"; }
    }; }//directPredecessors(v)

   /** Only applicable to {@link #isDirected Directed} Graphs.
       An <I>increasing</I> Series of those Vertices 's' such that
       &lang;v,s&rang; is an Edge (includes v if there is a loop
       &lang;v,v&rang;). Note,&nbsp;this default is inefficient for sparse
       Graphs but see {@link Directed.Sparse#directSuccessors}. */
   public Series.Int directSuccessors(final int v)                         // ?
    { if( isUndirected() )
         error( ".directSuccessors(" + v + ") of undirected" );
      return new Series.Int()
       { private int s = -1, pos = -1;
         { advance(); }//constructor
         public int       elt_n() { return s; }
         public int    position() { return pos; }
         public boolean hasSome() { return s < vSize(); }
         public void    advance()
          { if( hasNone() ) error( ".advance(), hasNone, s=" + s );
            do{ s ++ ; }while( s < vSize() && ! isEdge(v, s) );
            pos ++ ;  // 0, 1, 2, ...
          }//advance()
         public String toString() { return "directSuccessors(" + v + ")"; }
    }; }//directSuccessors(v)


   /** A Series of {@link #lo lo}- to {@link #hi hi}-sized, (weakly-)
       connected, {@link Induced Induced} subGraphs of 'this' Graph.
       Note, 'lo' to 'hi' inclusive. Also see {@link #subGraphs(int,int)}. */
   public class SubGraphs extends Series                     // Graph.SubGraphs
    { /** subGraphs of Vertex-size lo to hi inclusive. */
      public  final int lo, hi;        // size bounds
      private final int [] vs;         // Vertices
      private final BitSet [] ns, fs;  // free and neighbour sets
      private int tos, pos;            // top of stack & position in series
      private Graph.Induced G;         // current elt() Graph
      private final BitSet vertices;   // Vertices in G

      /** {@link #lo lo} and {@link #hi hi} inclusive are bounds on the
          vertex-size of the subGraphs in the Series. */
      public SubGraphs(int lo, int hi)
       { this.lo = lo; this.hi = hi;
         assert lo > 0 && hi >= lo : " lo="+lo + ", hi="+hi;
         vs = new int    [hi];   // stacks...
         fs = new BitSet [hi];
         ns = new BitSet [hi];
         vs[0] = -1;
         tos =  0;
         pos = -1;
         vertices = new BitSet(Graph.this.vSize()); // { }
         advance();              // !
       }//constructor

      /** Is there at least a {@link #elt() current} Graph
          remaining in the Series? */
      public boolean hasSome() { return vs[0] < vSize(); }

      /** Return the position, p, of the current subGraph in the Series,
          or how many were generated if {@link #hasNone hasNone()}.
          Counting starts at zero.  */
      public int    position() { return pos; }

      /** Return the current subGraph in 'this' Series if
          {@link #hasSome() hasSome()}. After
          {@link #advance() advance()}, a different (disjoint)
          <I>data-structure</I> will be returned although the current
          and next subGraphs may well overlap.  */
      public Induced elt()
       { assert hasSome() : ".elt(), none left, position()=" + pos;
         if( G == null )
          { int [] ans = new int [tos];
            for( int i = 0; i < tos; i ++ )
             { final int v = vs[i];
               int j = i;
               for( /* ! */ ; j > 0; j -- )  // ans[] sorted...
                  if( v >= ans[j-1] ) break; else ans[j] = ans[j-1];
               ans[j] = v;
             }//for i
            G = new Induced(ans);  // cache the Graph
          }
         return G;
       }//elt()

      /** Return the set of Vertices (of 'this' Graph) that are in
          the {@link #elt() current} subGraph of 'this' Series.
          Beware: The result BitSet is changed by {@link #advance()}.
          And on <EM>no account</EM> change its contents!  */
      public BitSet vertices()
       { if( vertices.cardinality() != tos )                // ZZZ paranoia
            error( ".vertices(), " + vertices.cardinality() + "!=" + tos );
         return vertices;
       }

      /** Move on to the next {@link #elt() subGraph} in the Series.
          If there are no more {@link #elt() elt()} becomes invalid.  */
      public void advance()
       { if( hasNone() ) error( ".advance(), none left, position()=" + pos );
         pos ++ ;  // 0, 1, 2, ...
         G = null; // cache is now invalid
         while( true )
          { if( tos == 0 )
             { vs[0] ++;                       // the "seed" Vertex
               if( vs[0] >= vSize() ) return;  // No subGraphs remain
               vertices.set(vs[0]);            // set of Vertices {vs[0]}
               fs[0] = new BitSet(vSize());    // free ...
               fs[0].set(vs[0]+1, vSize());    // all v > vs[0]
               ns[0] = new BitSet(vSize());    // neighbours ...
               Series.Int js = joinedTo(vs[0]);
               while( js.hasSome() )           // ... of vs[0]
                { ns[0].set(js.elt_n()); js.advance(); }
               tos = 1;                        // unlikely, but ...
               if( tos >= lo ) return;         // have a big enough subGraph
             }
            else /* tos > 0 */
            if( tos < hi )   // try to extend subGraph
             { // vs[0] ... vs[tos-1] is a part SubGraph
               BitSet eligible = (BitSet) fs[tos-1].clone();
               eligible.and( ns[tos-1] );      // free * ns[tos-1]
               int v = eligible.nextSetBit(0); // Vertex v
               if( v >= 0 )  // there is such a 'v'
                { fs[tos-1].set(v, false);     // free - {v}
                  BitSet n_v  = (BitSet) ns[tos-1].clone(); // + ...
                     Series.Int js = joinedTo(v); // ... v's neighbours, ...
                     while( js.hasSome() )
                      { n_v.set(js.elt_n()); js.advance(); }
                  vertices.set(v);  // + {v}
                  vs[tos] = v;
                  fs[tos] = (BitSet) fs[tos-1].clone();
                  ns[tos] = n_v;
                  tos ++ ;
                  if( tos >= lo ) return;      // have a big enough subGraph
                  // else loop and try to extend
                }
               else // v < 0, no such v, fall back a level
                { tos -- ; vertices.clear(vs[tos]); }
             }
            else // tos == hi, cannot extend, fall back a level
             { tos -- ; vertices.clear(vs[tos]); }
          }//while
       }//advance()
    }//SubGraphs class, the Series

   /** {@link #subGraphs(int,int) subGraphs(1, hi)}. */
   public SubGraphs subGraphs(int hi)
    { return subGraphs(1, hi); }

   /** Return a {@link SubGraphs Series} of all Vertex-size lo to hi,
       (weakly-) connected, Vertex-induced subgraphs of 'this' Graph. */
   public SubGraphs subGraphs(int lo, int hi)
    { return new SubGraphs(lo, hi); }
   // .........................................................................

   /** Calls {@link   Directed#dense(Type,la.maths.Matrix)}
       or    {@link Undirected#dense(Type,la.maths.Matrix)}
       according to 't'.  Return an unlabelled Graph of
       {@link Type Type} 't' and Adjacency Matrix&nbsp;'A'.
       For a    Directed Graph, A must be square.
       For an Undirected Graph, A must be square symmetric. */
   public static Graph dense(final Type t, Matrix A)
    { return t.isDirected ? Directed.dense(t,A) : Undirected.dense(t,A); }

   /** Calls {@link   Directed#dense(Type,int[][])}
       or    {@link Undirected#dense(Type,int[][])} according to t.
       For a    Directed Graph, A must be square.
       For an Undirected Graph, A must be upper-right <I>triangular</I>. */
   public static Graph dense(final Type t, final int[][] A)
    { return t.isDirected ? Directed.dense(t,A) : Undirected.dense(t,A); }

   /** Return a Graph of equivalent structure but with any trail of
       inducing, renumbering, etc., fixed and the trail lost.
       This implementation is inefficient for large Sparse Graphs.  */
   public Graph flatten()                                          // flatten()
    { final Graph g = this;
      return new Graph()
       { final Type      t = g.type();
         final int       V = g.vSize();
         final int [] [] A = g.arrayA();  // adjacency (array) matrix
         final Value []    vLabels = g.vLabels();
         final Value [] [] eLabels = g.eLabels();
         public Type                  type() { return t; }
         public int                  vSize() { return V; }
         public boolean isEdge(int u, int v) { return A[u][v] == 1;  }
         public Value          vLabel(int v) { return vLabels[v];    }
         public Value   eLabel(int u, int v) { return eLabels[u][v]; }
         public Matrix.Ints              A() { return Matrix.ints(A); } // !
         public int [] []           arrayA() { return A; }        // !
         public Value []           vLabels() { return vLabels; }  // !
         public Value [] []        eLabels() { return eLabels; }  // !
         public Graph              flatten() { return this; }     // !
         public String toString() { return Matrix.ints(A).toString(); }
    }; }

   /** Provided 'this' Graph is in fact Directed, return
       it as an instance of {@link ToDirected that&nbsp;class}.
       Also see {@link #toUndirected() toUndirected()}. */
   public Directed toDirected()                                 // toDirected()
    { return new ToDirected(); }

   /** Provided 'this' Graph is in fact Directed, be
       an instance of {@link Directed that&nbsp;class}.
       Also see {@link ToUndirected ToUndirected}. */
   protected class ToDirected extends Directed                    // ToDirected
                                      implements Child
    { public Graph parent() { return Graph.this; }
      public ToDirected()
       { if( isUndirected() ) error( "ToDirected(), type=" + type() );
         assert checkProperties() : this + " fails " + type();
       }//constructor
      public Type type() { return Graph.this.type();  }
      public int vSize() { return Graph.this.vSize(); }
      public boolean isEdge(int v0, int v1)
       { return Graph.this.isEdge(v0, v1); }
      public Value vLabel(int v) { return Graph.this.vLabel(v); }
      public Value eLabel(int v0, int v1)
       { return Graph.this.eLabel(v0, v1); }
      public String toString()
       { return Graph.this.toString() + ".toDirected()"; }
    }//ToDirected() class


   /** Provided 'this' Graph is in fact Undirected, return
       it as an instance of {@link ToUndirected that&nbsp;class}.
       Also see {@link #toDirected() toDirected()}. */
   public Undirected toUndirected()                           // toUndirected()
    { return new ToUndirected(); }

   /** Provided 'this' Graph is in fact Undirected, be
       an instance of {@link Undirected that&nbsp;class}.
       Also see {@link ToDirected ToDirected}. */
   protected class ToUndirected extends Undirected              // ToUndirected
                                        implements Child
    { public Graph parent() { return Graph.this; }
      public ToUndirected()
       { if( isDirected() ) error( "ToUndirected(), type=" + type() );
         assert checkProperties() : this + " fails " + type();
       }//constructor
      public Type type() { return Graph.this.type();  }
      public int vSize() { return Graph.this.vSize(); }
      public boolean isEdge(int v0, int v1)
       { return Graph.this.isEdge(v0, v1); }
      public Value vLabel(int v) { return Graph.this.vLabel(v); }
      public Value eLabel(int v0, int v1)
       { return Graph.this.eLabel(v0, v1); }
      public String toString()
       { return Graph.this.toString() + ".toUndirected()"; }
    }//ToUndirected class


   /** Is 'this' Graph structurally identical to 'g',
       <NOBR>this.v<SUB>i</SUB> : g.v<SUB>i</SUB></NOBR>,
       in terms of the <I>existence</I> of Edges?
       Checks vSize() of 'this' and 'g' match, then calls
       {@link #edgesCorrespond edgesCorrespond(g)}.
       Can be used to test for {@link #isomorphic isomorphism} of g1 and g2
       <I>provided</I> g1 and g2 are both {@link #canonical canonical}. */
   public boolean structurallyIdentical(Graph g)    // structurallyIdentical(g)
    { if( isDirected() != g.isDirected() ) error( type() + " v. " + g.type() );
      if( vSize() != g.vSize() ) return false;
      return edgesCorrespond(g);
    }//structurallyIdentical(g)

   /** <I>Assumes</I> this.vSize()&nbsp;=&nbsp;g.vSize()
       and gets on with checking Edge correspondence,
       this.v<SUB>i</SUB>&nbsp;:&nbsp;g.v<SUB>i</SUB>.
       Note, it relies on {@link #directSuccessors} and {@link #joinedTo}
       generating Vertices in ascending order.
       Time complexity "ought" to be O(|E|).
       Also see {@link #isomorphic isomorphic(g)} and
       {@link #structurallyIdentical structurallyIdentical(g)}.  */
   public boolean edgesCorrespond(Graph g)
    { assert vSize() == g.vSize() : "|V|: " + vSize() + " v. " + g.vSize();
      for( int v = 0; v < vSize(); v ++ )
       { Series.Int /*this*/ vs, /*g*/ gs;
         if( isDirected() )  // doing successors is enough!...
          { vs = directSuccessors(v); gs = g.directSuccessors(v); }
         else // isUndirected
          { vs = joinedTo(v); gs = g.joinedTo(v); }
         while( vs.hasSome() )
          { if( gs.hasNone() ) return false;
            if( vs.elt_n() != gs.elt_n() ) return false; // NB., assumes
            vs.advance(); gs.advance();                  // same ordering
          }
         if( gs.hasSome() ) return false;
       }//for v
      // { for( int w = isDirected() ? 0 : v; w < vSize(); w ++ )
           // if( isEdge(v, w) != g.isEdge(v, w) ) return false;
      // }//for v
      return true;
    }//edgesCorrespond(g)

   /** Is 'this' Graph isomorphic to 'g' (in terms of the existence of
       Edges)?  It makes use of {@link #canonical canonical()} and
       {@link #edgesCorrespond edgesCorrespond(g)}.  */
   public boolean isomorphic(Graph g)                          // isomorphic(g)
    { if( isDirected() != g.isDirected() ) error( type() + " v. " + g.type() );
      // following tests faster than canonical(), so probably worth it
      if( vSize() != g.vSize() || hashCode() != g.hashCode() ) return false;
      // minimum required: if( vSize() != g.vSize() ) return false;
      return canonical().edgesCorrespond(g.canonical()); // heavy compute
    }//isomorphic(g)

   /** Return a canonical {@link Renumbered Renumbering} of 'this' Graph,
       that is the one yielding the largest adjacency matrix when
       it is read as a binary number, in a "certain order".
       The renumbering is not unique if the Graph has non-trivial
       automorphisms. The implementation is adequate for small Graphs.
       Currently calls {@link #canonical2 canonical2(.)}.
       Also see {@link #isomorphic isomorphic(g)} and
       {@link #nAutomorphisms() nAutomorphisms()}.  */
   public Canonical canonical()                                  // canonical()
    { return canonical2(/*dont care*/ new RefInt()); }
    // { return canonical1(/*dont care*/ new RefInt()); }

   /** The number of automorphisms of 'this' Graph.
       Currently calls {@link #canonical2 canonical2(.)}.
       Also see {@link #canonical() canonical()}.  */
   public int nAutomorphisms()                              // nAutomorphisms()
    { RefInt nA = new RefInt();
      canonical2( nA );
      // canonical1( nA );
      return nA.n;
    }//nAutomorphisms()


    /** Does the hard work of finding a 'bestG' -- 'this' Graph Renumbered --
        for {@link #canonical() canonical()}. It partitions the set of
        Vertices on their (quick) {@link #localHash localHas(v)} property
        so as to reduce the number of permutations considered. It then calls
        {@link #canonical2R canonical2R(...)} to do the really hard work.
        And, 'nAuto' returns the number of automorphisms discovered.  */
   protected Canonical canonical2(RefInt nAuto)               // canonical2(nA)
    { final int V = vSize();
      final int [] vs   = new int[ V ],   // Vertices, ordered in ns[.]
                   cut  = new int[ V+1 ], // cut points, usually part full
                   best = new int[ V ],   // best permutation so far
                   perm = new int[ V ];   // working permutation
      int [] ns = best;  // temporarily, see later!
      for( int v = 0; v < V; v ++ )  // each Vertex 'v'
       { int j = v;  final int n_v = localHash(v);
         for( ; j > 0 && n_v > ns[j-1]; j -- ) // NB "stable"
          { vs[j] = vs[j-1]; ns[j] = ns[j-1]; }
         vs[j] = v; ns[j] = n_v;   // j is the position for v
       }//for v;  ns[0..V-1] is now monotonic decreasing.

      int nParts = 0;  // ?How many parts to the partition of [0,V)?
      cut[0] = 0;
      for( int i = 1; i < V; i ++ )
         if( ns[i-1] != ns[i] ) { nParts ++ ; cut[nParts] = i; }// a step
      nParts ++ ; cut[nParts] = V;  // the 'parts' of ==localHash(,) are...
      // [cut[0]=0,cut[1]), [cut[1],cut[2]), ..., [cut[nParts-1],cut[nParts]=V)
      ns = null;  // 'ns' no longer needed

      for( int i = 0; i < V; i ++ )
         best[i] = vs[i];  // the 1st guess at 'best' renumbering
      Renumbered permG = new Renumbered( perm ),     // current attempt ...
                 bestG = new Renumbered( best );     // ... best to date
      BitSet free = new BitSet(V); // = empty set!
      canonical2R( vs, cut, /*part*/ 0,
                   0, 0, free, permG, bestG, nAuto); // recursive search
      // obsolete as of 2017: ??? return bestG;
      return new Canonical( best );
    }//canonical2(nA)


   /** Does the hard work of finding a 'bestG' -- 'this' Graph Renumbered --
       for {@link #canonical2 canonical2(nAuto)}.  Recursively enumerates
       renumberings, permG, in lexicographically increasing order while
       trying to "short cut" when it can.  Hence, it comes across the
       lexicographically least representative of each equivalence
       class first.  */
   protected boolean canonical2R(int[] vs, int[] cut, int part,   // ...2R(...)
                                 int level, int diff, BitSet free,
                                 Renumbered permG, Renumbered bestG,
                                 RefInt nAuto )
    { // enumerates renumberings, permG, in lexicographically increasing order,
      // hence it comes across the lex.least rep. of each equiv.class first.
      // assert: 0 <= level <= vSize and diff >= 0
      final int V = vSize();
      boolean improved = false;
      if( level < V )                // make a choice for permG.vs[level]
       { int partL = part;  // see below

         if( level == cut[part] )    // start on new part of Vertex partition
          { // now juggle vs[cut[part], cut[part+1]) ...
            if( free.cardinality() != 0 ) error( "NOT_PARANOID!" ); // ZZZ
            assert free.cardinality() == 0
                 : "|free|=" + free.cardinality()
                 + " part=" + part + " level=" + level;
            for( int i = cut[part]; i < cut[part+1]; i ++ ) free.set(vs[i]);
            partL ++ ;
          }//fi

         for( int choice = free.nextSetBit(0); choice >= 0;
                  choice = free.nextSetBit(choice+1) )
          { permG.vs[ level ] = choice;
            int diffL = improved ? 0 : diff; // NB. diff>0 => already better

            for( int i = 0; diffL == 0 && i <= level; i ++ ) // extend...
             { diffL = (permG.isEdge(i, level) ? 1 : 0)      // check for...
                     - (bestG.isEdge(i, level) ? 1 : 0);     // improvement
                                                             // 1 2 4
                                                             // - 3 5 scan if
                                                             // - - 6 Undir.
             }//for i, i.e, extension of the check

            if( isDirected() )
            for( int i = 0; diffL == 0 && i < level; i ++ )  // NB. '<'
             { diffL = (permG.isEdge(level, i) ? 1 : 0)      // 1 2 5...
                     - (bestG.isEdge(level, i) ? 1 : 0);     // 4 3 6... scan
                                                             // 8 9 7... order
             }//for i, i.e, extension of the check

            if( diffL < 0 ) continue; // permG is worse, oh well, try again
          //else, note diffL>=0, at least as good as bestG so far, so...

            free.clear(choice);
            improved |= canonical2R(vs, cut, partL,
                                   level+1, diffL, free, permG, bestG,
                                   nAuto);  // RECN
            free.set(choice);  // restore
            // NB. improved => bestG.vs[0..level]==permG.vs[0..level] now;
            //     note recursive call will have changed bestG !
          }//for choice

         if( level == cut[part] ) free.clear();  // restore

       }
      else // level == V, a full renumbering, a leaf of the search
       { if( diff > 0 ) // have a new best
          { improved = true;
            for( int i = 0; i < V; i ++ ) bestG.vs[i] = permG.vs[i];
            nAuto.n = 1;  // as far as we know, so far
          }
         else // diff == 0, i.e., a lex. greater perm of the best ...
            nAuto.n ++ ;  // equiv.class (so far) & hence another
                                   // (tentative) automorphism of G.
       }// level : V
      return improved;  // did we?
    }//canonical2R(...)


   /** Like {@link #canonical() canonical()} but
       also counting the number of automorphisms.
       Does some initialisation and then calls {@link #canonical1R}.
       ???Superceded by canonical2(nA)???  */
   protected Graph canonical1(RefInt nAuto)                // canonical1(nAuto)
    { final int V = vSize();
      final int [] perm = new int[ V ], best = new int[ V ];
      for( int i = 0; i < V; i ++ ) best[i] = i;    // identity perm
      final BitSet free = new BitSet(V);            // free Vertices
      free.set(0, V);                               // all
      Renumbered permG = new Renumbered( perm ),    // current attempt ...
                 bestG = new Renumbered( best );    // ... best to date
      canonical1R(0, 0, free, permG, bestG, nAuto); // recursive search
      return bestG;
    }//canonical1(nAuto)

   /** Does the hard work of finding a 'bestG' -- 'this' Graph Renumbered --
       for {@link #canonical1 canonical1}.  Note, one might initialise
       bestG.vs[] with a "good bet" suggested by some heuristic -
       speeding the search?  ???Superceded by canonical2R(nA)???  */
   protected boolean canonical1R( int level, int diff, BitSet free,
                                  Renumbered permG, Renumbered bestG,
                                  RefInt nAuto )
    { // enumerates renumberings, permG, in lexicographically increasing order,
      // hence it comes across the lex.least rep. of each equiv.class first.
      // assert: 0 <= level <= vSize and diff >= 0
      final int V = vSize();
      boolean improved = false;
      if( level < V )                        // make a choice for vs[level]
       { for( int choice = free.nextSetBit(0); choice >= 0;
                  choice = free.nextSetBit(choice+1) )
          { permG.vs[ level ] = choice;
            int diffL = improved ? 0 : diff; // NB. diff>0 => already better

            for( int i = 0; diffL == 0 && i <= level; i ++ ) // extend...
             { diffL = (permG.isEdge(i, level) ? 1 : 0)      // check for...
                     - (bestG.isEdge(i, level) ? 1 : 0);     // improvement
                                                             // 1 2 4
                                                             // - 3 5 scan if
                                                             // - - 6 Undir.
             }//for i, i.e, extension of the check

            if( isDirected() )
            for( int i = 0; diffL == 0 && i < level; i ++ )  // NB. '<'
             { diffL = (permG.isEdge(level, i) ? 1 : 0)      // 1 2 5...
                     - (bestG.isEdge(level, i) ? 1 : 0);     // 4 3 6... scan
                                                             // 8 9 7... order
             }//for i, i.e, extension of the check

            if( diffL < 0 ) continue; // permG is worse, oh well, try again
          //else, note diffL>=0, at least as good as bestG so far, so...

            free.clear(choice);
            improved |= canonical1R(level+1, diffL, free, permG, bestG,
                                   nAuto);  // RECN
            free.set(choice);
            // NB. improved => bestG.vs[0..level]==permG.vs[0..level] now;
            //     note recursive call will have changed bestG !
          }//for choice
       }
      else // level == V, a full renumbering, a leaf of the search
       { if( diff > 0 ) // have a new best
          { improved = true;
            for( int i = 0; i < V; i ++ ) bestG.vs[i] = permG.vs[i];
            nAuto.n = 1;  // as far as we know, so far
          }
         else // diff == 0, i.e., a lex. greater perm of the best ...
            nAuto.n ++ ;  // equiv.class (so far) & hence another
                                   // (tentative) automorphism of G.
       }// level : V
      return improved;  // did we?
    }//canonical1R(...)


   /** Return 'this' Graph {@link Renumbered Renumbered} with Vertices in
       decreasing order of {@link #degree(int) degree}.
       Also see {@link #byDegree(boolean) byDegree(boolean)}.  */
   public Graph byDegree()                                        // byDegree()
    { return byDegree(true); }

   /** Return 'this' Graph {@link Renumbered Renumbered} with Vertices in
       decreasing or increasing order of {@link #degree(int) degree}
       as 'descending' is true or false.
       Also see {@link #byDegree() byDegree()}. */
   public Graph byDegree(boolean descending)            // byDegree(descending)
    { final int V = vSize();
      int [] deg = new int [V], vs = new int [V];
      for( int i = 0; i < V; i ++ )
       { final int di = degree(i); deg[i] = descending ? di : - di; }
      for( int i = 0; i < V; i ++ ) // each Vertex 'i'
       { int j = i, d_i = deg[i];     // insertion sort
         for( /* j=i */ ; j > 0 && d_i > deg[vs[j-1]]; j -- ) vs[j] = vs[j-1];
         // either j == 0 or ( j > 0 and d_i <= deg[vs[j-1]] )
         vs[j] = i;
       }
      return new Renumbered(vs);
    }//byDegree(descending)

   /** Return some kind of hash-code <I>invariant</I> under
       Vertex-renumbering.  If G1 and G2 are isomorphic then
       <NOBR> G1.hashCode() == G2.hashCode(), </NOBR> but equal hashCodes do
       <I>not</I> imply G1 and G2 are {@link #isomorphic isomorphic}.  */
   public int hashCode()                                          // hashCode()
    { final int multiplier = 5, V = vSize();
      int [] vn = new int[ V ];
      for( int i = 0; i < V; i ++ ) // fill sorted vn[]
       { int j = i, v_i = localHash(i);
         for( /* j=i */ ; j > 0 && v_i > vn[j-1]; j -- ) vn[j] = vn[j-1];
         // either j == 0 or ( j > 0 and v_i <= vn[j-1] )
         vn[j] = v_i;  // vn[0..i] is monotonic decreasing
       }//for i
      int hc = 0;
      for( int i = 0; i < V; i ++ ) hc = hc * multiplier + vn[i];
      return hc;
    }//hashCode()
   // .........................................................................

   /** A Vertex of 'this' Graph. Also see {@link Edge}. */
   public class Vertex implements Child                         // Graph.Vertex
    { public Graph parent() { return Graph.this; }
      /** The number of this Vertex (not its {@link #vLabel label}, if any). */
      public final int v;
      public Vertex(int v)
       { assert v >= 0 && v < vSize() : ".constructor(" + v + ")";
         this.v = v;
       }//constructor
      public boolean labelled() { return Graph.this.vLabelled(); }
      public Value      label() { return Graph.this.vLabel(v); }
      public String  toString() { return v+""; }
    }//Graph.Vertex class

   /** An Edge &lang;v0,&nbsp;v1&rang; (Directed), or
       (v0,&nbsp;v1) (Undirected, v0&le;v1), of 'this' Graph.
       Also see {@link Vertex}. */
   public abstract class Edge implements Child                  // * Graph.Edge
    { public Graph     parent() { return Graph.this; }
      /** v0 and v1, the Vertices of 'this' Edge. */
      public final int v0, v1;
      /** If 'this' is {@link #isDirected} &lang;v0,v1&rang;, otherwise
          'this' is {@link #isUndirected}, (v<SUB>s</SUB>,v<SUB>b</SUB>)
          where v<SUB>s</SUB> is the smaller and v<SUB>b</SUB>
          is the larger Vertex.  */
      public Edge(int v0, int v1)
       { assert isEdge(v0, v1) : ".constructor(" + v0 + "," + v1 + ")";
         if( isDirected() || (v0 <= v1) )
          { this.v0 = v0; this.v1 = v1; }
         else // Undirected() & v0 > v1
          { this.v0 = v1; this.v1 = v0; } // standard (v, w) where v<=w
       }//constructor
      /** Are (all) Edges labelled in 'this' Graph? */
      public boolean labelled() { return Graph.this.eLabelled(); }
      /** The Edge label, if any. */
      public Value      label() { return Graph.this.eLabel(v0, v1); }
      public abstract String toString();                                   // *
    }//Graph.Edge class
   // .........................................................................

   /** this.new&nbsp;Derived() is <EM>just like</EM> 'this' Graph, but
       a Function on Graphs may return a sub-class of Derived &mdash;
       extend it &mdash; with one or two methods overridden.  Note, a
       Derived Graph need not be (but can be) a {@link SubGraph SubGraph}
       of the {@link Derived#parent() parent()} Graph.
       Note that {@link Graph} is abstract but Derived is not &mdash;
       it sometimes needs a little thought!
       (Pre-2015(?) javac was not happy about the use of Derived.)  */
   // NOTE TO SELF: class Graph is abstract with some abstract methods
   //    but Derived is *not* abstract.
   //    If a new class G2 extends Derived it only has to define any
   //    method that it is changing.
   //    If G2 extends Graph it has to define all Graph's abstract methods.
   public class Derived extends Graph implements Child         // Graph.Derived
    { /** Returns 'Graph.this', that is, returns the
          Graph of which 'this' is a Derivation. */
      public Graph parent()    { return Graph.this; }

      /** The {@link #parent() parent}'s type()
          (often the same). */
      public Type       type() { return Graph.this.type();  }

      /** The {@link #parent() parent}'s vSize()
          (sometimes the same). */
      public int       vSize() { return Graph.this.vSize(); }

   // Note, we must not put eSize()=Graph.this.eSize() here...
   // ...the eSize() inherited from Graph works well enough.

   // ditto  degree(Vertex v), etc.,  vLabel(v), eLabel(v,w) etc.

      /** The {@link #parent() parent}'s isEdge(v0,v1)
          (is very likely to be changed, overridden).  */
      public boolean isEdge(int v0, int v1)
       { return Graph.this.isEdge(v0, v1); }

    }//Graph.Derived class
   // .........................................................................

   /** A SubGraph of a Graph is a {@link Child} and must also provide a
       {@link #v2pv(int) mapping} from its Vertices to those of the
       {@link #parent() parent} Graph.
       Also see {@link Induced}, {@link Renumbered} and {@link Derived}.  */
   public interface SubGraph extends Child                  // * Graph.SubGraph
    { /** Vertex 'v' of 'this' Graph corresponds to v2pv(v)
          of the {@link #parent parent} Graph. */
      public int v2pv(int v);
    }//SubGraph interface

   /** See {@link Sparse}, {@link Directed.Dense} and {@link Undirected.Dense}.
       Basically a "tag" that an algorithm can test for. 'Dense' and 'Sparse'
       are not precise terms.  */
   public static interface  Dense { }                          // * Graph.Dense

   /** See {@link Dense}, {@link Directed.Sparse} and {@link Undirected.Sparse}.
       Basically a "tag" that an algorithm can test for. 'Dense' and 'Sparse'
       are not precise terms.  */
   public static interface Sparse { }                         // * Graph.Sparse
   // .........................................................................

   /** Convenience function.
       Can't(?) have the {@link Induced} return type because of
       {@link   Directed.Sparse#induced} and
       {@link Undirected.Sparse#induced}.  */
   public Graph induced(int [] vs) { return new Induced(vs); }

   /** A SubGraph of 'this' Graph <I>induced</I> by selecting Vertices
       <NOBR>[vs[0], vs[1], ...]</NOBR>, and the Edges between them.
       Also see {@link #induced induced(vs)}. */
   public class Induced extends Derived implements SubGraph    // Graph.Induced
    {
      /** Vertex 'v' of 'this' Induced sub-graph is Vertex vs[v] of
          the {@link #parent() parent()} Graph. */
      public final int [] vs;

      /** vs must be ascending and a subset of the parent Graph's
          {0,&nbsp;...,&nbsp;vSize()-1}.  */
      public Induced(int [] vs)
       { this.vs = vs;  // could do more tests...
         if( vs.length > Graph.this.vSize() ) error( ".constructor" );
       }//constructor

      /** Vertex 'v' of 'this' (Vertex-)Induced sub-graph is Vertex vs[v] of
          the {@link #parent() parent()} Graph. */
      public int v2pv(int v) { return vs[v]; }

      /** Return |{@link #vs vs}|. */
      public int vSize() { return vs.length; }

      /** {@link #parent() parent()}.isEdge({@link #vs vs}[v0],&nbsp;vs[v1]).
          That is, was there a corresponding Edge in the parent Graph?  */
      public boolean isEdge(int v0, int v1)
       { return Graph.this.isEdge(vs[v0], vs[v1]); }

      /** Vertex labels, if any, as per {@link #vs vs} and
          the {@link #parent() parent()} Graph. */
      public Value vLabel(int v) { return Graph.this.vLabel(vs[v]); }

      /** Edge labels, if any, as per {@link #vs vs} and
          the {@link #parent() parent()} Graph. */
      public Value eLabel(int v0, int v1)
       { return Graph.this.eLabel(vs[v0], vs[v1]); }

      public String toString()
       { return parent() + ".induced(" + Vector.ints(vs) + ")"; }
    }//Graph.Induced class
   // .........................................................................

   /** Convenience function. */
   public Contraction contraction(int [] vs) { return new Contraction(vs); }

   /** A Vertex-Contraction of 'this' Graph, collapsing (identifying) Vertices
       <NOBR>[vs[0], vs[1], ...]</NOBR> onto the single Vertex vs[0].
       Vertex-contraction is the opposite of Vertex-expansion.
       Also see {@link #contraction contraction(vs)}. */
   public class Contraction extends Derived                // Graph.Contraction
                            implements SubGraph
    {
      /** The contracted vertices of the {@link #parent() parent} Graph.
          Note, vs <EM>is</EM> in ascending order!
          vs[0] has some special uses.  */
      public final int [] vs;  // ascending!

      /** [vs[0], vs[1], ...] are identified with 'specialV'. */
      private final int specialV;

      /** vs must be a subset of the parent Graph's
          <NOBR>{0, ..., vSize()-1}</NOBR>. */
      public Contraction(int [] vs)
       { this.vs = new int [vs.length];  // ensure sorted, ascending...
         this.vs[0] = vs[0];                    // 0
         for( int i = 1; i < vs.length; i ++ )  // 1, 2, ...
          { final int vs_i = vs[i];
            int j = i;
            for( ; j > 0; j -- )
             { final int vs_j1 = this.vs[j-1];
               if( vs_j1 < vs_i ) break;
               else if( vs_j1 == vs_i )
                  error( " vs[" + i + "]=" + vs_i + " duplicate" );
               else this.vs[j] = vs_j1;  // shuffle up
             } // j == 0 || this.vs[j-1] < vs_i
            this.vs[j] = vs_i;
          }//for i; this.vs[] is ascending
         specialV = this.vs[0];
         if( vs.length > Graph.this.vSize() ) error( ".constructor" );
       }//constructor

      /** The number of Vertices in 'this' Contracted Graph.  For a
          non-trivial contraction this is less than the parent's.  */
      public int vSize() { return Graph.this.vSize() - vs.length + 1; }

      /** The Vertex of the {@link #parent() parent} that 'v'
          corresponds to.  If, say, vertices x&lt;y&lt;z
          (of the parent) are the contracted vertices then
          a Vertex v in the Contraction maps to its old Vertex
          number in the parent, which is the same number
          if v&le;x and is &gt;v otherwise. */
      public int v2pv(int v)
       { //   p: 0 1 2* 3 4 5* 6 7* 8 ~ 0 1 2 3 4 _ 6 _ 8
         // p.C:                        0 1 2 3 4   5   6
         // vs=[2,5,7]; 0->0, 1->1; 2->2; 3->3; 4->4;  5->6;  6->8; 7->9
         // v -> v+m where m > 0 is least s.t. v+m > vs[m]
         // vs[1],vs[2],... "became" equiv to specialV
         if( vs.length < 2 || v < vs[1] ) return v;
         int lo = 0, hi = vs.length;  // v >= vs[lo]
         while( lo < hi-1 )
          { int mid = (lo+hi) / 2;    // lo < mid < hi <= vs.length
            if( v+mid > vs[mid] ) lo = mid;
            else   /* <= */        hi = mid;
          } // lo == hi-1 & v >= vs[lo]
         return v + lo;
       }//v2pv(v)

      /** If v0 and/or v1 is vs[0] to which &gt;1 vertices of
          the parent Graph are contracted, was there an Edge
          from/to <EM>any</EM> of the latter in the parent?  */
      public boolean isEdge(int v0, int v1)
       { final int vp0 = v2pv(v0), vp1 = v2pv(v1);
         if( vp0 == specialV )
            if( vp1 == specialV ) // both v0 & v1 contracted
             { if( ! selfLoops() ) return false;  // !
               for( int i = 0; i < vs.length; i ++ )
                  for( int j = 0; j < vs.length; j ++ )
                     if( Graph.this.isEdge(vs[i], vs[j]) ) return true;
               return false;
             }
            else // vp1 != specialV
             { for( int i = 0; i < vs.length; i ++ )
                  if( Graph.this.isEdge(vs[i], vp1) ) return true;
               return false;
             }
         else // vp0 != specialV
            if( vp1 == specialV )
             { for( int j = 0; j < vs.length; j ++ )
                  if( Graph.this.isEdge(vp0, vs[j]) ) return true;
               return false;
             }
            else /* neither */ return Graph.this.isEdge(vp0, vp1);
       }//isEdge(v0,v1)

      /** Vertex labels, if any, as per {@link #v2pv v2pv} and the
          {@link #parent() parent} Graph.  If v corresponds to &gt;1
          Vertices of the parent, return any one of their labels.  */
      public Value vLabel(int v) { return Graph.this.vLabel(v2pv(v)); }

      /** Edge labels, if any, as per {@link #v2pv v2pv} and
          the {@link #parent() parent} Graph.  Involves an arbitrary
          choice if v0 and/or v1 corresponds to &gt;1
          Vertices of the parent Graph.  */
      public Value eLabel(int v0, int v1)
       { return Graph.this.eLabel(v2pv(v0), v2pv(v1)); }

      public String toString()
       { return parent() + ".contraction(" + Vector.ints(vs) + ")"; }
    }//Graph.Contraction class
   // .........................................................................

   /** Convenience function. */
   public Graph renumbered(int [] vs) { return new Renumbered(vs); }

   /** A version of 'this' Graph with the Vertices renumbered
       according to {@link #vs vs[]}.
       Also see {@link #renumbered renumbered}. */
   public class Renumbered extends Derived implements SubGraph
    { /** Vertex 'v' of 'this' is Vertex vs[v] of the
          {@link #parent() parent()} Graph.  vs&nbsp;must be a
          permutation of <NOBR>[0, ..., vSize()-1]</NOBR>.  */
      public final int [] vs;
      /** vs must be a permutation of <NOBR>[0, ..., vSize()-1]</NOBR>. */
      public Renumbered(int [] vs)
       { this.vs = vs;  // could do more tests...
         if( vs.length != vSize() ) error( ", vs[" + vs.length + "]" );
       }//constructor

      /** Vertex 'v' in 'this' Renumbered is Vertex vs[v] of
          the {@link #parent() parent()} Graph. */
      public int v2pv(int v) { return vs[v]; } // makes it a SubGraph

      /** The {@link #parent() parent}'s degree(vs[v]). */
      public int degree(int v) { return Graph.this.degree(vs[v]); }

      /** The {@link #parent() parent}'s isEdge(vs[v0],isEdge[v1]). */
      public boolean isEdge(int v0, int v1)
       { return Graph.this.isEdge(vs[v0], vs[v1]); }

      /** The {@link #parent() parent}'s vLabel(vs[v]), if any. */
      public Value vLabel(int v) { return Graph.this.vLabel(vs[v]); }

      /** The {@link #parent() parent}'s eLabel(vs[v0],isEdge[v1]). */
      public Value eLabel(int v0, int v1)
       { return Graph.this.eLabel(vs[v0], vs[v1]); }

      public String toString()
       { return parent() + ".renumbered(" + Vector.ints(vs) + ")"; }
    }//Graph.Renumbered class
   // .........................................................................

   /** Canonical is mostly just class {@link Renumbered}
       <EM>except</EM> that it
       (i)&nbsp;declares 'this' to already be in a
           {@link Graph#canonical canonical} vertex-numbering, and thus
       (ii)&nbsp;{@link Canonical#canonical() Canonical.canonical()}
           returns 'this'.
       Also see {@link #isomorphic isomorphic(g)}.  */
   public class Canonical extends Renumbered // Needs Java 7+, ?Apple?
    { /** Note, the constructor does not (could not easily)
         check that 'vs' really is a canonical renumbering. */
      public Canonical(int [] vs) { super(vs); }

      /** Returns 'this' &ndash; it's canonical, right?
          Also see @link Graph#canonical() canonical()}. */
      public Canonical canonical() { return this; }

      public String toString()
       { return parent() + ".canonical(" + Vector.ints(vs) + ")"; }
    }//Graph.Canonical class
   // -------------------------------------------------------------------------

   /** Calls {@link Test#main}. */
   public static void main(String[] argv)                               // main
   { System.out.println("-- test Graph --");
     Test.main(argv);
     System.out.println("-- done Graph --");
   }//main

 }//Graph class,  https://www.cantab.net/users/mmlist/MML/A/

// (c) Lloyd Allison July 2014, Mon 11 Dec 2017. Licence to 'Graph' is Gnu GPL.
