//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789
                                        // Matrix, 2D Rectangular  Matrices ...

package la.maths;

import la.la.*;  // for Value etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** (Abstract) Matrix, for two-dimensional, rectangular,
    Vectors of Vectors; the new, principal operations are
    <NOBR>&bull; {@link #nCols() nCols()},</NOBR> and
    <NOBR>&bull; {@link #elt(int,int) elt(r,c)},</NOBR> plus as for
    all <TT>{@link Vector}s</TT>.  Also see {@link GP2}. And note
    that a non-rectangular, "jagged" Vector of Vectors can be created
    as, say, a {@link Vector.GP} of <TT>GPs</TT>. */
public abstract class Matrix extends Vector                         // * Matrix
 {
   private int nCols = -1;  // flagged as unset, for now.

   /** Constructor for use when the number of columns, nC, is known
       <I>before</I> calling the constructor, and the Matrix is known to
       be proper (rectangular). In this case {@link #setNcols} must not be
       called and {@link #nEltsRaw} need not be implemented --
       also see Matrix(). */
   public Matrix(int nC)
    { this.nCols = nC;
      if( nElts() < 1 || nC < 1 ) error( " cannot have an empty Matrix" );
    }// constructor

   /** The trivial constructor for use when the number of columns is
       <I>not</I> known before calling a Matrix constructor.
       Also see {@link #setNcols}, and constructor Matrix(nC). */
   public Matrix() { }

   /** If the trivial constructor, <TT>Matrix()</TT>, has been used, a
       sub-class of Matrix must override {@link #nEltsRaw} and call
       setNcols(nC) before {@link #nCols() nCols()}
       (or {@link #nElts(int) nElts(r)}) is used. */
   protected void setNcols(int nC)                              // setNcols(nC)
    { if( nCols >= 0 ) error( ", nCols=" + nCols + " >= 0 already" );
      assert nC >= 0 : "nC=" + nC + " < 0";
      nCols = nC;         // cache
      for( int r = 0; r < nElts(); r ++ ) // check
         if( nEltsRaw(r) != nCols )
            error( " not rectangular, row=" + r + " of " + nElts() );
      if( nElts() < 1 || nC < 1 ) error( " cannot have an empty Matrix" );
    }//setNcols(nC)

   /** Called in the rectangularity-check within {@link #setNcols(int)}
       <I>if</I> the constructor Matrix() was used; it need not be implemented
       (overridden) if Matrix(nC) was used. See {@link #nElts(int)}. */
   protected int nEltsRaw(int r)                               // ? nEltsRaw(r)
    { throw UOE( ".nEltsRaw(" + r + ") has not been implemented" ); }
   // -------------------------------

   /** The element Type of 'this' Matrix must be
       {@link la.la.Type#VECTOR VECTOR}.
       Also see {@link Vector#eltType()}. */
   public Type eltType() { return Type.VECTOR; }                   // eltType()

   /** The number of rows (<I>top-level</I> elements) in 'this' Matrix.
       Also see {@link #nCols nCols()}. */
   public abstract int nElts();                                    // * nElts()

   /** The number of colums; note every Matrix is rectangular. */
   public int nCols()                                                // nCols()
    { assert nCols >= 0 : errMsg( ", nCols=" + nCols );
      return nCols;
    }//nCols()

   /** Row 'r' (top-level element r) of 'this' Matrix; a sub-class of
       Matrix might be able to provide a more efficient version than
       this default. Also see {@link #elt(int,int) elt(r,c)}.
       <NOBR>Note that elt(r).wt(c) is 1.0, &forall;&nbsp;c.</NOBR>  */
   public Vector elt(final int r)                             // row r,  elt(r)
    { return new Vector()
       { public int      nElts() { return Matrix.this.nCols();   }
         public Value elt(int c) { return Matrix.this.elt(r, c); }
    }; }//elt(r)

   /** Return {@link #nCols()}; note that every Matrix is rectangular. */
   public int nElts(int r) { return nCols(); }                      // nElts(r)

   /** The Matrix element at position c of {@link #elt(int) elt(r)},
       that is at column c of row r of 'this' Matrix. */
   public abstract Value elt(int r, int c);                       // * elt(r,c)

   /** Return the sub-Matrix with rows, rs, and columns, cs.
       Also see {@link Vector#slice}.  */
   public Matrix elts(final int[] rs, final int[] cs)            // elts(rs,cs)
    { final Matrix m = this;  final int nE = nElts();
      for( int i = 0; i < rs.length; i ++ ) // check...
       { final int rs_i = rs[i];
         if( rs_i < 0 || rs_i >= nE )
            error( ".elts(rs,cs), rs[" + i + "]=" + rs_i );
       }
      for( int i = 0; i < cs.length; i ++ ) // check...
       { final int cs_i = cs[i];
         if( cs_i < 0 || cs_i >= nCols )
            error( ".elts(rs,cs), cs[" + i + "]=" + cs_i );
       }
      return new Matrix(cs.length)
       { public int      nElts() { return rs.length;   }
         public Value  elt(int r, int c) { return m.elt(rs[r], cs[c]); }
         public double wt(int r) { return m.wt(rs[r]); }
         public double constWt() { return m.constWt(); }
         // NB. AoMs of elts (of rows) may vary, even if m's do not!
    }; }//elts(rs,cs)

   /** True, every Matrix is rectangular by definition. */
   public boolean isRectangular() { return true; }

   /** Every row of 'this' Matrix has {@link #nCols() nCols()} elements. */
   public int rowLength() { return nCols(); }

   /** ?<TT>{@link #nElts() nElts()}=={@link #nCols() nCols()}</TT>? */
   public boolean isSquare() { return nElts() == nCols(); }       // isSquare()

   /** Return 'this'; also see {@link Vector#asMatrix()}. */
   public Matrix asMatrix() { return this; }
   // .........................................................................

   /** Return the N&times;N identity Matrix of Ints;
       note, is <TT>static</TT>. */
   public static Matrix id(final int N)                                // id(N)
    { return new Matrix(N)
       { public int           nElts() { return N; }
         public Int elt(int r, int c) { return r==c ? one : zero; }
         public boolean    isSquare() { return true; }
         public boolean nearlySymmetric(double e) { return true; }
         public Matrix      inverse() { return this; }
         public Matrix    transpose() { return this; }
    }; }//id(N)

   /** 'this' Matrix, M, times the Vector of Cts, v, as a column-Vector,
        as in <NOBR>M v.</NOBR> Also see {@link #times(Matrix)}. */
   public Vector times(Vector v)                                    // times(v)
    { return times(v.as_nx1()).col(0); }

   /** Matrix-multiplication of two matrices of Cts, 'this' and M. */
   public Matrix times(Matrix M)                                    // times(M)
    { /* final Vector m1 = this; */
      final int r1 = /* m1. */ nElts(),  r2 = M.nElts(),
                c1 = /* m1. */ nCols(),  c2 = M.nCols();
      if( c1 != r2 ) error( ", shapes " + r1+"x"+c1 + " * " + r2+"x"+c2 );
      final double[][] ans = new double[r1][c2];
      for( int r = 0; r < r1; r ++ )
         for( int c = 0; c < c2; c ++ )
          { double ans_rc = 0.0;
            for( int k = 0; k < c1 /* c1==r2 */; k ++ )
               ans_rc += /* m1. */ elt(r,k).x() * M.elt(k,c).x();
            ans[r][c] = ans_rc;
          }
      return Matrix.doubles(ans); // "exact"
    }//times(M)

   /** Return the transpose of 'this' Matrix.  Note,
       transpose().{@link #wt(int) wt(r)} is 1.0 for all&nbsp;r. */
   public Matrix transpose()                                     // transpose()
    { if( nElts() == 0 ) return this; // ?, else ...
      return new Matrix(/* nCols = */ this.nElts())
       { public int             nElts() { return Matrix.this.nCols();   }
         public Value elt(int r, int c) { return Matrix.this.elt(c, r); }
         public Vector       elt(int r) { return Matrix.this.col(r);    }
         public Vector       col(int c) { return Matrix.this.elt(c);    }
         public Matrix      transpose() { return Matrix.this; } // !
    }; }//transpose()
   // .........................................................................

   /** Return the inverse of a (square, non-singular) Matrix
       by {@link #GJ_fp()}. */
   public Matrix inverse()                                         // inverse()
    { if( ! isSquare() ) error( ", Matrix is not square" );
      // return GJ_naive();  // take
      // return GJ_pp();     // your
      return GJ_fp();     // pick
    }//inverse(m)

   /** GJ_fp, the Gauss-Jordan (GJ) elimination algorithm with full row
       and column pivoting (fp); also see {@link #inverse()}.  */
   public Matrix GJ_fp()                                             // GJ_fp()
    { final int N = nElts();  // 'this' is NxN
      final double[][] m = new double[N][N], //  (i) copy of 'this', and
                     ans = new double[N][N]; // (ii) the answer, later
      // Note, we do unto 'ans' as we do unto 'm' and,
      // by the end, m = id, and ans = inv(m).
      for( int r = 0; r < N; r ++ )
       { for( int c = 0; c < N; c ++ ) // copy 'this' into m
          { m[r][c] = elt(r,c).x(); ans[r][c] = 0.0; }
         ans[r][r] = 1.0;  // ans = Id, at the moment
       }
      final int[] cp = new int[N];   // cp, a Column Permutation
      for( int i = 0; i < N; i ++ ) cp[i] = i;    // the id perm
      // operate on m[row][cp[col]] (& ans) rather than on m[row][col]...

      for( int row = 0; row < N; row ++ )
       { double big = 0;
         int big_r = row, big_c = row;          // pivoting...
         for( int row2 = row; row2 < N; row2 ++ )
            for( int col = row; col < N; col ++ )
             { double mrc = Math.abs(m[row2][cp[col]]);
               if( mrc > big )  // looking for biggest
                { big = mrc; big_r = row2; big_c = col; }
             }
         double[] temp = m[row]; m[row] = m[big_r]; m[big_r] = temp; //swap
         temp = ans[row]; ans[row] = ans[big_r]; ans[big_r] = temp;  // -"-
         int cp_t = cp[big_c]; cp[big_c] = cp[row]; cp[row] = cp_t;  // &c
         // http://www.allisons.org/ll/
         double pvt = m[row][cp[row]]; // i.e., the big elt we just found
         if( pvt == 0 ) error( ".GJ_fp(,), row=" + row + " pvt=0" );
         for( int col = N - 1; col >= row; col -- )
            m[row][cp[col]] /= pvt;
         // and in particular, m[row][cp[row]] = 1.0 now
         for( int col = N - 1; col >= 0; col -- )
            ans[row][cp[col]] /= pvt;
         for( int row2 = row+1; row2 < N; row2 ++ )
          { double elt = m[row2][cp[row]];
            //In rows [row+1,N) of m[][cp[]], zero the 'col' entries
            for( int col = N - 1; col >= row; col -- )
               m[row2][cp[col]] -= m[row][cp[col]] * elt;
            // and in particular, m[row2][cp[row]] = 0
            for( int col = N - 1; col >= 0; col -- )
               ans[row2][cp[col]] -= ans[row][cp[col]] * elt;
          }//for row2
       }//for row
      // m[][cp[]] is now upper triangular.

      // Next, back-substitution ...
      for( int row = N - 1; row >= 1; row -- )
       { for( int row2 = row - 1; row2 >= 0; row2 -- )
          { double elt = m[row2][cp[row]];
            for( int col = row2+1; col <= row;  col ++ )
               m[row2][cp[col]] -= m[row][cp[col]] * elt;
            for( int col = 0; col < N; col ++ )
               ans[row2][cp[col]] -= ans[row][cp[col]] * elt;
          }//for row2
       }//for row

      // Finally, undo the column permutation, pc, on the ans rows!
      final double[][] ans2 = new double[N][];
      for( int row = 0; row < N; row ++ )
         ans2[cp[row]] = ans[row];
      return doubles( ans2 ); // i.e. a Matrix
    }//GJ_fp(m,ans)


   /** The Jacobi algorithm finds the Eigen-Values and Eigen-Vectors of
       'this' <I>square, symmetric</I> Matrix of Cts; it is usually
       invoked via {@link #eigen}, {@link #eigenValues}, or
       {@link #eigenVectors}. Also see {@link #logDeterminant()}.  */
   public class Jacobi                                                // Jacobi
    { private final int N = nElts();
      private int state;
      private int[] big = new int[N];
      private boolean[] changed = new boolean[N];
      private double sin_theta, cos_theta;
      private double[][] S = new double[N][N];  // copy of 'this'
      /** The Eigen-values, as double[]. */
      protected final double[] eVals = new double[N];
      /** The Eigen-vectors, as double[][], one E-vector per row. */
      protected final double[][] eVecs = new double[N][N];

      /** Set big[row] to the col of largest elt in S[row][row+1..N-1]. */
      private void find_big(int row)
       { int col = row+1;  final double[] Sr = S[row];
         double Src = -1.0;
         for( int c = row+1; c < N; c ++ )
          { final double Sri = Math.abs(Sr[c]);
            if( Sri > Src) { col = c; Src = Sri; }
          }
         big[row] = col;
       }//find_big(row)

      /** Adjust the estimate of the k-th Eigen-value by +t. */
      private void update (int k, double t)
       { final boolean alters = Math.abs(t) > 0.0;  // 0.001; // ?
         eVals[k] += t;
         if( changed[k] )
          { if( ! alters ) { changed[k] = false; state -- ; } } // no change
         else // ! changed[k]
            if(   alters ) { changed[k] = true;  state ++ ; }
       }//update(k,t)

      /** Rotate S[r2][c2] S[r1][c1] anticlockwise by &theta;. */
      private void rotate (int r1, int c1, int r2, int c2 )
       { final double S1 = S[r1][c1], S2 = S[r2][c2]; // old
         S[r1][c1] = cos_theta * S1 - sin_theta * S2; // new
         S[r2][c2] = sin_theta * S1 + cos_theta * S2; // -"-
       }//rotate(r1,c1,r2,c2)

      /** The "constructor" carries out the computation of the
          Eigen-Values and Eigen-Vectors. */
      public Jacobi()
       { if( ! nearlySymmetric() ) error( ", not (nearly) symmetric" );

         for( int r = 0; r < N; r ++ )
          { for( int c =0; c < N; c ++ )
             { S[r][c] = elt(r,c).x();  // S := this
               eVecs[r][c] = 0.0; // ...
             }
            eVecs[r][r] = 1.0;    // ...eVecs := id(N), and
            eVals[r] = S[r][r];   // ...eVals, as a 1st guess
            changed[r] = true;
            find_big(r);
          }
         state = N;
         int iteration = 0;  // The upper-right triangle of S[][] may be
         while( state > 0 )  // changed during the following loop
          { if( ( ++ iteration) > 100 )
             { System.out.println(
                 " {Jacobi might be looping (" + iteration + "): stop.}" );
               break;
             }
            int Pr = 0;  // find pivot, P, a big off-diagonal element
            for( int k = 1; k < N-1; k ++ )   // NB off-diagonal!
               if( Math.abs(S[k][big[k]]) > Math.abs(S[Pr][big[Pr]]) )
                  Pr = k;
            // found pivot, P, at row, Pr, and col, big[Pr].
            int Pc = big[Pr];  double P = S[Pr][Pc];  double P2 = P*P;
            if( P == 0 ) break;

            double y   = (eVals[Pc] - eVals[Pr]) / 2;
            double t   = Math.abs(y) + Math.sqrt(P2 + y*y);
            double hyp = Math.sqrt(P2 + t*t);
            cos_theta = t / hyp;
            sin_theta = P / hyp;
            t = P2 / t;
            if( y < 0 ) { sin_theta = -sin_theta; t = -t; }

            S[Pr][Pc] = 0;    // changed S's upper-right triangle
            update(Pr, -t);  update(Pc, t);   // Eigen-values and ...
            for( int i = 0;    i < Pr; i ++ ) rotate(i,  Pr, i,  Pc);
            for( int i = Pr+1; i < Pc; i ++ ) rotate(Pr, i,  i,  Pc);
            for( int i = Pc+1; i < N;  i ++ ) rotate(Pr, i,  Pc, i );
            for( int i = 0; i < N; i ++ )     // ... Eigen-vectors
             { double Eri = cos_theta*eVecs[Pr][i] - sin_theta*eVecs[Pc][i],
                      Eci = sin_theta*eVecs[Pr][i] + cos_theta*eVecs[Pc][i];
               eVecs[Pr][i] = Eri;
               eVecs[Pc][i] = Eci;
             }//for i

            find_big(Pr); find_big(Pc);
          }//while state > 0

         S = null; big = null; changed = null; // free the working space
       }//constructor
    }//Jacobi class


   /** Return the Eigen-Values and Eigen-Vectors,
       <NOBR>(eVals, eVecs),</NOBR> of 'this' <I>square, symmetric</I>
       Matrix using the {@link Jacobi} algorithm.
       Also see {@link #eigenValues} and {@link #eigenVectors}.  */
   public Value.Tuple eigen()                                        // eigen()
    { Jacobi J = new Jacobi();
      return Value.pair( doubles(J.eVals), doubles(J.eVecs) );
    }//eigen()

   /** Return the Eigen-Values of 'this' <I>square, symmetric</I> Matrix.
       Also see {@link #eigen()}. */
   public Vector eigenValues()                                 // eigenValues()
    { return doubles(new Jacobi().eVals); }

   /** Return the Eigen-Vectors of 'this' <I>square, symmetric</I> Matrix.
       Also see {@link #eigen()}. */
   public Matrix eigenVectors()                               // eigenVectors()
    { return doubles(new Jacobi().eVecs); }

   /** Return the log of the determinant of 'this'
       <I>square, symmetric</I> Matrix.  This is the sum of
       the logs of its {@link Jacobi#eVals Eigen-values}. */
   public double logDeterminant()                           // logDeterminant()
    { final double [] evs = new Jacobi().eVals;
      double logD = 0;
      for( int i = 0; i < evs.length; i ++ ) logD += Math.log(evs[i]);
      return logD;
    }//logDeterminant()

   /** Calculate the determinant of 'this' Matrix by the slow, recursive
       (naive) method. It is probably OK for 3x3 and smaller but it will
       be slow and may have numerical problems for a large Matrix.
       Also see {@link #logDeterminant}. */
   public double slowDeterminant()                         // slowDeterminant()
    { final int N = nElts();
      if( N == 1 ) return elt(0, 0).x();
      int [] rows = new int [N-1], cols = new int [N-1];
      for( int i = 0; i < N-1; i ++ )
       { rows[i] = i+1; cols[i] = i+1; } // i.e., [1,2,...,N-1]
      double det = 0;
      boolean even_i = true;
      for( int i = 0; ; i ++ )
       { double d = elt(i, 0).x() * elts(rows, cols).slowDeterminant();
         det += even_i ? d : - d;
         if( i >= N-1 ) break;
         even_i = ! even_i;
         rows[i] = i;  // [0,2,...,N-1]; [0,1,3,...,N-1]; ...; [0,1,...,N-2].
       }// for i
      return det;
    }//slowDeterminant()
   // .........................................................................

   /** Read a Matrix of Cts from {@link Lexical} analyser, 'lex', such as a
       <TT>Lexical.{@link Lexical#forFile(String) forFile(fname)}</TT>, say.
       Elements (columns) are separated by white space,
       rows by newline(s).  AoM is guessed from the data.  */
   public static Matrix fromLexical(Lexical lex)            // fromLexical(lex)
    { long pORm  = (1L << Lexical.plus) | (1L << Lexical.minus);
      long start = pORm | (1L << Lexical.numeralR) | (1L << Lexical.numeral);
      int lines = 0, cols = 0;
      double AoM = Double.MAX_VALUE;
      double x[] = new double[101];  // cludge ???
      java.util.ArrayList <double[]> rows =
         new java.util.ArrayList <double[]> ();
      int sy = lex.sy();
      while( ! lex.eoi() ) // [+/-]31.4e-1 etc. ...
       { do{ if( cols >= x.length ) lex.error( " row is too long" );
             if( (1L << sy & pORm) > 0 ) lex.insymbol();  // [+|-]
             double AoM1 = 1.0;
             switch( lex.sy() )
              { case Lexical.numeral:
                   x[cols] = lex.theInt(); break;
                case Lexical.numeralR:
                   x[cols] = lex.theDbl(); AoM1 = lex.theAcc(); break;
                default: lex.error( " bad number" );
              }
             if( sy == Lexical.minus ) x[cols] = - x[cols];
             AoM = Math.min(AoM, AoM1);  // a very simple rule!
             lex.insymbol();  sy = lex.sy();  cols ++ ;
           }while( lex.sy() != Lexical.eofSy && ! lex.lines1stSy() );
         double[] x2 = new double[cols];
         for( int i = 0; i < cols; i ++ ) x2[i] = x[i];
         rows.add(x2);
         lines ++ ; cols = 0;
       } // note rows.toArray() would return Object[], so ...
      return doubles(rows.toArray(new double[0][0]), AoM);
    }//fromLexical(lex)
   // .........................................................................

   /** A simple, general-purpose implementation of a {@link Matrix}.
       Like all Matrices, a GP2 must be rectangular.
       (Also see the "1D" {@link Vector.GP}.)  */
   public static class GP2 extends Matrix                                // GP2
    { /** The [][] of Values held in 'this' GP2. */
      protected final Value[][] elts;
      /** Given a 2D Java array, <TT>[] []</TT>, of Values,
          construct a GP2, a {@link Matrix Matrix}. */
      public GP2(Value[][] elts)
       { this.elts = elts;
         setNcols(elts.length == 0 ? 0 : elts[0].length);
       }//constructor
      protected int  nEltsRaw(int r) { return elts[r].length;  }
      public int             nElts() { return elts.length;     }
      public Vector       elt(int r) { return values(elts[r]); }
      /** Return the (r,c)-th of {@link #elts elts}, {@link #force forced}. */
      public Value elt(int r, int c)
       { Value erc = elts[r][c].force(); elts[r][c] = erc; return erc; }
    }//GP2 class

   /** Convenience function,
       <TT>values : Value[][] &rarr; {@link GP2 GP2}</TT>.
       Also see the 1D {@link #values(Value[])}.  */
   public static GP2 values(Value [][] vs) { return new GP2(vs); }
   // .........................................................................

   /** Matrices of Ints (ints). */
   public static abstract class Ints extends Matrix                   // * Ints
    { public Ints() { super(); }//needs nEltsRaw(row)
      public Ints(int nCols) { super(nCols); }
      public Type    type() { return Type.MATRIX_INT; }
      public Type eltType() { return Type.VECTOR_INT; }
    }//Ints abstract class

   /** Convenience function,
       <TT>ints : int[][] &rarr; {@link Matrix}</TT>.
       Also see the 1D {@link #ints(int[])}.  */
   public static Ints ints(final int [][] ns)
    { return new Ints()
       { { setNcols(ns.length == 0 ? 0 : ns[0].length); }
         protected int  nEltsRaw(int r) { return ns[r].length; }
         public int             nElts() { return ns.length;    }
         public Vector       elt(int r) { return ints(ns[r]);  }
         public Value elt(int r, int c) { return new Value.Int(ns[r][c]); }
    }; }//ints(ns)

   /** Make a symmetric Int Matrix from 'ns' which is an
       upper-right <I>triangular</I> array of array of int.
       If 'incDiag' is false, 'ns' omits the main diagonal and all
       diagonal Values of the Matrix are taken to equal 'diagV'.
       If incDiag is true we don't care about diagV. */
   public static Ints ints_UR(final boolean incDiag, final int [][] ns,
                              final int diagV)
    { final int N = ns.length + (incDiag ? 0 : 1);
      return new Ints(/* nCols= */ N)
       {  { for( int r = 0; r < ns.length; r ++ )
               if( ns[r].length != ns.length-r )
                  error( ", ints_UR(" + incDiag + ",ns[][]," + diagV
                       + "), N=" + N + ", |ns[" + r + "]|=" + ns[r].length );
          }//constructor
         public int nElts() { return N; }
         public Value elt(int r, int c)
          { assert r >= 0 && r < N && c >= 0 && c < N : "r=" + r + ", c=" + c;
            return new Value.Int(
               incDiag ? (r<=c ? ns[r][c-r]   :          ns[c][r-c])
               /* ! */ : (r< c ? ns[r][c-r-1] : (r > c ? ns[c][r-c-1] : diagV))
               );
          }//elt(r,c)
         public boolean isSquare() { return true; }
         public boolean nearlySymmetric(double e) { return true; }
         public String toString()
          { Vector [] rows = new Vector [ns.length];
            for( int i = 0; i < rows.length; i ++ )
               rows[i] = ints(ns[i]);
            return "ints(" + incDiag + "," + values(rows) + "," + diagV + ")";
          }//toString()
    }; }//ints(,,)
   // .........................................................................

   // NB. There is no 2D version, double(nlAoM, double[][]), of
   //     the 1D doubles(nlAoM,[]).

   /** Matrices of Reals (Cts, doubles). */
   public static abstract class Doubles extends Matrix             // * Doubles
    { public Doubles() { super(); }//needs nEltsRaw(row)
      public Doubles(int nCols) { super(nCols); }
      public Type    type() { return Type.MATRIX_CTS; }
      public Type eltType() { return Type.VECTOR_CTS; }
    }//Doubles abstract class

   /** Convenience function,
       <TT>doubles : double[][] &times; AoM &rarr; Matrix</TT>, where
       every element of the Matrix has the same {@link #AoM() AoM()}.
       Also see the 1D {@link #doubles(double[],double)}.  */
   public static Doubles doubles(final double[][] xs, final double AoM)
    { return new Doubles()
       {  { setNcols(xs.length == 0 ? 0 : xs[0].length); }
         protected int  nEltsRaw(int r) { return xs[r].length; }
         public int             nElts() { return xs.length; }
         public Vector       elt(int r) { return doubles(xs[r], AoM); }
         public Value elt(int r, int c) { return Value.cts(xs[r][c], AoM); }
         /** The negative log AoM of every element, i.e., every <I>row</I>. */
         public double constNlAoM() { return - nCols() * Math.log(AoM); }
         public double   constAoM() { return Math.exp(-constNlAoM()); } // !
    }; }//doubles(xs,AoM)

   /** Convenience function
       <TT> : double[][] &rarr; Matrix</TT>, where each
       <TT>elt(r,c)</TT> is an <I>exact</I> {@link la.la.Value.Real Real}.
       Also see the 1D {@link #doubles(double[])}.  */
   public static Doubles doubles(final double[][] xs)
    { return new Doubles()
       {  { setNcols(xs.length == 0 ? 0 : xs[0].length); }
         protected int  nEltsRaw(int r) { return xs[r].length;   }
         public int             nElts() { return xs.length;      }
         public Vector       elt(int r) { return doubles(xs[r]); }
         public Value elt(int r, int c) { return Value.real(xs[r][c]); }
         public double            AoM() { return 0; }
         public double       constAoM() { return 0; } // every elt (row)
    }; }//doubles(xs)
   // -------------------------------------------------------------------------

   private static String show(Value m) { return m + ":" + m.type(); }

   /** main() allows Vector to be (slightly) tested, in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Matrix --");
      System.out.println("   see Test.main(.)");
      System.out.println("-- done Matrix --");
    }//main

 }//Matrix class

// (c) Lloyd Allison 2012, Sat 7 May 2017. Licence to 'Matrix' is Gnu GPL.
