//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.maths;

import la.la.*; // for Value etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Q, the (abstract) class of quaternions.
    Quaternions generalize Complex numbers and can be used to
    {@link Vector#rotate(double,Vector) rotate} 3D Vectors. */
public abstract class Q extends Vector.Doubles              // * Quaternions, Q
 {
   /** Error: wt(i) is not a Quaternion operation. */
   public double wt(int i) { throw RTE( ".wt(" + i + ") not for class Q" ); }

   /** Error: wts() is not a Quaternion operation. */
   public double wts()     { throw RTE( ".wts() not for class Q" ); }

   /** The "real" part, elt(0).x(). */
   public double  x() { return elt(0).x(); }                             // x()

   /** The i-part, elt(1).x(). */
   public double  i() { return elt(1).x(); }                             // i()

   /** The j-part, elt(2).x(). */
   public double  j() { return elt(2).x(); }                             // j()

   /** The k-part, elt(3).x(). */
   public double  k() { return elt(3).x(); }                             // k()

   /** Return 4. */
   public int nElts() { return 4; } // !

   /** Multiply Quaternions 'this' and 'q'.
       It generalizes complex multiplication. */
   public Q times(Q q)                                              // times(q)
    { final double q10 =   x(), q11 =   i(), q12 =   j(), q13 =   k(),
                   q20 = q.x(), q21 = q.i(), q22 = q.j(), q23 = q.k();
      return doubles( q10*q20 - q11*q21 - q12*q22 - q13*q23,   // 1
                      q10*q21 + q11*q20 + q12*q23 - q13*q22,   // i
                      q10*q22 - q11*q23 + q12*q20 + q13*q21,   // j
                      q10*q23 + q11*q22 - q12*q21 + q13*q20 ); // k
    }//times(q)

   /** Return the conjugate, (a,-b,-c,-d), of
       'this' Quaternion (a,b,c,d). */
   public Q conjugate()                                          // conjugate()
    { final double[] ds = new double[4];
      ds[0] = x(); ds[1] = -i(); ds[2] = -j(); ds[3] = -k();
      return doubles(ds);
    }//conjugate()

   /** Return the inverse of 'this' Quaternion, that is, its
       {@link #conjugate() conjugate} divided by its
       {@link #norm() norm}<SUP>2</SUP>. */
   public Q inverse()                                              // inverse()
    { final Q c = conjugate();
      final double n2 = sumSq(); // the norm squared
      final double[] ds = new double[4];
      for( int i = 0; i < 4; i ++ ) ds[i] = c.elt(i).x() / n2;
      return doubles(ds);
    }//inverse()

   /** For 'this' Quaternion, (a,b,c,d), return the 3D Vector (b,c,d). */
   public Vector asV3()                                               // asV3()
    { return new Vector()
       { public int      nElts() { return 3; }// !
         public Value elt(int i) { return Q.this.elt(i+1); }
    }; }//asV3()

   /** Return 'this' -- also see {@link Vector#asQ()}. */
   public Q asQ() { return this; }

   /** Show 'this' Quaternion as "(a&plusmn;bi&plusmn;cj&plusmn;dk)". */
   public String toString()                                       // toString()
    { StringBuilder ans = new StringBuilder( "("+x() );
      if( norm() == 0 ) return "(0Q)"; // else...
      for( int i = 1; i < nElts(); i ++ )
       { final double qi = elt(i).x(); final double absqi = Math.abs(qi);
         ans.append( absqi > 0
          ? ((qi > 0 ? "+" : "-" ) + absqi
             + (i==1 ? "i" : i==2 ? "j" : i==3 ? "k" : "?"))
          : "" );
       }//for i
      ans.append( ")" );
      return ans.toString();
    }//toString()
   // .........................................................................

   /** Given an array of doubles <NOBR>{a,b,c,d},</NOBR> return the
       Quaternion <NOBR>(a + bi +cj + dk).</NOBR> */
   public static Q doubles(final double[] abcd)
    { return new Q()
       { public final double[] elts;
         { this.elts = abcd;
           if( abcd.length != 4 ) error( " is not a Quaternion" );
         }//init
         public Real elt(int i) { return real(elts[i]); }//NB. Real
    }; }//q(abcd)

   /** Return the Quaternion <NOBR>(a + bi +cj + dk).</NOBR> */
   public static Q doubles(double a, double b, double c, double d)
    { return doubles(new double[]{a, b, c, d}); }

   /** Return the Quaternion <NOBR>(a + 0i +0j + 0k).</NOBR> */
   public static Q double1(double a)
    { return doubles(new double[]{a, 0, 0, 0}); }
   // -------------------------------------------------------------------------

   /** main() allows Q to be (slightly) tested, in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Q --");
      System.out.println("   see Test.main(.)");
      System.out.println("-- done Q --");
    }//main

 }//Q class   http://www.allisons.org/ll/

// (c) Lloyd Allison June 2013, Wed 13 Aug 2014. Licence to 'Q' is Gnu GPL.
