//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.util;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A simple timer that can {@link #start() start()},
    {@link #stop() stop()} (i.e., pause), and (re-)start(). */
public class Timer
 { private String  id;
   private int     count;         // number of start()
   private long    total, lastStart, lastStop;
   private boolean running;       // status

   /** Anonymous, not running (timing). */
   public Timer() { this("anonymous"); }

   /** Named, not running (timing). */
   public Timer(String id) { this(id, false); }

   /** Anonymous, running as specified. */
   public Timer(boolean running) { this("anonymous", running); }

   /** Named, running as specified. */
   public Timer(String id, boolean running) 
    { count = 0; total = 0;
      this.id = id;
      this.running = false; // but...
      if( running ) start();
    }//constructor

   /** Start the clock. If the Timer was already running,
       do a {@link #stop() stop()} first. */
   public void start()
    { if( running ) stop();
      running = true;
      count ++ ;
      lastStart = System.currentTimeMillis();
    }

   /** Stop (pause) 'this' <I>running</I> Timer and add the
       time since its last {@link #start() start()} to its total. */
   public void stop()
    { if( ! running ) error(".stop(), already stopped" );
      lastStop = System.currentTimeMillis();
      total += lastStop - lastStart;
      running = false;
    }

   /** {@link #stop() stop()},
       {@link #lastTime() lastTime()} (is returned),
       {@link #start() start()}.  */
   public long markTime()
    { stop(); long t = lastTime(); start(); return t; }

   /** Time since the last {@link #start() start()} in milliseconds
       to the subsequent stop(), if any, otherwise to 'now'.
       Note, it does <I>not</I> do a {@link #stop() stop()}.
       Also see {@link #markTime() markTime()}.  */
   public long lastTime()
    { if( count < 1) error( ".lastTime(), count=" + count );
      return (running ? System.currentTimeMillis() : lastStop) - lastStart;
    }

   /** Total time in milliseconds between all {@link #start() start()} -
       {@link #stop() stop()} pairs, plus current elapsed if started and
       not yet stopped.  Note, it does <I>not</I> do a 'stop().'  */
   public long totalTime()
    { if( count < 1) error( ".totalTime(), count=" + count );
      return total + (running ? lastTime() : 0);
    }

   /** Mean time in milliseconds of all periods of 'running'.
       Note, it does <I>not</I> do a 'stop().'  */
   public long meanTime()
    { if( count < 1 ) error( ".meanTime(), count=" + count );
      return Math.round( (double) totalTime() / count );
    }

   /** Throw a RuntimeException.
       The exception includes this.toString() and 'msg'. */
   public void error(String msg)
    { Util.error(this + msg ); }

   public String toString()
    { return "{Timer " + id + (running ? " (running)" : " (stopped)" )
           + ", count=" + count + ", total=" + totalTime() + " (msec) }" ;
    }

 }//Timer class

// (c) Lloyd Allison, Dec 2014, Wed 23 Mar 2015; Licence to 'Timer' is Gnu GPL.
