//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Type;
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised Adapative Model of Bounded Discrete data;
    also see&nbsp;{@link M M}.  The model (idea) is that a data-set is
    homogeneous, and the data are independent, but {@link Adaptive.M M}
    <I>does not care</I> what the probabilities of the various Values
    actually are. It offers no opinion about the probabilities so
    {@link Adaptive.M M} has no statistical parameters.
    Note, it is debatable whether 'Adaptive' should be a Model of
    Bounded Discretes, as here, or rather a Model of <I>Vector</I> of
    Bounded Discretes. These are not necessarily exclusive, and we do
    want the present form as a comparison to {@link Discretes.Bounded}.
    Note, for a data-set 'ds',
    <NOBR> {@link M#nlLH nlLH}({@link #stats stats}(ds)) </NOBR> and
    <NOBR> &sum;<SUB>i</SUB> {@link M#nlPr nlPr}(ds<SUB>i</SUB>) </NOBR>
    may give very different results.
    <I>Adapative should not be used in structured Models, such as
    {@link Mixture}, that require nlPr(d).</I>  */
public class Adaptive extends Discretes.Bounded                     // Adaptive
 { /** The data-space bounds, as ints. */
   private final int lwb_n, upb_n;

   /** The initial counts, the "offsets" added to the frequencies.
       The prior on the probabilities of the data values is
       {@link Dirichlet Dirichlet}.M({@link #alpha &alpha;}). Note that
       <NOBR> &alpha;<SUB>i</SUB>=1, &forall;
       i = {@link #lwb lwb} .. {@link #upb upb} </NOBR>
       amounts to a uniform prior. */
   private final double [] alpha;

   /** &sum; alpha;<SUB>i</SUB>. */
   private final double sum_alpha;

   /** Definition parameters <NOBR> dp = (bounds, &alpha;) </NOBR> where
       <NOBR> bounds = (lwb, upb) </NOBR> on 'this' Model's data-space,
       and <NOBR>&alpha;<SUB>k</SUB> &gt; 0</NOBR> is the "offset" to be
       added to the frequency of value&nbsp;v<SUB>k</SUB> by
       {@link M#nlLH nlLH(ss)}.  Setting all
       <NOBR>&alpha;<SUB>k</SUB> = 1</NOBR> gives the uniform "prior"
         distribution over probabilities (frequencies), while
       <NOBR>&alpha;<SUB>k</SUB> &gt; 1</NOBR> raises
         probabilities in the interior of the Simplex, and
       <NOBR>&alpha;<SUB>k</SUB> &lt; 1</NOBR> has the opposite effect.
       Note, &alpha; cannot be sensibly estimated from a data-set
       because the next data-set may be very different. */
   public Adaptive(Value dp)
    { super(dp);
      lwb_n = lwb_n(); upb_n = upb_n();     // see above
      final Tuple  dpT    = dp.isTuple(2);  // (bounds, alpha)
      Vector alphaV = (Vector) dpT.snd();
      if( alphaV.nElts() != upb_n - lwb_n + 1 )
         error( ", alpha.nElts()=" + alphaV.nElts() );
      this.alpha = new double[upb_n - lwb_n +1];
      double sum_alpha = 0;
      for( int i = 0; i <= upb_n - lwb_n; i ++ )
       { final double a_i = alphaV.elt(i).x();
         if( a_i <= 0 ) error( ", alpha[" + i + "]=" + a_i + "<=0" );
         this.alpha[i] = a_i;
         sum_alpha += a_i;
       }
      this.sum_alpha = sum_alpha;
    }//constructor

   public Tuple bounds()                                            // bounds()
    { return  (Tuple) ((Tuple) defnParams()) .fst(); } // data-space bounds

   /** The "offsets" to be the initial values of the frequency
       counters of data Values by {@link M#nlLH nlLH(ss)}. */
   public Vector alpha()                                             // alpha()
    { return (Vector) ((Tuple) defnParams()) .snd(); }

   /** Return sufficient statistics, that is {@link #freqs frequency}
       counts, for elements [lo,&nbsp;hi) of data-set 'ds'.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { return freqs(ds, lo, hi); }

   /** For sufficient statisticses 'ss0' and 'ss1', either combine
       ss0 and ss1 (add=true), or remove ss1 from ss0 (add=false).
       Note, it is only valid to remove from ss0 the contribution
       of elements previously counted in ss0.  Also see
       {@link #stats(Vector,int,int) stats(ds,lo,hi)}.  */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { return freqs(add, ss0, ss1); }

   /** Return a {@link M Model} with 2-part message lengths, msg1=0
       and msg2, and statistical parameter sp=(), triv. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /** Return the trivial Estimator,
       Mdl.{@link M#asEstimator asEstimator(t)},
       of {@link #Mdl Mdl}. */
   public Estimator estimator(Value t /* = triv */ )
    { return Mdl.asEstimator(t); }

   /** {@link Adaptive Adaptive} fully (trivially) parameterised. */
   public final M Mdl = new M(0, 0, triv);  // "The" Adaptive M-odel.
   // .........................................................................

   /** Model {@link #Mdl Mdl} should be enough
       for most purposes but here is its class, fully (trivially)
       parameterised&nbsp;(M). Also see the
       {@link Adaptive UnParameterised} Model. Note the remarks on
       {@link M#nlLH nlLH(ss)} and {@link M#nlPr_n nlPr_n(n)}. */
   public class M extends Discretes.Bounded.M                     // Adaptive.M
    { /** Requires msg1=0, sp=triv; msg2 may be &gt;&nbsp;0
          for fair competition on training data. */
      public M(double msg1, double msg2, Value sp)
       { super(0, msg2, triv); zeroTriv(msg1, sp); }//constructor

      /** The negative log probability of Discrete datum 'd', but pay careful
          attention to the remarks on {@link #nlLH nlLH(ss)}.
          ?Arguably nlPr(d) should be an Unsupported Operation? */
      public double nlPr(Value d)
       { return super.nlPr(d); }

      /** The negative log probability of datum int 'n', but pay careful
          attention to the remarks on {@link #nlLH nlLH(ss)}.
          ?Arguably nlPr_n(n) should be an Unsupported Operation? */
      public double nlPr_n(int n)
       { return nlLH(stats(Vector.ints( new int[]{n} ))); }  // !

      /** The negative log likelihood of data-set 'ds', where
          statistics <NOBR>ss = {@link Adaptive#stats stats(ds)}</NOBR>.
          The running frequency "counter" for Value v<SUB>k</SUB> is,
          in effect, initialised to&nbsp;&alpha;<SUB>k</SUB>.
          Note that nlLH(ss) may be quite different to
          <NOBR> &sum;<SUB>i</SUB> {@link #nlPr nlPr}(ds<SUB>i</SUB>) </NOBR>
          because the latter does not, and cannot, "adapt".  */
      public double nlLH(Value ss)                                  // nlLH(ss)
       { // ds:    v0      v1      v0      ...,           v_i in [0,k]
         //
         //  0: a0    a0+1    a0+1    a0+2                counter for v0
         //  1: a1    a1      a1+1    a0+1                etc.
         //  ...
         //  k: ak    ak      ak      ak
         //      s    s+1     s+2          ...,           where s = SUM a_i
         //
         //        a0      a1     a0+1     ...            --top
         // pr:    --  x  ---  x  ----  x
         //         s     s+1      s+2     ...  s+|d|-1   --bottom
         Vector freqs = (Vector) ss;  // frequencies of the possible values
         double sum_freqs = 0, ans = 0;
         for( int i = 0; i < freqs.nElts(); i ++ )
          { final double f_i = freqs.elt(i).x();
            if( f_i <= 0 ) continue;  // probably worth the test
            sum_freqs += f_i;
            ans -= alpha[i] == 1                            // NB. '-='
                    ? logFactorial( f_i )
                    : logGamma( alpha[i] + f_i ) -     // top
                      logGamma( alpha[i] );
          }
         ans += logGamma( sum_alpha + sum_freqs ) -    // bottom
                logGamma( sum_alpha );
         // Note, Gamma n = (n-1)!,  and n! = Gamma(n+1)
         return ans;
       }//nlLH(ss)

      /** Not implemented, arguably it is not possible (do not think that
          using {@link #random(int) random(1)} truly gets around this).
          For a data-set, {@link #random(int) random(n)} and
          {@link #randomSeries() randomSeries()} are fine.  */
      public int random_n() { throw UOE( ".random() not implemented" ); }

      /** Uses {@link #randomSeries() randomSeries()} to
          sample 'n' random <I>homogeneous</I> Values. */
      public Vector random(int n)
       { int [] rns = new int [n];
         la.util.Series.Discrete ss = randomSeries();
         for( int i = 0; i < n; i ++ ) { rns[i] = ss.elt_n(); ss.advance(); }
         return Vector.ints((Type.Discrete) lwb().type(), rns);
       }//random(n)

      /** Return a {@link la.util.Series.Discrete} which is
          random and <I>homogeneous</I>: Sample probabilities 'prs[]'
          from {@link Dirichlet.M}({@link #alpha() alpha)} and then
          repeatedly sample Ints from {@link MultiState.M}(prs),
          i.e., the Series is random and homogeneous.  Don't forget to
          {@link la.util.Series#advance() advance()}!  */
      public la.util.Series.Discrete randomSeries()
       { return new la.util.Series.Discrete((Type.Discrete) lwb().type())
          { private int pos = -1, elt_n;   // position, current int
            private final MultiState.M ms; // Model for this Series
             { Dirichlet.M  dm = new Dirichlet(new Value.Int(upb_n-lwb_n))
                                    . sp2Model(0, 0, alpha());
               ms = new MultiState(bounds())
                                    . sp2Model(0, 0, dm.random());
               advance(); /* ! */
             }// constructor
            public boolean hasSome() { return true; }
            public int    position() { return pos; }
            public int       elt_n() { return elt_n; }
            public void    advance() { pos ++ ; elt_n = ms.random_n(); }
       }; }//randomSeries()

    }//Adaptive.M class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** See {@link Test}. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- test Adaptive --");

      Tuple bnds = pair(zero, one);
      Adaptive A01 = new Adaptive(pair(bnds,
                                       Vector.doubles(new double[]{7, 3})));
      Vector r = A01.Mdl.random(100);
      System.out.println( "   " + A01 + ": " + r + ": " + A01.freqs(r) );

      System.out.println("   see mml.Test" );

      System.out.println("-- done Adaptive --");
    }//main

 }//Adaptive class, the UnParameterised Model

// (c) Lloyd Allison, 2014, Tue 12 April 2016. Licence to 'Adaptive' is Gnu GPL.
