//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A class that may be useful in defining (UnParameterised) Models over
    continuous data-spaces such as, but not limited to,
    {@link la.la.Value.Cts}, and as defined by a probability density function
    {@link ByPdf.M#pdf pdf(d)} and {@link ByPdf.M#nlPdf nlPdf(d)}.
    The fully parameterised Model is {@link ByPdf.M}.
    Also&nbsp;see interface {@link HasPdf}.  */
public abstract class ByPdf extends UPModel                          // * ByPdf
 { public ByPdf(Value dp)
    { super(dp); }//constructor
   // .........................

   /** The (abstract) class of fully parameterised {@link ByPdf} Models;
       specify {@link #nlPdf nlPdf(d)}.  The UnParameterised Model
       is {@link ByPdf}. Also see interface {@link HasPdf}. */
   public abstract class M extends UPModel.M implements HasPdf     // * ByPdf.M
    { public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Probability density function, pdf(d), of a datum d;
          returns <CODE>exp(-{@link #nlPdf nlPdf(d)})</CODE>. */
      public double pdf(Value d)                                      // pdf(d)
       { return Math.exp( - nlPdf(d) ); }

      /** The negative log pdf, nlPdf(d), of a datum d,
          must be specified to define a <TT>ByPdf</TT>.
          Also see {@link #pdf(Value) pdf(d)}.  */
      public abstract double nlPdf(Value d);                      // * nlPdf(d)

      /** A convenience function using a negative log AoM, 'nlAoM',
          other than that set in the datum, d, itself.  This might,
          for example, be useful where the datum, d, has gone through
          a transformation that has zapped its nlAoM().  */
      public double nlPr(double nlAoM, Value d)                // nlPr(nlAoM,d)
       { final double nits = nlPdf(d) + nlAoM;
         assert nits >= 0 : "nits=" + nits + " < 0 in " + this;
         return nits;
       }//nlPr(nlAoM,d)

      /** The negative log probability of datum d; returns
          <NOBR>{@link #nlPdf nlPdf(d)}
                + d.{@link la.la.Value#nlAoM() nlAoM()}</NOBR>. */
      public double nlPr(Value d)                                    // nlPr(d)
       { final double nits = nlPdf(d) + d.nlAoM();
         assert nits >= 0 : "nits=" + nits + " < 0 in " + this;
         return nits;
       }//nlPr(d)

    }//ByPdf.M (abstract) class, the fully parameterised Model

 }//ByPdf (abstract) class, the UnParameterised UPModel

// (c) Lloyd Allison, 2012, Tues 3 Dec 2013. Licence to 'ByPdf' is Gnu GPL.
