//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** An UnParameterised Dependent Model of data pairs,
    &lang;id,&nbsp;od&rang;, made from an UnParameterised Model,
    {@link #upm upm}, of the input (independent) datum, id, and from an
    UnParameterised FunctionModel, {@link #upfm upfm}, of the output
    (dependent) datum, od, conditional on id.
    Also see the fully parameterised {@link Dependent.M},
    and {@link Independent}.  */
public class Dependent extends Multivariate                        // Dependent
 {
   /** UnParameterised input Model of input datum id. */
   public final         UPModel  upm;

   /** UnParameterised FunctionModel of id&rarr;od. */
   public final UPFunctionModel upfm;

   /** Given problem-definition parameters,
       dp&nbsp;=&nbsp;&lang;{@link #upm},&nbsp;{@link #upfm}&rang;,
       construct a Dependent UPModel. */
   public Dependent(Value dp)
    { super(dp);   // note, width
      Value.Tuple i_f = dp.isTuple(2);
      upm  = (UPModel) i_f.fst();
      upfm = (UPFunctionModel) i_f.snd();
    }//constructor

    /** Two columns, width&nbsp;=&nbsp;2.  */
   public int width() { return 2; }

   /** Given 2-part message lengths, m1 &amp; m2, and statistical
       parameters sps&nbsp;=&nbsp;&lang;upm's,&nbsp;upfm's&rang;,
       return a fully parameterised {@link M M Model}.
       "upm's" is the statistical parameter given to upm etc..  */
   public M sp2Model(double m1, double m2, Value sps)
    { return new M(m1, m2, sps); }

   /** Given a data-set, ds, return sufficient statistics,
       ss&nbsp;=&nbsp;&lang;upm's,&nbsp;upfm's&rang;.
       ss&nbsp;is a pair. "upm's" is the stats from upm.stats(.) etc..
       Note that upm's depends on only the first column of ds,
       but upfm's, being for an UnParameterised FunctionModel,
       depends on both columns.
       May or may not equal {@link M#stats(Vector,int,int) M.stats(,,)}.
       More on stats {@link UPModel#stats here}. */
   public Value.Tuple stats(Vector ds, int lo, int hi)
    { if( ds.nElts() > 0 && ds.nCols() != 2 )  // hmmm?
         error( "ds is not two-columned" );
      return Value.pair(  upm.stats(ds.col(0),lo,hi),    // col 0
                         upfm.stats(ds,       lo,hi) );  // both cols
    }//stats(ds)

   /** Calls {@link #upm}'s and {@link #upfm}'s stats. May or may not
          equal {@link M#stats(boolean,Value,Value) M.stats(,,)}. */
   public Value.Tuple stats(boolean add, Value ss0, Value ss1)
    { Value.Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
      return Value.pair( upm.stats(add, ss0T.fst(), ss1T.fst()),
                        upfm.stats(add, ss0T.fst(), ss1T.fst()) );
    }//stats(add,ss0,ss1)

   /** The Estimator's parameter is a pair,
       ps&nbsp;=&nbsp;&lang;upm's,&nbsp;upfm's&rang;.  */
   public Est estimator(final Value ps)
    { final Value.Tuple psT = ps.isTuple(2);
      return new Est(ps)
       { public M ss2Model(Value ss)
          { Value.Tuple ssT = ss.isTuple(2);
            Model im =  upm.estimator(psT.fst()).ss2Model(ssT.fst());
            FunctionModel fm = (FunctionModel)
               upfm.estimator(psT.snd()).ss2Model(ssT.snd());
            return new M( im.msg1()+fm.msg1(), im.msg2()+fm.msg2(),
                          Value.pair(im.statParams(), fm.statParams()) );
          }//ss2Model(ss)
    }; }//estimator(ps)
   // .........................................................................

   /** A fully parameterised {@link Dependent Dependent} Model of
       data pairs, &lang;id,&nbsp;od&rang;, made from a Model, im,
       of the input (independent) datum, id, and from a {@link FunctionModel},
       fm, of the ouput (dependent) datum, od, conditional upon id.
       Note that id and/or od may be multivariate.
       Also see {@link Independent.M}.  */
   public class M extends Multivariate.M                         // Dependent.M
    {
      /** 'im', the fully parameterised input {@link Model Model}
           of the input datum, id. */
      public final Model im;

      /** 'fm', the fully parameterised {@link FunctionModel FunctionModel}
          of the output datum, od, conditional upon the input datum, id. */
      public final UPFunctionModel.M fm;

      /** The two-part message lengths are msg1 and msg2, and
          sps&nbsp;=&nbsp;(iSp,&nbsp;fSp), where iSp is the stat param
          for the input Model {@link #im im}, and fSp is for the
          FunctionModel {@link #fm fm}.  */
      public M(double msg1, double msg2, Value sps)
       { super(msg1, msg2, sps);
         Value.Tuple spT = sps.isTuple(2);
         im = upm.apply(spT.fst());
         fm = (UPFunctionModel.M) upfm.apply(spT.snd());
       }//constructor

      /** The negative log probability of datum
          iod&nbsp;=&nbsp;(id,&nbsp;od).  */
      public double nlPr(Value iod)
       { Value.Tuple id_od = iod.isTuple(2);
         Value id = id_od.fst();
         return im.nlPr(id) + fm.condNlPr(id, id_od.snd());
       }//nlPr(iod)

      /** Given a data-set, ds, return statistics,
          ss&nbsp;=&nbsp;&lang;{@link #im im}'s,&nbsp;{@link #fm fm}'s&rang;.
          ss&nbsp;is a pair. "im's" is the stats from im.stats(.) etc..
          Note that im's depends on only the first column of ds, but fm's,
          being for an UnParameterised FunctionModel, depends on both columns.
          May or may not equal {@link Dependent#stats(Vector,int,int)}.
          More on stats {@link Model#stats here}. */
      public Value.Tuple stats(Vector ds, int lo, int hi)
       { if( ds.nElts() > 0 && ds.nCols() != 2 )  // hmmm?
            error( "ds is not two-columned" );
         return Value.pair( im.stats(ds.col(0),lo,hi),    // col 0
                            fm.stats(ds,       lo,hi) );  // both cols
       }//stats(ds)

      /** Calls {@link #im}'s and {@link #fm}'s stats. May or may not
          equal {@link Dependent#stats(boolean,Value,Value)}. */
      public Value.Tuple stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
         return Value.pair( im.stats(add, ss0T.fst(), ss1T.fst()),
                            fm.stats(add, ss0T.fst(), ss1T.fst()) );
       }//stats(add,ss0,ss1)

      /** Given sufficient statistics,
          ss&nbsp;=&nbsp;{@link #stats(Vector,int,int) stats(ds)} of
          a data-set, ds, return the negative log likelihood of ds. */
      public double nlLH(Value ss)
       { Value.Tuple ssT = ss.isTuple(2);
         return im.nlLH(ssT.fst()) + fm.nlLH(ssT.snd());
       }

      /** Return a random (id,&nbsp;od), assuming both {@link #im}
          and <TT>{@link #fm}.condModel(.)</TT> implement
          {@link Model#random random()}. */
      public Value.Tuple random()
       { Value id = im.random();
         return Value.pair(id, fm.random(id));
       }//random()

   }//Dependent.M class, the Model

 }//Dependent class, the UPModel

// (c) Lloyd Allison 2012, Mon 4 Dec 2017. Licence to 'Dependent' is Gnu GPL.
