//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Function;
import la.la.Value.Tuple;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** An Estimator estimates (fits) a fully parameterised Model to a
    data-set, thus solving the inference problem posed by some
    {@link UPModel UnParameterised} Model.  Most Estimators will be
    instances of {@link UPModel.Est}, {@link UPFunctionModel.Est} or
    {@link UPSeriesModel.Est}.
    Note that an Estimator can do all that a UPModel can do,
    {@link #sp2Model sp2Model(m1,m2,sp)} &amp;
    {@link #stats    stats(ds)}, and in addition
    {@link #ds2Model ds2Model(ds)} &amp;
    {@link #ss2Model ss2Model(ss)}. In general these last two
    depend on the Estimator's own {@link #params parameter(s)}
    which may, for example, control a <I>prior</I>.  */
public abstract class Estimator extends Function                 // * Estimator
 { /** The Estimator's parameters, notably the parameters
       of a <I>prior</I> on the estimated-Model's statistical parameters.
       Note that 'params' may be <CODE>triv</CODE>ial.  */
   public final Value params;

   /** An Estimator may have parameters, notably
       parameters of a <I>prior</I>.  */
   public Estimator(Value params)
    { this.params = params; }// constructor

   /** Apply, 'this' Estimator to a data-set, <CODE>ds</CODE>, that is,
       return the Model {@link #ds2Model ds2Model(ds)}.
       Also see {@link Function#apply}.  */
   public final Model apply(Value ds)                            // ! apply(ds)
    { return ds2Model((Vector) ds); }

   /** Return problem-defining parameter,
       for example, by {@link UPModel#defnParams()}.  */
   public abstract Value defnParams();                        // * defnParams()

   /** Given part&nbsp;1 and part&nbsp;2 message lengths, m1 and m2, and
       statistical parameter(s), sp, return a fully parameterised Model.
       Also see UPModel.{@link UPModel#sp2Model sp2Model(m1,m2,sp)}.  */
   public abstract Model sp2Model(double m1, double m2, Value sp);         // *

   /** {@link #stats(Vector,int,int) stats(0,ds.nElts())}.  There is
       unlikely to be any need for a subclass to override stats(ds).  */
   public Value stats(Vector ds)                                   // stats(ds)
    { return stats(ds, 0, ds.nElts()); }

   /** Given a data-set 'ds', return sufficient statistics for elements
       [lo,&nbsp;hi), lo&nbsp;inclusive to hi&nbsp;exclusive,
       <SPAN STYLE="white-space:nowrap">ss = stats(ds,lo,hi)</SPAN>.
       'ss'&nbsp;is used to estimate a Model as in
       {@link #ss2Model ss2Model(ss)}, say.
       More on stats [{@link UPModel#stats(Vector,int,int) here}].  */
   public abstract Value stats(Vector ds, int lo, int hi);       // * stats(,,)

   /** Combine {@link #stats(Vector,int,int) statisticses}
       ss0 and ss1, '+'&nbsp;if add=true otherwise&nbsp;'-'.  */
   public abstract Value stats(boolean add, Value ss0, Value ss1);         // *

   /** {@link #stats(boolean,Value,Value) Add} (remove)
       {@link #stats(Vector,int,int) statistics} for ds.[lo,hi) to (from)
       statistics ss0. There is unlikely to be any need for a subclass
       to override stats(add,ss0,ds,lo,hi).  */
   public Value stats(boolean add, Value ss0, Vector ds, int lo, int hi)
    { return stats(add, ss0, stats(ds, lo, hi)); }

   /** Given a data-set, ds, return
       {@link #ss2ModelSp ss2ModelSp}({@link #stats stats(ds)});
       read {@link #ss2ModelSp <B>ss2ModelSp</B>(ss)} carefully!  */
   public final Tuple ds2ModelSp(Vector ds)                 // ! ds2ModelSp(ss)
    { return ss2ModelSp(stats(ds)); }

   /** Given a data-set, ds, estimate a Model by
       <SPAN STYLE="white-space:nowrap">
          mdl = {@link #ss2Model ss2Model}({@link #stats stats(ds)})</SPAN>.
       Also see {@link #ds2ModelSp ds2ModelSp(ds)}!  */
   public final Model ds2Model(Vector ds)                     // ! ds2Model(ds)
    { return ss2Model(stats(ds)); }

   /** Given stats, ss, estimate a Model and parameters; this default is
       (mdl,&nbsp;sp), where mdl=ss2Model(ss) and sp=mdl.statParams()
       (usually).  <I>However</I>, there is at least one exception &ndash;
       Model.{@link Model#asUPModel() asUPModel()}'s estimator(()) &ndash;
       which returns (mdl,&nbsp;triv). The distinction ss2ModelSp:ss2Model,
       and ds2ModelSp:ds2Model, can matter to a compound UnParameterised
       Model handed a Model.asUPModel().  */
   public Tuple ss2ModelSp(Value ss)                          // ss2ModelSp(ss)
    { final Model mdl = ss2Model(ss);
      return ss2Model(ss).pair(mdl, mdl.statParams());
    }//ss2ModelSp

   /** Given sufficient statistics, <CODE>ss={@link #stats stats(ds)}</CODE>,
       of a data-set, ds, estimate a fully parameterised Model, ss2Model(ss).
       Also see {@link #ss2ModelSp ss2Model(ss)}!  */
   public abstract Model ss2Model(Value ss);                  // * ss2Model(ss)

   /** Return a String representation of 'this' Estimator. */
   public String toString()                                       // toString()
    { return "(" + getClass().getName() + " " + params + ")"; }

   /** In some context, it might(?) be necessary to treat 'this'
       Estimator as an {@link UPModel UnParameterised} Model, upm, where
       upm.{@link UPModel#estimator estimator(triv)} always returns
       'this' Estimator.  Also see {@link Model#asUPModel()}.  */
   public UPModel asUPModel()                          // Estimator.asUPModel()
    { return new UPModel(defnParams())
       { /** Given msg1, msg2, and statistical parameters 'sp',
             return a Model, as per 'this' Estimator.  */
         public Model sp2Model(double msg1, double msg2, Value sp)
          { return Estimator.this.sp2Model(msg1, msg2, sp); }
         /** Calculate sufficient statistics, as per 'this' Estimator.
             More on stats {@link UPModel#stats here}.  */
         public Value stats(Vector ds, int lo, int hi)
          { return Estimator.this.stats(ds, lo, hi); }
         public Value stats(boolean add, Value ss0, Value ss1)
          { return Estimator.this.stats(add, ss0, ss1); }
         /** Return 'this' Estimator; note that ps must be triv.
             We already have 'this' Estimator so we don't need
             any more (non-trivial) parameters.  */
         public Estimator estimator(Value ps /* = triv! */)
          { if( ps != triv ) error( ".estimator(" + ps + ")" );
            return Estimator.this;
          }//estimator(triv)
         public String toString()
          { return Estimator.this + ".asUPModel()"; }
    }; }//asUPModel()

 }//Estimator class

// (c) Lloyd Allison, 2012, Tue 15 Aug 2017. Licence to 'Estimator' is Gnu GPL.
