//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Lexical;
import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised Geometric0 Model. For most purposes
    {@link MML#Geometric0} and {@link Geometric0UPM.M} should be enough,
    with little or no need to call upon Geometric0UPM directly?
    (The '0' is to remind us that the Model is for integers &ge;&nbsp;0.) */
public class Geometric0UPM extends Discretes                   // Geometric0UPM
 { public Geometric0UPM(Value t /* = triv */)
    { super(t);
      if( t != triv ) error( ", t=" + t );
    }//constructor

   /** Given m1, m2 and &mu; return a fully parameterised
       {@link M Geometric} Model. */
   public M sp2Model(double m1, double m2, Value mu)
    { return new M(m1, m2, mu); }

   /** Returns {@link Discretes#NandSum NandSum(ds,lo,hi)}.
       More on statistics {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { return Discretes.NandSum(ds, lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return (Vector)
         ((Vector) ss0).bOp(add?Lexical.plus:Lexical.minus, (Vector) ss1);
    }

   /** The Estimator has a parameter, 'AA', the parameter of
       the {@link ExponentialUPM.M Exponential} prior on
       the mean of the Geometric distribution.  */
   public Est estimator(final Value AA /* Cts */)
    { return new Est(AA)
       { /** 'A' is the {@link #priorPdf prior}'s parameter, as a double. */
         public final double A;
         { this.A = AA.x(); }//constructor

         /** The pdf of the <I>prior</I> on the Geometric's &mu;,
             <NOBR>h(&mu;) = (1/A).e<SUP>-&mu;/A</SUP>;</NOBR> the
             prior itself has mean&nbsp;A so it is usually quite easy to
             think up a suitable value. */
         protected double priorPdf(double mu)
          { return Math.exp( - mu / A) / A; }

         /** Given sufficient statistics,
             <NOBR>ss = {@link Geometric0UPM#stats stats(ds)},</NOBR>
             of a data-set, ds, return a fully parameterised
             {@link Geometric0Model Geometric} Model. */
         public M ss2Model(Value ss)
          { Vector v = (Vector) ss;
            double N = v.elt(0).x(), sum = v.elt(1).x();
            double mu = (1 - N - 1/A +
               Math.sqrt(N*N + 1/(A*A) + 1 + 2*N/A -2*N + 4*sum/A)) / (2/A);
            double halfLogF = (Math.log(N)-Math.log(mu)-Math.log(mu+1))/2.0;
            Value.Real muR = Value.real(mu);
            double msg1 = - Math.log(priorPdf(mu))            // pdf
                        + halfLogF
                        + Math.log(latticeConstant(1)) / 2;
            msg1 = Math.max(0.0, msg1);                       // paranoia
            double msg2 = new M(0.0,0.0,muR).nlLH(ss) + 0.5;  // i.e. + D/2
            return new M(msg1, msg2, muR);
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(AA)

   public String toString()
    { return this == Geometric0 ? "Geometric0" : super.toString(); }
   // ..............................................

   /** The fully parameterised Geometric Model (probability distribution)
       on integers in <NOBR>[<B>0</B>, &infin;);</NOBR> the {@link #mu mean},
       &mu;, is its one statistical-parameter.  Also see the
       UnParameterised {@link MML#Geometric0} and class
       {@link Geometric0UPM} for estimation etc., and
      [<A HREF="http://www.allisons.org/ll/MML/Discrete/Integers/Geometric/"
          TARGET="EX">www</A>].  */
   public class M extends Discretes.M                     // Geometric0UPM.M
    { /** Statistical parameter &mu;, the mean, as a double. */
      public  final double mu;
      private final double logMu, logMu1, lm1_lm;

      public M(double msg1, double msg2, Value mu)
       { super(msg1, msg2, mu);
         this.mu = ((Value.Cts) mu) .x();
         logMu   = Math.log(this.mu);
         logMu1  = Math.log(this.mu + 1);
         lm1_lm  = logMu1 - logMu;
       }//constructor

      /** The negative log probability of n; <B>note</B>, n&ge;0. */
      public double nlPr_n(int n)
       { if( n < 0 ) error( ".nlPr_n(n), n=" + n + "<0" );
         return n * lm1_lm + logMu1;
       }//nlPr_n(n)

      /** Given sufficient statistics, ss = stats(ds), of a data-set ds,
          return the negative log LikeliHood of ds. */
      public double nlLH(Value ss)
       { Vector v = (Vector) ss;  // i.e. [N, sum]
         double N = v.elt(0).x(), sum = v.elt(1).x();
         return sum * lm1_lm + N * logMu1;
       }//nlLH(ss)

      /** Return a random number from 'this' Geometric distribution;
          calls {@link #random_n()}. */
      public Value.Int random()
       { return new Value.Int(random_n()); }

      /** ???<B>random_n() could be made much(!) more efficient!!!</B>
          This implementation is just for completeness!  */
      public int random_n()
       { int i;  double rndm = RNG.nextDouble();
         for( i = 0; rndm >= 0 && i < mu*100; i++ ) rndm -= pr_n( i );
         return i-1;
       }//random_n()

    }//Geometric0UPM.M, the Model

 }//Geometric0UPM, the UPModel

// (c) L. Allison, 2012, Fri 25 Nov 2015. Licence to 'Geometric0UPM' is Gnu GPL.
