//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A Tuple of UnParameterised Models makes an UnParameterised Model
    of Tuples, i.e., of Multivariate data, and a Tuple of statistical
    parameters make the statistical parameter of the Model, etc..
    The components of a datum Tuple are modelled <I>independently</I>.
    Also see {@link Dependent Dependent}, and the fully parameterised
    {@link Independent.M} Model.  */
public class Independent extends Multivariate                    // Independent
 {
   /** The sub-UPModels making 'this' Independent.
       Also see {@link M#ms M.ms[]}.  */
   public  final UPModel[] upms;  // the sub-UPModels

   /** The number of {@link #upms upms} and of columns in the data. */
   private final int       width; // # of columns (variables)

   /** Given a Tuple of UnParameterised Models, 'upms',
       construct an UnParameterised Model of Tuples &mdash;
       of multivariate data. See {@link #upms upms[]}.  */
   public Independent(Value upms)
    { super(upms);
      Value.Tuple upmsT = (Tuple) upms;
      width = upmsT.nElts();
      this.upms = new UPModel[width];
      for( int i = 0; i < width; i ++ )
         this.upms[i] = (UPModel) upmsT.elt(i);
    }//constructor

   /** The number of variables (columns) and of {@link #upms upms[]}. */
   public int width() { return width; }

   /** Statistical parameter
       sps&nbsp;=&nbsp;&langle;upms[0]'s,&nbsp;...&nbsp;etc.&rangle;. */
   public M sp2Model(double m1, double m2, Value sps)
    { return new M(m1, m2, sps); }

   /** Given a multivariate data-set, ds, return sufficient statistics,
       ss, a Tuple of course, one element per sub-{@link #upms upms}.
       Note that the components of a datum are modelled
       <I>independently</I> so an element of ss depends
       on just the corresponding column of&nbsp;ds.
       May or may not equal {@link M#stats(Vector,int,int)}.
       More on stats {@link UPModel#stats here}. */
   public Value.Tuple stats(Vector ds, int lo, int hi)
    { if( ds.nCols() != width )
         error( "data-set ds does not have " + width + " columns" );
      Value[] ss = new Value[width];
      for( int i = 0; i < width; i ++ )  // per column of the data
         ss[i] = upms[i].stats( ds.col(i), lo, hi );
      return Value.tuple(ss);
    }//stats(ds,lo,hi)

   /** Calls upon the stats(,,) of the {@link #upms upms[]}.
       May or may not equal {@link M#stats(boolean,Value,Value)}. */
   public Value.Tuple stats(boolean add, Value ss0, Value ss1)
    { Value.Tuple ss0t = ss0.isTuple(width), ss1t = ss1.isTuple(width);
      Value [] ss = new Value[width];
      for( int i = 0; i < width; i ++ )  // per column of the data
         ss[i] = upms[i].stats(add, ss0t.elt(i), ss1t.elt(i));
      return Value.tuple(ss);
    }//stats(add,ss0,ss1)

   /** Given a Tuple ps, where ps.elt(i) is the parameter of sub-Model
       Estimator {@link #upms upms}[i].estimator(ps.elt(i)), return an
       Estimator for the Independent {@link M Model}. */
   public Est estimator(final Value ps /* a Tuple */)
    { final Value.Tuple psT = ps.isTuple(width);
      return new Est(ps)
       { /** Given ss={@link #stats(ds)}, estimate a {@link M Model}. */
         public M ss2Model(Value ss)
          { /** ssT the sub-Models' sufficient statistics. */
            Value.Tuple ssT = ss.isTuple(width);
            Value [] sps = new Value[width];  // stat params []
            double msg1 = 0, msg2 = 0;
            for( int i = 0; i < width; i ++ )
             { Model m = upms[i].estimator(psT.elt(i)).ss2Model(ssT.elt(i));
               msg1 += m.msg1();  msg2 += m.msg2();
               sps[i] = m.statParams();
             }
            return new M(msg1, msg2, Value.tuple(sps));
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(t)
   // .........................................................................

   /** A fully parameterised Model of Tuples made from a Tuple of
       independent Models by the UnParameterised {@link Independent} Model.
       Also see {@link Dependent.M}. */
   public class M extends Multivariate.M                       // Independent.M
    {
      /** The sub-Models making 'this' Independent.M Model.
          Also see {@link #upms upms[]}. */
      public final Model[] ms;

      /** Given two-part message lengths, msg1 and msg2, and a
          Tuple of statistical parameters, sps, construct a Model
          of Tuples from the upms. */
      public M(double msg1, double msg2, Value sps)
       { super(msg1, msg2, sps);  Value.Tuple spsT = sps.isTuple(width);
         ms = new Model [width];
         for( int i = 0; i < width; i ++ )
            ms[i] = upms[i].apply(spsT.elt(i));
       }//constructor

      /** The negative log probability of a Tuple-datum, 'd', where the
          elements (fields, components, columns) of d are modelled
          independently. */
      public double nlPr(Value d)
       { Value.Tuple dT = d.isTuple(width);
         double ans = 0;
         for( int i = 0; i < width; i ++ ) ans += ms[i].nlPr(dT.elt(i));
         return ans;
       }//nlPdf(d)

      /** Given a multivariate data-set, ds, return statistics,
          ss, a Tuple of course, one element per sub-{@link #ms ms}.
          Note that the components of a datum are modelled
          <I>independently</I> so an element of ss depends
          on just the corresponding column of&nbsp;ds.  May or may
          not equal {@link Independent#stats(Vector,int,int)}.
          More on stats {@link UPModel#stats here}. */
      public Value.Tuple stats(Vector ds, int lo, int hi)
       { if( ds.nCols() != width )
            error( "data-set ds does not have " + width + " columns" );
         Value[] ss = new Value[width];
         for( int i = 0; i < width; i ++ )  // per column of the data
            ss[i] = ms[i].stats( ds.col(i), lo, hi );
         return Value.tuple(ss);
       }//stats(ds,lo,hi)

      /** Calls upon the stats(,,) of the {@link #ms ms[]}.  May or may
          not equal {@link Independent#stats(boolean,Value,Value)}. */
      public Value.Tuple stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(width), ss1t = ss1.isTuple(width);
         Value [] ss = new Value[width];
         for( int i = 0; i < width; i ++ )  // per column of the data
            ss[i] = ms[i].stats(add, ss0t.elt(i), ss1t.elt(i));
         return Value.tuple(ss);
       }//stats(add,ss0,ss1)

      /** Given sufficient statistics,
          ss&nbsp;=&nbsp;{@link #stats(Vector,int,int) stats(ds)}, of a
          data-set, ds, return the negative log likelihood of&nbsp;ds. */
      public double nlLH(Value ss)
       { Value.Tuple ssT = ss.isTuple(width);
         double sum = 0;
         for( int i = 0; i < width; i ++ ) sum += ms[i].nlLH(ssT.elt(i));
         return sum;
       }//nlLH(ss)

      /** Return a random Tuple, assuming each of the
          {@link #ms sub-Models} can play its part. */
      public Value.Tuple random()
       { Value [] vs = new Value[ width ];
         for( int i = 0; i < width; i ++ )
            vs[i] = ms[i].random();
         return Value.tuple(vs);
       }//random()

    }//Independent.M class, the Model

 }//Independent class, the UPModel

// (c) L. Allison 2012, 4 December 2017. Licence to 'Independent' is Gnu GPL.
