//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
import la.maths.Vector.Derived; // see stats(...)
import la.maths.Vector.Slice;   // see stats(...)

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised Laplace. Usually {@link MML#Laplace} suffices, but
    here is its class. Fully parameterised is {@link M M<SUB>&mu;b</SUP>}.
    Also see the (negative-) {@link ExponentialUPM Exponential}.  */
public class LaplaceUPM extends Continuous                        // LaplaceUPM
 {
   /** Requires definition parameters dp = triv. */
   public LaplaceUPM(Value dp)
    { super(dp); if( dp != triv ) error( ".constructor(" + dp + ")" );
    }//constructor

   /** Given two part message lengths, msg1 and msg2, and
       statistical parameters, sp, return an {@link M M}. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }


   /** Given a data-set, ds, return statistics,
       <NOBR> ss = stats(ds,lo,hi), </NOBR> that is
       ds.slice[lo,hi).sorted(), a Sorted Slice.
       Can't think of anything better, not if you want the
       ability to add more elements of ds and update ss. */
   public Value stats(Vector ds, int lo, int hi)
    { return ds.slice(lo,hi).sorted(); }

   /** Given sufficient statisticses 'ss0' and 'ss1', either
       'add' them or remove (add=false) ss1 from ss0. Also
       see {#link #stats(Vector,int,int) stats(ds,lo,hi)}. */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { Vector ss0v = ((Derived) ss0).parent();
      Slice  ss1v = (Slice) ((Derived) ss1).parent();
      return ss1v.combine(add, ss1v).sorted();
      // Could do better with a suitable sorted 'merge' operation
    }//stats(add,ss0,ss1)


   /** Parameters ps = (&mu;<SUB>min</SUB>, &mu;<SUB>max</SUB>,
       b<SUB>min</SUB>, b<SUB>max</SUB>). A uniform prior on &mu; in its
       range; log(b) uniform, b~1/b, in its range. Return an Estimator for
       a fully {@link M parameterised} Laplace probability distribution. */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { Tuple psT = ps.isTuple(4);
         double muMin = psT.elt(0).x(), muMax = psT.elt(1).x(),
                bMin  = psT.elt(2).x(), bMax  = psT.elt(3).x();
          { assert muMin<muMax && 0<bMin && bMin<bMax : errMsg(" ps"); }

         /** Given ss = stats(ds) of a data-set, ds, return an M. */
         public M ss2Model(Value ss)
          { double R_mu = muMax - muMin,     // h(mu) uniform over R_mu
                   R_b  = Math.log(bMax) - Math.log(bMin), // h(b) ~ 1/b
                   k2   = latticeConstant(2);
            Vector ssV  = (Vector) ss;       // the sorted data-set
            final double N = ssV.wts();      // effective |data-set| (NB wts)
            final int midPt = ssV.wtMedian();
            double mu = N > 0 ? ssV.elt(midPt).x()     // est mu ~ median
                              : (muMax + muMin) / 2.0; // else mid range
            mu = Math.max(muMin, Math.min(muMax, mu)); // and bounds
            double diffs = 0;
            for( int i = 0; i < N; i ++ )
               diffs += Math.abs( ssV.elt(i).x() - mu ) * ssV.wt(i);  // wt!
            double b = N > 1 ? diffs / (N-1)       // est b ~ avg |diff|
                             : Math.sqrt(bMin * bMax); // else geom mean
            b = Math.max(bMin, Math.min(bMax, b));         // and bounds
            M mdl = (M) sp2Model(0, 0, pair(real(mu), real(b)));
            // h(mu,b) = 1/(R_mu b R_b), where R_b is the range of log b,
            // Fisher, F = N^2 / b^4
            // msg1 = -log(h(mu.b)) + 1/2 log F + 2/2 log k2
            //      = log(R_mu R_b) log N - log b + log k2
            double msg1 = Math.log(R_mu*R_b) + Math.log(N) - 2 * Math.log(b)
                        + Math.log(k2),
                   msg2 = mdl.nlLH(ss) + 1;  // NB. +1 = 2*0.5
            assert msg1 >= 0 && msg2 >= 0 : "msg1=" + msg1 + ", msg2=" + msg2;
            return (M) mdl.asGiven(msg1, msg2);
          }//ss2Model(ss)

    }; }//estimator(ps)

   public String toString()
    { return this == Laplace ? "Laplace" : super.toString(); }
   // ....................................

   /** The fully parameterised Laplace probability distribution.
       Also see it {@link MML#Laplace UnParameterised}. */
   public class M extends Continuous.M                       // LaplaceUPM.M
    { /** The median, &mu;, and the scale, b, of this Model. */
      public final double mu, b;

      /* Statistical parameters sp = ({@link #mu &mu;}, {@link #b b}).
         See {@link #nlPdf_x nlPdf_x}.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple spT = sp.isTuple(2);
         mu = spT.fst().x(); b = spT.snd().x();
       }//constructor

      /** The pdf is <NOBR><SUP>1</SUP>/<SUB>2b</SUB>
          e<SUP><FONT SIZE="-2">-|x-&mu;|/b</SUP></FONT></NOBR>,
          its negative log being <NOBR>|x-&mu;|/b + log(2b)</NOBR>.
          This amounts to two (negative)
          {@link ExponentialUPM.M Exponentials} back to back at &mu;. */
      public double nlPdf_x(double x)
       { return Math.abs(x-mu)/b + Math.log(2*b); }

      /** Assumes that {@link LaplaceUPM#stats(Vector,int,int) statistics}
          are ~the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }

      /** Generate a {@link #random random} (double) from 'this' Model. */
      public double random_x()
       { double d = RNG.nextDouble();               // d in [0, 1)
         double l = b * Math.log(RNG.nextDouble()); // l < 0
         return d < 0.5 ? mu - l : mu + l;  // also see (neg-) exp
       }//random_x()

    }//LaplaceUPM.M class, the Model

 }//LaplaceUPM class, the UPModel

// (c) Lloyd Allison 2012, Fri 27 June 2014. Licence to 'LaplaceUPM' is Gnu GPL.
