//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import static la.la.Value.*;
import la.maths.Vector;
import la.maths.Matrix;
import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParametrised LinearD function-model; also see the
    fully {@link LinearD.M parameterised} LinearD function-model.
    The function-model of <B>R</B><SUP>D</SUP>&rarr;<B>R</B> is
    y=&sum;<SUB>j=0..D-1</SUB>{a<SUB>j</SUB>x<SUB>j</SUB>
     + b + N<SUB>0,&sigma;</SUB>}.
    Matrix X=[x<SUB>i,j</SUB>], i=0..N-1, j=0..D-1, and
    vector Y =[y<SUB>i</SUB>], i=0..N-1.  As is customary,
    define X<SUB>+</SUB> to be X augmented to the right with an
    extra column of 1s, and 'b' becomes a<SUB>D</SUB>.
    Also&nbsp;see {@link Linear1}, <B>R</B>&rarr;<B>R</B>.  */
public class LinearD extends UPFunctionModel                         // LinearD
 {
   /** See {@link #D(). */
   private final int D;
   /** D(), the dimension of the input variable x:R<SUP>D</SUP> in x&rarr;y. */
   public int D() { return D; }

   /** The problem definition parameter gives {@link #D()}&gt;1.  */
   public LinearD(Value D)
    { super(D);
      this.D = ((Int) D).n();
    }//constructor

   /** Calculate the  sufficient statistics of ds[lo,hi).  That is
       &langle;N,&nbsp;X<SUB>+</SUB><SUP>T</SUP>&nbsp;X<SUB>+</SUB>,
          Y<SUP>T</SUP>&nbsp;X<SUB>+</SUB>, Y.Y, nlAoM&rangle;.
       Note that X<SUB>+</SUB> is the matrix of x-values augmented
       with a column of '1's to the right, and that nlAoM is the
       nlAoM of the y-values (the xs being common knowledge).  */
   public Tuple stats(Vector ds, int lo, int hi)
    { if( ds.nCols() != 2 )
         error( ".stats(ds,lo,hi), ds is not 2-columns" );
      Vector xs = ds.col(0), ys = ds.col(1);
      if( xs.nCols() != D )
         error( ".stats(ds,lo,hi), x is not " + D + "-columns" );
      double       y2   = 0;  // ys.ys
      double []    yT_x = new double [D+1];        // y dot (cols of xs)
      double [] [] xT_x = new double [D+1] [D+1];  // xs' dot xs
      for( int i = 0; i <= D; i ++ )      // initialise...
       { yT_x[i] = 0;
         for( int j = 0; j <= D; j ++ ) xT_x[i][j] = 0;
       }//for i
      for( int i = lo; i < hi; i ++ )                       // i
       { double wi = ds.wt(i),  yi = ys.elt(i).x();;
         y2 += yi * yi * wi;
         for( int j = 0; j <= D; j ++ )                     // j
          { double xij = j < D ? xs.elt(i, j).x() : 1.0;
            // pretend x[][] has an extra column of ^^ ones
            yT_x[j] += yi * xij * wi;
            for( int k = 0; k <= D; k ++ )                  // k
             { double xik = k < D ? xs.elt(i, k).x() : 1.0;
               xT_x[j][k] += (xij * xik) * wi;
             }//for k
          }//for j
       }//for i
      return tuple( new Value [] {
         real(ds.wts(lo,hi)),           // i.e., the N of the data
         Matrix.doubles(xT_x),          // xs' dot xs
         Vector.doubles(yT_x),          // (cols of xs) dot y
         real(y2),                      // ys.ys
         real(ys.wtNlAoM(lo, hi)) } );  // nlAoM of ys
    }//stats(ds,lo,hi)

   /** Combine {@link #stats(Vector,int,int) statistics} ss0 and ss1,
       either ss0 "+" ss1, or ss0 "-" ss1,
       depending on 'add' being true or false.  */
   public Tuple stats(boolean add, Value ss0, Value ss1)
    { return (Tuple)
        ((Tuple) ss0) . bOp(add?la.la.Lexical.plus:la.la.Lexical.minus,
         (Tuple) ss1    );
    }//stats(add,ss0,ss1)

   /** Given first- and second-part message lengths, msg1 and msg2, and
       statistical parameter(s), sp, return a fully parameterised
       {@link M LinearD.M} Model. */
   public LinearD.M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); } // for the sake of the type


   /** Return an {@link Est Estimator} of a fully parameterised
       {@link M LinearD.M} Model.  */
   public Est estimator(Value ps) { return new Est(ps); }

   /** Estimator for a fully parameterised
       {@link M LinearD.M Linear}-Model. */
   public class Est extends UPFunctionModel.Est                  // LinearD.Est
    { /** Bounds on &sigma;; {@link #nlPrior prior},
          pr(&sigma;)~1/&sigma;.  */
      public final double sigmaMin, sigmaMax;
      /** 1/&sigma; "range", log(sigmaMax)-log(sigmaMin).   */
      protected final double R_s;
      /** {@link MML#latticeConstant latticeConstant(D)}. */
      protected final double kD;

      /** The estimators parameter 'ps' is the bounds on &sigma;. */
      public Est(Value ps)
       { super(ps);
         final Tuple bds = ps.isTuple(2);
         sigmaMin = bds.fst().x(); sigmaMax = bds.snd().x();
         assert 0 < sigmaMin && sigmaMin < sigmaMax
              : errMsg( ", Est(ps)" );
         R_s = Math.log(sigmaMax) - Math.log(sigmaMin); // pr(s) ~ 1/s
         kD  = latticeConstant(D);
       }//constructor

      /** Negative log prior (probability density). */
      public double nlPrior(Vector a, double sigma)
       { double ans = Math.log(R_s);                     // sigma and ...
         for( int i = 0; i <= D; i ++ )
          { final double phi  = Math.atan(a.elt(i).x());
            ans += logPI - 2*Math.log(Math.cos(phi));    // ... a[i]
          }
         return ans;
       }//nlPrior(,)

      /** Log Fisher. */
      public double logF(double N, double sigma, Matrix xT_x)
       { // NB. remember blah[D1] refers to entry for const "b" (and x[D1]=1)
         return log2 + Math.log(N)           // from sigma's term
              - 2 * (D+2) * Math.log(sigma)  // every row/col has a 1/s^2 factor
              + xT_x.logDeterminant();
       }//logF(,,,)

      /** Given sufficient {@link LinearD#stats(Vector,int,int) statistics}
          ss=stats(ds) of a data-set 'ds', estimate a fully
          {@link LinearD.M parameterised} Linear Model.  */
      public M ss2Model(Value ss)
       { final Tuple ssT = ss.isTuple(5); // unpack ss ...
         double N     = ssT.elt(0).x(),
                y2    = ssT.elt(3).x(),
                nlAoM = ssT.elt(4).x();
         Matrix xT_x  = (Matrix) ssT.elt(1);  // (D+1)x(D+1)
         Vector yT_x  = (Vector) ssT.elt(2);
         Vector a  = xT_x.inverse().times( yT_x );  // NB. last elt is "b"
         //   sq_error = errs.errs,   where errs_i = y_i - x_i a
         // = (y - x a).(y - x a)
         // = y.y - 2 (y^t x) a + a^t (x^t x) a
         double s2 = (y2 - 2 * yT_x.dot(a) + a.dot(xT_x.times(a))) / (N - 1);
         double s  = Math.sqrt(s2);
                s  = Math.max(sigmaMin, Math.min(sigmaMax, s)); // & bounds
         // System.out.print( " {a=" + a + ",s=" + s + "} " ); // DEBUG
         Tuple sp = pair(a, real(s));  // stat params
         M mdl = LinearD.this.sp2Model(0, 0, sp);
         double msg1 = nlPrior(a, s)
                     + 0.5 * logF(N, s, xT_x)
                     + Math.log(kD),
                msg2 = mdl.nlLH(ss) + D*0.5;
         return LinearD.this.sp2Model(msg1, msg2, sp);
       }//ss2Model(ss)
    }//Est class

   // .........................................................................
   /** The fully parameterised function-model;
       {@link LinearD} is the UnParametrised function-model.
       Also see {@link Linear1.M}.  */
   public class M extends UPFunctionModel.M                        // LinearD.M
                  implements HasPdf
    { /** Statistical parameter, 'a[D+1]' of the LinearD Model
          y=a&middot;x+b+N(0,&sigma;) where 'b' <I>is</I> a[D].
          (The maths is easier if we define X<SUB>+</SUB> to be X
          with an extra column of 1s, and 'b' becomes a[D].)  */
      public final double [] a;  // of length D
      /** Vector version of a[] for condModel(x). */
      private Vector aV;
      /** Statistical parameter of the Model
          y=a&middot;x+b+N(0,&sigma;).  */
      public final double    sigma;
      /** Real version of sigma for condModel(x).  */
      private final Real sigmaR;
      /** log(2&pi;)/2+log(&sigma;), and
          2 &sigma;<SUP>2</SUP> (used by {@link #nlLH}).  */
      private final double log_root_2_pi_sigma, two_sigma_2;

      /** Two part message lengths, msg1 and msg2,
          and statistical parameters sp.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple spT = sp.isTuple(2);
         aV = (Vector) spT.fst();
         a  = new double[D+1];
              for( int i = 0; i <= D; i ++ ) a[i] = aV.elt(i).x();
         sigmaR = (Real) spT.snd();  sigma = sigmaR.x();
         log_root_2_pi_sigma = 0.5*(log2 + logPI) + Math.log(sigma);
         two_sigma_2         = 2.0*sigma*sigma;
       }//constructor

      /** Conditional Model of output datum y given input datum x,
          i.e., the {@link NormalUPM.M Normal} Model,
          N({@link #a a}&middot;x+{@link #a b},{@link #sigma &sigma;)}. */
      public NormalUPM.M condModel(Value x)
       { // remember, 'b' *is* a[D] ...
         double y = aV.slice(0,D).dot( ((Vector) x).slice(0,D) ) + a[D];
         return Normal.sp2Model(0, 0, pair(real(y), sigmaR));
       }//condModel(id)

      /** Negative log pdf of 'y' {@link #condModel given}&nbsp;'x'. */
      public double nlPdf(Value xy)
       { Tuple xyT = xy.isTuple(2);
         return condModel(xyT.fst()).nlPdf(xyT.snd());
       }//nlPdf(xy)

      /** pdf of 'y' {@link #condModel given}&nbsp;'x'. */
      public double pdf(Value xy)
       { Tuple xyT = xy.isTuple(2);
         return condModel(xyT.fst()).pdf(xyT.snd());
       }//nlPdf(xy)

      /** Given sufficient {@link LinearD#stats(Vector,int,int) statistics},
          ss=stats(ds), of a data-set ds, return the
          negative log LikeliHood of ds. */
      public double nlLH(Value ss)
       { final Tuple ssT = ss.isTuple(5); // unpack ss ...
         double N     = ssT.elt(0).x(),
                y2    = ssT.elt(3).x(),
                nlAoM = ssT.elt(4).x();
         Matrix xT_x  = (Matrix) ssT.elt(1);  // (D+1)x(D+1)
         Vector yT_x  = (Vector) ssT.elt(2);
         //   sq_error = errs.errs,   where errs_i = y_i - x_i a
         // = (y - x a).(y - x a)
         // = y.y - 2 (y^t x) a + a^t (x^t x) a
         return N * log_root_2_pi_sigma
              + (y2 - 2 * yT_x.dot(aV) + aV.dot(xT_x.times(aV))) / two_sigma_2
              + nlAoM;
       }//nlLH(ss)

    }//LinearD.M class, the fully parameterised FunctionModel
// ----------------------------------------------------------------------------

   /** One very little test. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run LinearD --");

      // y ~ 2*x1 - x2 + 3
      double [] [] xx = new double [] []
             {{1,1}, {1,3}, {1,4}, {2,1}, {2,3}, {2,4}, {4,1}, {4,3}, {4,4}};
      double [] y = new double []
             { 4.0,   2.3,   0.7,   6,     4.3,   2.7,   10,    8.3,   6.7 };
      Vector xxy = Matrix.doubles(xx).zip(Vector.doubles(y, 0.1));
      Tuple ps = pair(point1, tenR);  // sigma bounds
      LinearD L2 = new LinearD(/* D= */ two);
      mml.Test.exercise( L2.estimator(ps), xxy );

      System.out.println( "(Also see mml.Test)" );

      System.out.println("-- done LinearD --");
    }//main

 }//LinearD class, the UPFunctionModel

// (c) Lloyd Allison 2017, Thu 11 May 2017. Licence to 'LinearD' is Gnu GPL.
