//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import java.util.BitSet;

import la.util.*;
import la.la.Value;
import static la.la.Value.*;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;
import la.maths.Matrix;
import static mml.MML.*;

import graph.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** 'MotifA', the UnParameterised adaptive Motif Model of'                          {@link Graph Graphs}.  The fully parameterised Model
    is&nbsp;{@link MotifA.M MotifA.M}.
    Also see {@link MotifD}.  */
public class MotifA extends Graphs.Motifs                             // MotifA
 {
   // protected graph.Type gType;        // from Graphs, also
   // public    graph.Type gType()

   // protected Discretes upmV;          // from Graphs, also
   // public    Discretes upmV()

   // protected Discretes.M mdlNmotifs;  // from Graphs.Motifs,
   // protected Model mdlMotif;          // from Graphs.Motifs, and
   // public double msgMotifs(Graph [] ms)
   // public double msgMotifs(Vector vm)

   /** The UnParameterised ({@link Adaptive Adaptive}) <I>background</I>
       Model over the existence of possible Edges outside instances of
       {@link M#motifs motifs}.  Also see {@link M#mdlE mdlE}. */
   public final Discretes.Bounded upmE =                                // upmE
      new Adaptive(pair( pair(zero, one),  // bounds
             Vector.doubles(new double[]{9, 1}) ));  // alpha, ie., 10+/-%

   /** Bounds [sgMinV, sgMaxV] on |V| for the motifs (patterns)
       that the {@link #estimator estimator} will search for
       (but not necessarily find). Also see {@link #mdlMotif mdlMotif}.  */
   public final int sgMinV, sgMaxV;

   /** Problem definition parameters
       <NOBR> dp = &lang;{@link #gType gType}, {@link #upmV  upmV},
           &lang;{@link #sgMinV sgMinV},
                 {@link #sgMaxV sgMaxV}&rang;&rang;. </NOBR>  */
   public MotifA(Value dp)
    { super(dp); Tuple dpT = dp.isTuple(3);
      gType = (graph.Type) dpT.fst();     // the population of Graphs
      upmV  = (Discretes)  dpT.snd();     // the UPModel of |V| >= 1
      Tuple bnds = dpT.thrd().isTuple(2); // [sgMinV, sgMaxV]
      sgMinV = bnds.fst().n(); sgMaxV = bnds.snd().n();

      mdlNmotifs = Geometric0.sp2Model(0, 0, twoR); // model of #motifs >= 0
      mdlMotif = new Graphs.GERadaptive(            // model of each motif
                  triple(gType,
                         new Discretes.Uniform(pair(new Int(sgMinV), // size
                                                    new Int(sgMaxV)) ),
                         Vector.doubles(new double[]{1, 1})         // alpha
                        )          )
              .sp2Model(0, 0, triv /* sp of Discretes.Uniform */ );
    }//constructor

   /** Sufficient statistics are elements [lo,&nbsp;hi)
       of the data-set ds itself. */
   public Vector stats(Vector ds, int lo, int hi)            // stats(ds,lo,hi)
    { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)          // stats(,,)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return a fully parameterised MotifA {@link MotifA.M Model}. */
   public MotifA.M sp2Model(double msg1, double msg2, Value sp) // sp2Model(,,)
    { return new M(msg1, msg2, sp); }

   /** Return an Estimator for a Motif {@link M Model}. Parameters 'ps' are
       for {@link #upmV upmV}.estimator(ps).  Estimating {@link M#mdlV mdlV}
       is standard but finding good {@link M#motifs motifs} is hard so
       heuristics are used.  The time complexity should be approximately
       O(#G.V.d<SUP>sgMaxV-1</SUP>), given a data-set of #G Graphs of
       average size V and average degree&nbsp;d.  */
   public Est estimator(final Value ps)                        // estimator(ps)
    { return new Est(ps)
       { public M ss2Model(Value ss)
          { final int topN = 20;  // # of candidate Motifs fed into greedy()
            Vector ds = (Vector) ss;           // i.e., data-set
            java.util.Vector <Graph>  mtfV = new java.util.Vector <Graph>  ();
            java.util.Vector <Int>    scrV = new java.util.Vector <Int>    ();
            java.util.Vector <BitSet> bitV = new java.util.Vector <BitSet> ();

            Discretes.M mdlV = ds2mdlV(ps, ds);  // of |V|

            int instances = 0;                      // next, seek motifs
            for( int i = 0; i < ds.nElts(); i ++ )  // each Graph g_i in ds
             { Graph g_i = (Graph) ds.elt(i);       // ... g_i
               final int g_i_V = g_i.vSize();
               Graph.SubGraphs sgs = g_i.subGraphs(sgMinV, sgMaxV);

               if( i > 0 ) // clear out any old coVers of previous g_i
                  for( int j = 0; j < bitV.size(); j ++ )
                     bitV.set(j, new BitSet(g_i_V));

               while( sgs.hasSome() )         // each sized subGraph of g_i
                { Graph.Induced sg = sgs.elt();         // sg
                  // ?could just skip low-density subGraphs?
                  instances ++ ;  // trace, stats
                  final Graph sgC = sg.canonical().flatten();
                  // sgC: if sg is novel we need sgC to add it and
                  //      if not we still need it to confirm isomorphism
                  boolean novel = true;  // tentatively
                  int j = 0;
                  for( ; j < mtfV.size(); j ++ )
                   { Graph mtfV_j = (Graph) mtfV.elementAt(j);
                     if( sgC.structurallyIdentical(mtfV_j) )
                      { novel = false; break; }          // seen sg before
                   }//for j

                  if( novel )  // add new instance and init info...
                   { mtfV.addElement(sgC);               // instance
                     scrV.addElement(zero);              // score = 0
                     bitV.addElement(new BitSet(g_i_V)); // empty
                   }//if; note old or novel, motif is at position 'j'

                  // it is good for a Motif to cover many Edges of g_i ...
                  int newEdges = 0;                          // Edges
                  for( int v = 0; v < sg.vSize(); v ++ )     // newly
                     if( ! bitV.elementAt(j).get(sg.vs[v]) ) // covered
                        newEdges += sg.degree(v);            // count * 2
                  scrV.setElementAt( new Int(scrV.elementAt(j).n()
                                            + newEdges), j);
                  bitV.elementAt(j).or(sgs.vertices());  // + sg's Vs

                  sgs.advance(); // next subGraph of g_i
                }//while sgs...
             }//for i

            // Sort the counted subGraphs on decreasing score, 'scrV[.]'.
            for( int i = 1; i < mtfV.size(); i ++ )
             { Graph  mtfV_i = mtfV.elementAt(i);
               Int    scrV_i = scrV.elementAt(i); int scrV_i_n = scrV_i.n();
               int j = i;
               for( ; j > 0; j -- )
                  if( scrV.elementAt(j-1).n() < scrV_i_n )
                   { mtfV.setElementAt(mtfV.elementAt(j-1), j);
                     scrV.setElementAt(scrV.elementAt(j-1), j);
                   }
                  else break; // >=
               if( j == i ) continue; // else...
               mtfV.setElementAt(mtfV_i, j);
               scrV.setElementAt(scrV_i, j);
             }//for i

// ZZZ...
if( verbosity > 0 )
{ int lastShown = 0, maxExtras = 5;
  System.out.println("   MotifA.estimator: "
                    + mtfV.size()+" motifs, "
                    + instances+" instances" );
  for( int i = 0; i <  mtfV.size(); i ++ )
   if( i < topN || i >= mtfV.size()-5 ||
       (maxExtras > 0 &&
        mtfV.elementAt(i).eSize() >= mtfV.elementAt(i).vSize() /* "dense" */ ) )
    { if( i > lastShown + 1 ) System.out.println( " . . . etc." );
      Graph m_i = mtfV.elementAt(i);
      System.out.println( i + ":" + m_i + ": " + scrV.elementAt(i)
                        + " " + m_i.eSize() + "e #" + m_i.hashCode() );
      lastShown = i;
      if( i >= topN && i < mtfV.size()-5 ) maxExtras -- ;
    }
}//if(...)

            // Choose a small number of the "best" motifs greedily ...
            Graph [] mtfArr = mtfV.toArray(new Graph [mtfV.size()]);
            Vector mtfVec = Vector.values(mtfArr); // not java.util.Vector!-)
            return (MotifA.M) greedy(
                        mtfVec.take(Math.min(topN, mtfArr.length)), mdlV, ds );
          }//ss2Model(ss)
       };//new Est(ps)
    }//estimator(ps)
   // .........................................................................

   /** The fully parameterised MotifA Model of Graphs.
       The UnParameterised Model is {@link MotifA}.
       Also see {@link MotifD.M}.  */
   public class M extends Graphs.Motifs.M                           // MotifA.M
    {

      // protected final Discretes.M mdlV; // from Graphs.M
      // also                        mdlV()

      /** The fully parameterised ({@link Adaptive.M Adaptive}) Model of
          that part of the Adjacency Matrix not covered by instances of
          {@link #motifs motifs[.]}. Also see {@link #upmE}. */
      public final Discretes.Bounded.M mdlE;                            // mdlE

      /** The motifs (patterns, templates), sorted (down) on |V|,
          that can be used to compress a given Graph. There will be good
          compression if there are many instances of motifs[.] in the Graph.
          The motifs may be "given" or {@link MotifA#estimator estimated}. */
      protected final Graph [] motifs;                              // motifs[]

      /** Returns {@link #motifs motifs[]};
          see {@link Graphs.Motifs.M#motifs()}. */
      public Graph [] motifs() { return motifs; }

      private final int [] [] moFreqs;     // Edge/not per motifs[.]
      private final int []    bgFreqs = new int [2];  // bg Edge/not

      /** |V| for the smallest and largest of the {@link #motifs motifs[.]}.
          Also see the <I>different</I> {@link MotifA#sgMinV} and sgMaxV. */
      public final int sgMinV, sgMaxV;

      /** Statistical parameters
          <NOBR> sp = &lang;sp<SUB>{@link #mdlV mdlV}</SUB>,
                            {@link #motifs motifs}&rang;. </NOBR>  */
      public M(double msg1, double msg2, Value sp)                  // MotifA.M
       { super(msg1, msg2, sp); Value.Tuple spT = sp.isTuple(2);
         mdlV = upmV.apply(spT.fst());        // (a)
         mdlE = ((Adaptive) upmE).Mdl;
         Vector sgV = (Vector) spT.snd();     // (b)
         moFreqs = new int [sgV.nElts()] [2]; // workspace for nlPr
         motifs  = new Graph  [sgV.nElts()];
         for( int i = 0; i < motifs.length; i ++ )  // each elt of motifs
          { // motifs[0..i-1] is sorted (down) in |V|
            Graph sg_i = ((Graph) sgV.elt(i)) .canonical();
            int sz = sg_i.vSize();
            int j = i;
            for( /* j=i above */ ; j > 0; j -- )
             { if( motifs[j-1].vSize() >= sz ) break;
               motifs[j] = motifs[j-1];
             }// now, j == 0 || sg[j-1].vSize() >= sz
            motifs[j] = sg_i;
          }//for i
         if( motifs.length > 0 )
          { sgMinV = motifs[0].vSize();
            sgMaxV = motifs[motifs.length-1].vSize();
            if( sgMinV < MotifA.this.sgMinV || sgMaxV > MotifA.this.sgMaxV )
               error( "sizes of given motifs out of range" );
          }
         else { sgMinV = 1; sgMaxV = 1; }
       }//constructor

      /** ss = {@link #stats(Vector,int,int) stats(ds,lo,hi)}, requires
          that ss is ds, the Vector (data-set) of Graphs itself. */
      public double nlLH(Value ss)                                  // nlLH(ss)
       { return sumNlPr((Vector) ss); }


      /** The negative log probability of datum (Graph)&nbsp;G.
          <B>TODO, a work in progress.</B>  */ // ZZZ
      public double nlPr(Value G)                                    // nlPr(G)
       { final Graph GG = (Graph) G;
    //   final Graph g = ((Graph) G).byDegree(false);           // g
         final int   V = GG.vSize(),          // |V|
                     E = GG.eSize();          // |E|
         double costE = 0;                    // cost of edges & non-edges
         int i, j;
         int v, w;                            // two Vertices of g
         int nMatch;
         int [] freqs = new int [2];
         boolean [] matched = new boolean [motifs.length];

         // vs[] : "low" degree, then "high" degree vertices (not a full sort)
         double perm = logFactorial(V); int threshold = 8;
         int [] vs = new int [V];
         i = 0; j = V-1;
         while( i <= j )
          { if(      GG.degree(i) <  threshold ) { vs[i] = i; i ++ ; }
            else if( GG.degree(j) >= threshold ) { vs[j] = j; j -- ; }
            else // deg[i] >= threshold > deg[j]  (so i!=j)
             { vs[i] = j; vs[j] = i; i ++ ; j -- ; }
          }// 0 <= i <= V  and  V > j >= -1  and  i = j+1
         perm = logFactorial(V+1)  // adaptive code on lo v. hi degree
              - logFactorial(i)
              - logFactorial(V-i);
         final Graph g = GG.renumbered(vs);   // g

         if( verbosity > 1 ) // ZZZ
            System.out.println( "   MotifA.M.nlPr: th=" + threshold + ", "
                              + i + ":" + (V-i) );

         // i = 0;
                // double deg_i = g.degree(i), deg_j;
         // for( j = 1; j < V; j ++ )     // perm: the cost of
          // { deg_j = g.degree(j);       // sorting the vertices
            // if( deg_i < deg_j )        // by degree
             // { perm += -logFactorial(j-i); // j > 1
               // i = j;  deg_i = deg_j;
          // }  }
         // perm -= logFactorial(j-i); // j > 1   TOO EXPENSIVE

         bgFreqs[0] = 0;  bgFreqs[1] = 0;          // initialise...
         for( int c = 0; c < moFreqs.length; c ++ )  // counters...
          { final int [] mc = moFreqs[c];
            mc[0] = 1; mc[1] = 1;
          }

         // 'Sent sent' is the graph transmitted so far, upto and including
         // Vertex 'i', with the next *possible* i-Edge to be assumed, or not.
         class Sent extends Graph
          { public int i = -1, i_v_w = -1;  // state ...
            public boolean isUndir = g.isUndirected(),
                           is_vw = true;    // ... ? Edge (v, w) ?
            public graph.Type type() { return g.type(); }
            public int vSize() { return i; }
            public boolean isEdge(int v0, int v1)
             { if( isUndir && v0 > v1 ) error("alarm " + v0 + "," + v1);
               if( v0 < i && v1 < i ) return g.isEdge(v0, v1); // transmitted
               final int i_v0_v1 = vw2int(v0, v1);
               if( i_v0_v1 <  i_v_w ) return g.isEdge(v0, v1); // transmitted
               if( i_v0_v1 == i_v_w ) return is_vw;  // i.e., next ?Edge?
               return false;   // in the far future
             }//isEdge(v0,v1)
          };//class Sent

         Sent sent = new Sent();

         for( i = 0; i < V; i ++ )                    // i
          { int i_deg = 0;
            for( j = 0; j < i; j ++ )                 // j < i, no loops
             { // Note, Und => (j,i),  Dir => <j,i> and <i,j>.
               v = j; w = i;              // v < w
               for(int dc = 0; g.isDirected() && dc < 2; dc ++ )
                { // first, supposition...
                  nMatch = 0;
                  freqs[0] = 0; freqs[1] = 0; // shouldn't get used
                  sent.i = i; sent.i_v_w = vw2int(v, w); // "state" of sent
                  sent.is_vw = true;
                  if( i_deg > 0 ) // so degree *might* truly be > 1
                   { for( int m = 0; m < motifs.length; m ++ )
                        matched[m] = false;
                     for( int m = 0; m < motifs.length; m ++ )
                        if( subGiso(motifs[m], sent, i, j, true) )         // *
                         { if( nMatch == 0 ) { freqs[0] = 0; freqs[1] = 0; }
                           nMatch ++ ;
                           matched[m] = true;
                           freqs[0] += moFreqs[m][0];
                           freqs[1] += moFreqs[m][1];
                         }
                 // following made tiny improvement, but slowed greatly...
                 //  if( option == 0 )
                 //   { sent.is_vw = false;
                 //     for( int m = 0; m < motifs.length; m ++ )
                 //        if( subGiso(motifs[m], sent, i, j, false) )     // *
                 //         { option = 2; break; }
                 //   }//if
                   }//if
                  // now for reality...
                  final boolean is_vw = g.isEdge(v, w); // next Edge?
                  final int is_vw_n = is_vw ? 1 : 0;

                  // for simplicity on the fly, not collected for later...
                  if( nMatch > 0 )
                     costE += Math.log( (freqs[0] + freqs[1])
                                      / (double) freqs[is_vw_n] );
                  // - log of a proper probability^^^^^^^^^^
                  // else no (hypothetical) motif match, treat bg edges later

                  if( nMatch == 0 ) bgFreqs[is_vw_n] ++ ;     // update...
                  else for( int m = 0; m < motifs.length; m ++ ) // ... or
                     if( matched[m] )
                        moFreqs[m][is_vw_n] ++ ;
                  if( is_vw && i != j ) i_deg ++ ;

                  if( v == w ) break;     // no need to swap & retry
                  v = i; w = j;           // if directed, try v > w
                }//for dc
             }//for j
          }//for i
         costE += mdlE.nlLH(Vector.ints(bgFreqs));
         // "background" edges -- ^^^^^^^^^^^^^suff stats for a binomial

         final double costV = mdlV.nlPr_n(V);  // note, |V| > 0

if( verbosity > 1 )
{ System.out.print( "   MotifA.M.nlPr: "
                  + "freqs:{"+bgFreqs[0]+","+bgFreqs[1]+"}, {" );
  for( int m = 0; m < motifs.length; m ++ )   // XXX
    System.out.print( (m > 0 ? ", " : "") +     // XXX
      "{" + moFreqs[m][0] + "," + moFreqs[m][1] + "}"); // XXX
  System.out.println( "} " );  // XXX

  System.out.println( "   MotifA.M.nlPr: V+pi+E:{"
    + Util.x2String( (costV/Math.log(2.0)) ) + "+"
    + Util.x2String( (perm /Math.log(2.0)) ) + "+"
    + Util.x2String( (costE/Math.log(2.0)) ) + "_bits} " );
}//if(...)

         return costV + perm + costE;
       }//nlPr(G)

   /** mv:tv, and mw:mustUse (where mv--mw), and then the rest.  */
   public boolean subGiso(Graph m, Graph sent, int tv, int mustUse, boolean jnd)
    { int count = 0, dist2 = 0;
      final int mV = m.vSize();
      BitSet mUsed = new BitSet(mV),  // for m's breadth-first thing
             tUsed = new BitSet(sent.vSize());  // for sent's rec thing
      int [] mvs = new int [mV], parent = new int [mV],
             tvs = new int [mV];      // mvs[i] : tvs[i]
      tvs[0] = tv;
      for( int i = 1; i < mV; i ++ ) tvs[i] = -1;
      final int deg_tv = sent.degree(tv);
      final int minDeg = jnd ? 2 : 1;    // min degree of mv

      for( int mv = 0; mv < mV; mv ++ )  // choose some vertex mv in m
       { final int deg_mv = m.degree(mv);
         if( deg_mv < minDeg /* ! */ || deg_mv > deg_tv ) continue; // else...

         mUsed.clear(); mUsed.set(mv);
         mvs[0] = mv; parent[0] = -1;

         int p = 0, q = 1;    // put vertices of m into mvs[] in ...
         while( q < mV )      // ... breadth first order
          { Series.Int ms = m.joinedTo(mvs[p]);
            while( ms.hasSome() )
             { final int mse = ms.elt_n();
               if( ! mUsed.get(mse) )            // it's new
                { mvs[q] = mse;  parent[q] = p;  // mvs[p]--mvs[q]
                  mUsed.set(mse); q ++ ;
                }
               ms.advance();
             }//while ms ...
            if( p == 0 ) dist2 = q; // eventually, 1st vertex @ dist=2 from mv
            p ++ ;
            if( p >= mV ) error( "p=" + p + " >=" + mV ); // impossible!
          }//while q ...

         // NB. mvs[1,dist2) joined to mv; mvs[dist2,mV) not, if any
         int lo = dist2, hi = mV;
         if( jnd ) { lo = 1; hi = dist2; }
         for( int i = lo; i < hi; i ++ )
          { tvs[i] = mustUse;
            tUsed.clear();  tUsed.set(tv); tUsed.set(mustUse);
            // now for p=1,2,... try to match mvs[p] of m to
            // a Vertex joined to tvs[anc[p]] in 'sent'
            count = rec(1, m,mvs,parent, sent,tvs, tUsed);  // NB '1'. *
            tvs[i] = -1;
            if( count > 0 ) break;
          }//for
         if( count > 0 ) break;
       }//for mv, if deg_mv

      return count > 0;
    }//subGiso( ... )


   /** Try to match vertex mvs[depth] of m to a Vertex of 'sent' that
       is joined to tvs[parent[depth]] of 'sent'. */
   public int rec(int depth,
                  Graph m,     int [] mvs, int [] parent,
                  Graph sent, int [] tvs, BitSet tUsed)
    { if( depth >= m.vSize() )  // done, mvs[i]:tvs[i], i = 0..mV-1
       {
// XXX System.out.print( " {M:G" );
// XXX for( int i = 0; i < depth; i ++ )
   // XXX System.out.print( ", " + mvs[i] + ":" + tvs[i] );
// XXX System.out.println( "} " );
         return 1;
       }// else depth < mV

      final int save_tvs = tvs[depth];
      if( save_tvs >= 0 )  // i.e., ==mustUse, already set
       { if( check(depth, m, mvs, sent, tvs) )
            return rec(depth+1, m,mvs,parent, sent,tvs, tUsed);
         else return 0;
       }//else save_tvs < 0, ...

      int count = 0;
      final int mv = mvs[depth];  // Vertex in m
      Series.Int ts = sent.joinedTo( tvs[parent[depth]] );
      while( ts.hasSome() )       // ts is short if G is sparse
       { final int tv = ts.elt_n(); // tv is joined to match of mv's "parent"
         if( ! tUsed.get(tv) )    // i.e., tv is free
          { tvs[depth] = tv;      // tentatively...
            boolean ok = check(depth, m, mvs, sent, tvs);
            if( ok )  // mv:tv extends the match
             { tUsed.set(tv);
               count += rec(depth+1, m,mvs,parent, sent,tvs, tUsed);       // *
               tUsed.clear(tv);
               if( count > 0 ) break;
             }//if
          }//if
         ts.advance();
       }//while ts

      tvs[depth] = save_tvs;      // restore tvs[]

      return count;
    }//rec( ... )

   public boolean check(int depth, Graph m, int [] mvs, Graph sent, int [] tvs)
    { final int mv = mvs[depth], tv = tvs[depth];
      boolean ok = true;    // so far
      for( int i = 0; ok && i <= depth; i ++ )
         ok =   (m.isEdge(mvs[i], mv) ==   // ? clean up
            sent.isEdge(tvs[i], tv))       // dir v. undir
            &&  (m.isEdge(mv, mvs[i]) ==   // later.-)
            sent.isEdge(tv, tvs[i])) ;
      return ok;
    }//check(...)

   /** Return an increasing int for the Edge (v,w) as we descend the
       i-th column and (if) we traverse the i-th row. */
   private int vw2int(int v, int w)
    { if( v <= w )  return 2*v;      // even, descending col w
      /*  v >  w */ return 2*w + 1;  // odd,  traversing row v
    }//vw2int(v,w)

    }//MotifA.M, the fully parameterised Model

 }//MotifA class, the UnParameterised Model

// (c)Lloyd Allison, May 2015, Sat 11 Dec 2017. Licence to 'MotifA' is Gnu GPL.
