//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised MultiState Model (MultiState distribution)
    on data in <NOBR><TT>bounds = [lo, hi]</TT>,</NOBR> capable of
    estimating a fully parameterised {@link MultiState.M} Model.
    Also see {@link Multinomial} and
    [<A HREF="http://www.allisons.org/ll/MML/Discrete/Multistate/"
        TARGET="LL">www</A>].  */
public class MultiState extends Discretes.Bounded                 // MultiState
 { /** The data-space bounds, as ints. */
   private final int lwb_n, upb_n;

   /** bounds = [lwb, upb], on 'this' Model's data-space. */
   public MultiState(Value bounds)
    { super(bounds);
      lwb_n = lwb_n(); upb_n = upb_n();     // see above
    }//constructor

   public Tuple bounds()                                            // bounds()
    { return defnParams().isTuple(2); }

   /** sp2Model(msg1, msg2, sp), where sp is the statistical parameters,
       that is the probabilities, return a fully parameterised
       MultiState {@link M Model}.  */
   public M sp2Model(double m1, double m2, Value sp)
    { return new M(m1, m2, (Vector) sp); }

   /** Return sufficient statistics, that is {@link #freqs frequency}
       counts, for elements [lo,&nbsp;hi) of data-set 'ds'.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { return freqs(ds, lo, hi); }

   /** For sufficient statisticses 'ss0' and 'ss1', either combine
       ss0 and ss1 (add=true), or remove ss1 from ss0 (add=false).
       Note, it is only valid to remove from ss0 the contribution
       of elements previously counted in ss0.  Also see
       {@link #stats(Vector,int,int) stats(ds,lo,hi)}.  */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { return freqs(add, ss0, ss1); }

   /** Return an Estimator for a MultiState Model; this version has
       no non-trivial parameters, but another version could --
       if it used a non-uniform prior, say. */
   public Est estimator(final Value t /* = triv */)
    { if( t != triv ) error( ".estimator(" + t + ")" );
      return new Est(t)
       { /** Given sufficient statistics,
             <TT>ss = {@link #stats stats(ds)}</TT>, of a data-set, ds,
             estimate a MultiState / Multinomial Model. The total message
             length, msg1+msg2, is always good, and msg1 and msg2 are
             always &ge;0, but for few data, msg1 may be a little high
             (where you should be grateful it works at all).  */
         public M ss2Model(Value ss)
          { final int N = upb_n - lwb_n + 1;       // width of data-space
            if( N == 1 ) // might happen? e.g., a "Mixture" of 1 class
               return new M(0, 0, Vector.doubles(new double []{1})); // !
         // else ...
            final Value freqs = (Vector) ss;       // suff stats, freqs
            double sum = 0;
            final double [] prsx = new double[N];  // MML estimate
            for( int i = 0; i < N; i ++ )
               sum += freqs.elt(i).x();            // no. of data
            final double sumPlus = sum + ((double) N)/2.0;   // !
            for( int i = 0; i < N; i ++ )                    // Estimate...
               prsx[i] = (freqs.elt(i).x() + 0.5) / sumPlus; // NB. the 0.5
            double adaptiveCode = logFactorial(sum + N - 1)
                                - logFactorial(      N - 1),
                   msg2 = 0;                        // 2nd part of msg...
            for( int i = 0; i < N; i ++ )
             { final double f_i = freqs.elt(i).x(); // i-th frequency
               msg2 -= f_i * Math.log(prsx[i]);
               adaptiveCode -= logFactorial(f_i);
             }   // msg2 is "robust", if slightly low
            double msg1 = adaptiveCode - msg2;     // 1st Model part of msg
            msg2 += informativeIncrement( N-1 );   // Informative v UnInf.
            assert msg1 >= 0 && msg2 >= 0
                 : "msg1=" + msg1 + " < 0 || msg2=" + msg2 + " < 0";
            // NB. the prior's pdf is (N-1)!, msg1=-log((N-1)!)+nlPrecision
            double nlPrecision = msg1 + logFactorial(N-1); // of prsx
         // Vector params = Vector.doubles(nlPrecision, prsx);  // ???or...
            Vector params = Vector.doubles(prsx, Math.exp(-nlPrecision/N));
            return new M(msg1, msg2, params); // precision with p of s!
          }//estimator(ds)
       };//new Estimator
    }//estimator(t)
   // ....................

   /** A fully parameterised MultiState (Multinomial) Model;
       also see the UnParameterised {@link MultiState} Model.
       The statistical {@link #statParams() parameters} is a Vector of
       probabilities, which must be (L1-) normalised. This is checked!  */
   public class M extends Discretes.Bounded.M                // MultiState.M
    { /** The probabilities, <TT>prs</TT>, are the
          {@link #statParams() statistical}-parameters of 'this'
          MultiState Model (in contrast to its {@link #bounds}). */
      private final Vector prs;

      /** The negative log probabilities (doubles); it is
          probably worth caching them as they are used a lot. */
      private final double [] nlPrs;

      /** The statistical parameters, prs, are the <I>probabilities</I>.
          It is required that <NOBR>&sum; prs[i] = 1</NOBR>.  */
      public M(double msg1, double msg2, Value prs)
       { super(msg1, msg2, prs);
         this.prs = (Vector) prs;
         nlPrs = new double[ this.prs.nElts() ];
         double sum = 0;
         for( int i = 0; i < nlPrs.length; i ++ )
          { double pr_i = this.prs.elt(i).x();
            nlPrs[i] = - Math.log(pr_i);
            sum += pr_i;
          }
         if( ! nearlyEqual(sum, 1) ) error( "sum prs[i] = " + sum + " != 1" );
       }//constructor

      /** Return <TT>{@link #prs}[dn-lwb_n()]</TT>. */
      public double pr_n(int dn)
       { assert dn >= lwb_n() && dn <= upb_n()
              : errMsg(" " + dn + " out of range");
         return prs.elt( dn - MultiState.this.lwb_n ).x();
       }//pr_n(dn)

      /** Return <TT>{@link #nlPrs}[dn-lwb_n()]</TT>. */
      public double nlPr_n(int dn)
       { assert dn >= lwb_n() && dn <= upb_n()
              : errMsg(" " + dn + " out of range");
         return nlPrs[ dn - MultiState.this.lwb_n ];
       }//nlPr_n(dn)

      /** Given sufficient statistics,
          <NOBR><TT> ss = {@link MultiState#stats stats(ds)}</TT>,</NOBR>
          of a data-set, ds, return nlLH(ss), the negative log LikeliHood
          (negative log probability) of ds.  */
      public double nlLH(Value ss)
       { final int N = MultiState.this.upb_n - MultiState.this.lwb_n + 1;
         double ans = 0;   // ss is the frequencies
         if( ss.nElts() != N ) error( " |ss|=" + ss.nElts() + "!=" + N );
         for( int i = 0; i < N; i ++ )
            ans += ss.elt(i).x() * nlPrs[i];
         return ans;
       }//nlLH(ss)

    }//MultiState.M class, the Model

 }//MultiState class, the UPModel

// (c) Lloyd Allison 2012, Sun 12 Oct 2014. Licence to 'MultiState' is Gnu GPL.
