//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import static la.maths.Maths.*;
import la.la.Value;
import mml.*;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** An UnParameterised FunctionModel, given a number of "categories"
    {@link #k 'k'}, from a number of trials {@link M.Trials#n 'n'}
    to a Vector of 'k' frequencies that sum to 'n'.
    The fully parameterised model is {@link M}.
    Also see {@link MultiState}.  */
public class Multinomial extends UPFunctionModel                 // Multinomial
 { /** The number of categories 'k'.  */
   public final int k;                                                  // k

   private final MultiState ms;     // see estimator(...) below
   private final Estimator  ms_est;

   /** Problem definition parameter {@link #k 'k'}&nbsp;&gt;&nbsp;0
       is the number of categories.  */
   public Multinomial(Value k)
    { super(k);
      this.k = ((Int) k).n();
      ms = new MultiState(pair(zero, new Int(this.k - 1)));
      ms_est = ms.estimator(triv);
    }//constructor

   /** Stats 'ss' is just the data-set ds[lo,hi).
       (One could maybe do <I>slightly</I> better.)  */
   public Value stats(Vector ds, int lo, int hi)
    { return ds.slice(lo, hi); }

   public Value stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Vector.Slice) ss1); }

   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /**  Return an Estimator of a {@link M};
        parameter 'ps' is 'triv', '(&nbsp;)'.  */
   public Est estimator(final Value ps)
    { return new Est(ps)
       { public M ss2Model(Value ss) // stats ss is the data-set ds
          { final Vector ods = ((Vector) ss).col(1);
            int nc = ods.nCols();  // must = k
            if( nc != k ) error(".ss2Model(ss), output data are not Nx" + k );
            double [] sum = new double [k];
            for( int c = 0; c < k; c ++ ) sum[c] = 0;
            for( int r = 0; r < ods.nElts(); r ++ )  // each row 'r'
               for( int c = 0; c < k; c ++ ) // each column 'c'
                  sum[c] += ods.elt(r,c).n();
            MultiState.M msm
              = (MultiState.M) ms_est.ss2Model(Vector.doubles(sum));
            return new M(msm.msg1(), msm.msg2(), msm.statParams());
          }//ss2Model(ss)
       };//new Est(ps)
    }//estimator(ps)
   // .....................................

   /** The fully parameterised FunctionModel from a number of trials
       {@link Trials#n 'n'} to a Vector of {@link #k 'k'} frequencies that
       sum to&nbsp;'n'.  M's&nbsp;{@link M#condModel(Value) condModel(id)}
       is a {@link Trials Trials(n)}.{@link Trials.TM <B>TM</B>()}.
       The UnParameterised function-model is {@link Multinomial}.  */
   public class M extends UPFunctionModel.M                    // Multinomial.M
    {
      /** The negative log probabilities of the {@link #k 'k'} categories.
          These are used by {@link Trials Trials}.{@link Trials.TM TM}.  */
      public double [] nlPrs = new double [k];                    // nlPrs[]

      /** 'prs' is the probabilities of the 'k' categories.
          'prs'&nbsp;is used to set {@link #nlPrs nlPrs}.  */
      public M(double msg1, double msg2, Value prs)
       { super(msg1, msg2, prs);
         Vector prs_v = (Vector) prs;
         double sum = 0;
         for( int i = 0; i < k; i ++ )
          { double pr_i = prs_v.elt(i).x();
            nlPrs[i] = - Math.log(pr_i);
            sum += pr_i;
          }
         if( ! nearlyEqual(sum, 1) ) error( "sum prs[i] = " + sum + " != 1" );
       }//constructor

      public double nlLH(Value ss) { return sumNlPr((Vector) ss); }

      // the default nlPr(id_od) is defined in terms of condModel ...

      /** Returns a {@link Trials Trials(n)}.{@link Trials.TM <B>TM</B>()}
          appropriate to {@link #n 'n'} trials. */
      public Trials.TM condModel(Value n)
       { return (new Trials(n)) . new TM(0, 0, triv); }

      // ......................................................................

      /** An UnParameterised Model, given {@link #n 'n'} trials over
          {@link #k 'k'} categories, of Vectors of 'k' frequencies
          f_i,&nbsp;i=0..k-1, that sum to 'n'.
          Note that 'Trials' is only a partially implemented model &ndash;
          just enough for the purposes of {@link Multinomial}.  */
      public class Trials extends UPModel               // Multinomial.M.Trials
       { /** The number of trials, 'n', of the {@link #k 'k'} categories. */
         public final int n;                                            // n

         /** {@link #n 'n'} is the number of trials.
             This is supplied by {@link #condModel}.  */
         public Trials(Value n)
          { super(n); this.n = ((Int) n).n(); }//constructor

         /** Not implemented, throw an {@link #RTE RTE}! */
         public Value stats(Vector ds, int lo, int hi)
          { throw RTE(".stats(ds,lo,hi) not yet implemented"); }

         /** Not implemented, throw an {@link #RTE RTE}! */
         public Value stats(boolean add, Value ss0, Value ss1)
          { throw RTE(".stats(add,ss0,ss1) not yet implemented"); }

         public Trials.TM sp2Model(double msg1, double msg2, Value sp)
          { return new Trials.TM(msg1, msg2, sp); }
         // ......................................

         /** A fully (trivially) parameterised Model of Vectors of
             {@link #k 'k'} frequencies that sum to {@link #n 'n'}.
             Note that 'Trials.TM' is only a partially implemented model
             &ndash; just enough for the purposes of {@link Multinomial.M}.
         <BR>(<B>Note</B>, the javac compiler objects to calling this
             class 'M' [error: class Multinomial.M is already defined],
             even though the other M is Multinomial.M and this class
             would be Multinomial.M.Trials.M &ndash;27/1/2017.)  */
         public class TM extends UPModel.M           // Multinomial.TM.Trials.M
          {
            /** Note, 'sp' is 'triv', '(&nbsp;)'.  */
            public TM(double msg1, double msg2, Value sp)
             { super(msg1, msg2, sp);
               if( sp != triv ) error( "sp != triv" );
             }//constructor

            /** Not implemented, throw an {@link #RTE RTE}! */
            public double nlLH(Value ss)
             { throw RTE(".stats(ds,lo,hi) not yet implemented"); }

            /** Negative log probability of 'f', frequencies of
                {@link #k 'k'} categories in {@link #n 'n'} trials.<BR>
                pr(f) = (n! / (&prod; f[i]!)) (&prod; pr[i]^f[i])<BR>
                The 'pr[i]' , actually the {@link #nlPrs nlPrs[i]},
                come from the {@link Multinomial.M}.
                Note, 'f'&nbsp;<EM>must</EM> be a Vector of
                'k' integers (frequencies) that sum to 'n';
                nlPr(f) throws an {@link #error error} if not.  */
            public double nlPr(Value f)
             { Vector fv = (Vector) f;
               double ans = - logFactorial(n);   int sum = 0;
               for( int i = 0; i < fv.nElts(); i ++ )
                { int f_i = fv.elt(i).n();
                  sum += f_i;
                  ans += logFactorial(f_i) + nlPrs[i] * f_i;
                }
               if( sum != n )
                  error(".nlPr(f=" + f + "), sum f[i] = " + sum + " != " + n);
               return ans;
             }//nlPr

          }//Multinomial.Trials.M class, the Model

       }//Multinomial.Trials class, the UPModel
      // ......................................................................

  }//Multinomial.M class, the Model

 }//Multinomial class, the UPFunctionModel

// (c) Lloyd Allison 28 Jan 2017. Licence to 'Multinomial' is Gnu GPL.
