//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised <I>Sequences</I> Model;
    the fully parameterised Model is {@link Sequences.M}.
    Here a Sequence is finite and is represented by a Vector.  */
public abstract class Sequences extends UPModel                    // Sequences
 {
   /** The UnParameterised Model of the <I>lengths</I> of Sequences.
       Also see {@link M#lengthMdl}.  A constructor must set it.  */
   protected Discretes lengthUPM;

   /** Problem definition parameter(s) 'dp'.  */
   public Sequences(Value dp)
    { super(dp); }//constructor

   /** Get {@link #lengthUPM}'s statistics on <I>lengths</I> out
       of 'ds[lo,hi)'. */
   public Value lengthStats(Vector ds, int lo, int hi)
    { int [] len = new int [hi - lo];
      for( int i = lo; i < hi; i ++ )
         len[i] = ((Vector) ds.elt(i)).nElts();
      return lengthUPM.stats( Vector.ints(len)
                            . weightLike(ds.slice(lo, hi)) ); // !
    }//lengthStats(ds,lo,hi)

   /** Get {@link #lengthUPM}'s stats(add,ss0,ss1) on <I>lengths</I>. */
   public Value lengthStats(boolean add, Value ss0, Value ss1)
    { return lengthUPM.stats(add, ss0, ss1); }
   // .........................................................................

   /** The fully parameterised Model of Sequences;
       {@link Sequences} is the UnParameterised Model.  */
   public abstract class M extends UPModel.M                     // Sequences.M
    {
      /** The parameterised Model of the <I>lengths</I> of Sequences.
          Also see {@link #lengthUPM}.  */
      protected Discretes.M lengthMdl;

      /** Two part message lengths, msg1 and msg2,
          and statistical parameter(s) sp.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Get {@link #lengthMdl}'s statistics on <I>lengths</I> out of
          'ds[lo,hi)'.  These may or may not be the same as
          {@link Sequences#lengthStats(Vector,int,int)}, it all depends on
          {@link #lengthUPM} versus {@link #lengthMdl}. */
      public Value lengthStats(Vector ds, int lo, int hi)
       { int [] len = new int [hi - lo];
         for( int i = lo; i < hi; i ++ )
            len[i-lo] = ((Vector) ds.elt(i)).nElts();
         return lengthMdl.stats( Vector.ints(len)
                               . weightLike( ds.slice(lo, hi) ) ); // !
       }//lengthStats(ds,lo,hi)

      /** Return the negative log probability of
          Sequence (Vector) datum&nbsp;'d'.  */
      public abstract double nlPr(Value d);                        // * nlPr(d)

    }//Sequences.M class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** Sequence.<B>K</B> uses the same Model, {@link #eltUPM} (parameterised),
       for every element of every Sequence datum.
       The parameterised Model is {@link K.M K.M}.  */
   public static class K extends Sequences                       // Sequences.K
    {
      /** The (UnParameterised) Model to be used as the
          Model of every element of every Sequence datum.
          Also see {@link M#eltMdl}.  */
      public final UPModel eltUPM;

      /** dp = &langle;{@link #lengthUPM},&nbsp;{@link #eltUPM}&rangle;.
          Also&nbsp;see the constructor for {@link K.M K.M}.  */
      public K(Value dp)
       { super(dp);
         Tuple dpT = dp.isTuple(2);
         lengthUPM = (Discretes) dpT.fst();
         eltUPM    = (UPModel)   dpT.snd();
       }//constructor

      /** Return &langle;{@link #lengthStats lengthStats}, eltStats&rangle;.
          Note, may or may not be the same as
          {@link M#stats(Vector,int,int)}. */
      public Value stats(Vector ds, int lo, int hi)
       { Value eltSS = eltUPM.stats( Vector.empty );  // !
         for(int i = lo; i < hi; i ++ )
            eltSS = eltUPM.stats( /* add= */ true,    // ! ...
                      eltSS,
                      eltUPM.stats(((Vector) ds.elt(i)).weightLike(ds.wt(i))));
         return pair( lengthStats(ds,lo,hi), eltSS );
       }//stats(ds,lo,hi)

      /** Return &langle;{@link #lengthStats lengthStats}, eltStats&rangle;.
          Note, may or may not be the same as
          {@link M#stats(boolean,Value,Value)}. */
      public Value stats(boolean add, Value ss0, Value ss1)
       { Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
         return pair( lengthUPM.stats(add,ss0T.fst(),ss1T.fst()),
                         eltUPM.stats(add,ss0T.snd(),ss1T.snd()) );
       }//stats(add,ss0,ss1)

      /** Return a fully parameterised {@link K.M K.M} Model.  */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Return the Estimator of a {@link K.M K.M}.  */
      public Est estimator(Value ps)
       { final Tuple psT = ps.isTuple(2);  // <for_lengths, for_elts>
         return new Est(ps)
          { public M ss2Model(Value ss)
             { final Tuple ssT = ss.isTuple(2);  // <length_ss, elt_ss>
               Model
                  lMdl = lengthUPM.estimator(psT.fst()).ss2Model(ssT.fst()),
                  eMdl =    eltUPM.estimator(psT.snd()).ss2Model(ssT.snd());
               return new M( lMdl.msg1() + eMdl.msg1(),
                             lMdl.msg2() + eMdl.msg2(),
                             pair(lMdl.statParams(), eMdl.statParams()) );
             }//ss2Model(ss)
          };//new Est
       }//estimator(ps)
      // ....................................

      /** K.M uses the same Model, {@link #eltMdl}, for every element of
         a Sequence.  The UnParameterised Model is&nbsp;{@link K K}.  */
      public class M extends Sequences.M                       // Sequences.K.M
       {
         /** The (parameterised) Model to be used as the
             Model of every element of the Sequence.
             Also see {@link #eltUPM}.  */
         public final Model eltMdl;

         /** sp = &langle;for_{@link #lengthMdl},
                          for_{@link #eltMdl}&rangle;. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            Tuple spT = sp.isTuple(2);
            lengthMdl = lengthUPM.sp2Model(0, 0, spT.fst());
            eltMdl    = eltUPM   .sp2Model(0, 0, spT.snd());
          }//constructor

         /** Return &langle;{@link #lengthStats lengthStats}, eltStats&rangle;.
             Note, may or may not be the same as
             {@link K#stats(Vector,int,int)}. */
         public Value stats(Vector ds, int lo, int hi)
          { Value eltSS = eltMdl.stats( Vector.empty );  // !
            for(int i = lo; i < hi; i ++ )
             eltSS = eltMdl.stats( /* add = */ true,     // !
                       eltSS,
                       eltMdl.stats(((Vector) ds.elt(i)).weightLike(ds.wt(i))));
            return pair( lengthStats(ds,lo,hi), eltSS );
          }//stats(ds,lo,hi)

         /** Return &langle;{@link #lengthStats lengthStats}, eltStats&rangle;.
             Note, may or may not be the same as
             {@link K#stats(boolean,Value,Value)}. */
         public Value stats(boolean add, Value ss0, Value ss1)
          { Tuple ss0T = ss0.isTuple(2), ss1T = ss1.isTuple(2);
            return pair( lengthMdl.stats(add,ss0T.fst(),ss1T.fst()),
                            eltMdl.stats(add,ss0T.snd(),ss1T.snd()) );
          }//stats(add,ss0,ss1)

         /** The negative log probability of Sequence (Vector)
             datum&nbsp;'d'.  */
         public double nlPr(Value d)
          { Vector v = (Vector) d;
            final int N = v.nElts();
            double ans = lengthMdl.nlPr_n(N);               // length, plus ...
            for( int i = 0; i < N; i ++ ) ans += eltMdl.nlPr(v.elt(i)); // elts
            return ans;
          }//nlPr(d)

         /** The negative log likelihood of data-set 'ds' where
             ss={@link #stats(Vector,int,int) stats(ds)}. */
         public double nlLH(Value ss)
          { Tuple ssT = ss.isTuple(2);
            double len = lengthMdl.nlLH(ssT.fst()),
                   elt = eltMdl.nlLH(ssT.snd());
            return lengthMdl.nlLH(ssT.fst()) + eltMdl.nlLH(ssT.snd());
          }//nlLH(ss)

         /** Return a random Sequence (Vector). */
         public Vector random()                                     // random()
          { return eltMdl.random( lengthMdl.random_n() ); }

       }//Sequences.K.M class, the parameterised Model

    }//Sequences.K class, the UPModel
   // -------------------------------------------------------------------------

   /** The UnParameterised Left-to-Right Model of Sequences.
       The fully parameterised model is {@link LtoR.M}.  */
   public static abstract class LtoR extends Sequences        // Sequences.LtoR
    {
      /** The UnParameterised Function Model for the "next" element.
          Also see {@link #lengthUPM}.  */
      public final UPFunctionModel nextUPFM;

      public LtoR(Value dp)
       { super(dp);
         Tuple dpT = dp.isTuple(2);
         lengthUPM = (Discretes) dpT.fst();
         nextUPFM  = (UPFunctionModel) dpT.snd();
       }//constructor

      /** The fully parameterised Left-to-Right Model of Sequenves.
          The UnParameterised Model is {@link LtoR}.  */
      public abstract class M extends Sequences.M           // Sequences.LtoR.M
       {
         /** The FunctionModel of the next element given the context.
             Also see {@link #lengthMdl}.  */
         public final FunctionModel nextFM;

         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            Tuple spT = sp.isTuple(2);
            lengthMdl = lengthUPM.sp2Model(0, 0, spT.fst());
            nextFM    = nextUPFM. sp2Model(0, 0, spT.snd());
          }//constructor

       }//Sequences.LtoR.M class, the fully parameterised Model

    }//Sequences.LtoR class, the UPModel
   // .........................................................................

 }//Sequences class, the UPModel

// (c) Lloyd Allison 3 Dec 2017. Licence to 'Sequences' is Gnu GPL.
