//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** For Models of data s : K_Simplex; a <I>standard K_Simplex</I> is the space
    of (K+1)-Vectors, {[s<SUB>0</SUB>,&nbsp;...,&nbsp;s<SUB>K</SUB>]}
    (NB.&nbsp;0..K) where <NOBR>0 &le; s<SUB>i</SUB> &le; 1</NOBR> and
    &sum;s<SUB>i</SUB>&nbsp;=&nbsp;1.
    <SMALL>(An alternative, equivalent definition is K-Vectors
    {[s[0],&nbsp;...,&nbsp;s[K-1]]}, <NOBR>0 &le; s[i] &le; 1</NOBR>
    where &sum;s[i]&nbsp;&le;&nbsp;1.)</SMALL> The present formulation
    fits conveniently with {@link MultiState.M}.<BR>
    The standard K_Simplex has K+1 vertices in <B>R</B><SUP>K+1</SUP>,
    {(1,0,...,0), (0,1,...,0), ..., (0,0,...,1)}. The length of a side,
    the distance between any two vertices, is &radic;2.
    The length of the standard 1_Simplex in <B>R</B><SUP>2</SUP>
    is &radic;2.
    The area   of the standard 2_Simplex in <B>R</B><SUP>3</SUP>
    is (&radic;3)/2.
    The volume of the standard 3_Simplex in <B>R</B><SUP>4</SUP>
    is 1/3 = (&radic;4)/3!.
    The K-volume of the standard K_Simplex in <B>R</B><SUP>K+1</SUP>
    is (&radic;(K+1))/K!. &nbsp; --
    [<A HREF="https://en.wikipedia.org/wiki/Simplex" TARGET="EX">wikip</A>]<BR>
    Also see {@link Direction Direction} for another
    K-dimensional subspace of <B>R</B><SUP>K+1</SUP>.  */
public abstract class Simplex extends R_D                          // * Simplex
 { public Simplex(Value dp)
    { super(dp); }//constructor

   /** D, the dimension of [0,&nbsp;1]<SUP>D</SUP> which contains
       the K-Simplex; D&nbsp;=&nbsp;{@link #K() K()}+1.
       Also see {@link Direction#D()}. */
   public int D() { return K() + 1; };                                   // D()

   /** K, the dimension (degrees of freedom) of the data-space
       which is a sub-space of [0,&nbsp;1]<SUP>K+1</SUP>.
       K&nbsp;=&nbsp;{@link #D() D}-1.
       Also see {@link Direction#K()}. */
   public abstract int K();                                            // * K()

   /* Note that we do not know how the constructor will learn K's value */
   private double logArea = 101;  // cache, see logArea() just below

   /** The area of the standard K-{@link Simplex Simplex} is
       (&radic;(K+1))/K!, so log of that.  E.g.,
         (i)&nbsp;Bool    ~ 1-Simplex, vol=&radic;2;
        (ii)&nbsp;{R|G|B} ~ 2-Simplex, vol=(&radic;3)/2!;
       (iii)&nbsp;DNA     ~ 3-Simplex, vol=1/3 = (&radic;(K+1))/K!;
       ...&nbsp;etc..  */
   public double logArea()
    { if( /* cache, above */ logArea > 100 ) // biggest is sqrt(2) for K=1
         logArea = 0.5 * Math.log(D()) - logFactorial(K());
      return logArea;
    }//logArea()

   /** Perform a validity check on datum&nbsp;v &mdash;
       that it is L<SUB>1</SUB>-normalised. */
   public void check(Value v)
    { final Vector vv = (Vector) v;  final int N = vv.nElts();
      if( N != D() ) error( " v is " + N + "D, should be " + D() + "D" );
      double sum = 0;
      for( int i = 0; i < N; i ++ )
       { final double vi = vv.elt(i).x();
         if( vi < 0 || vi > 1 ) error( " v[" + i + "]=" + vi );
         sum += vi;
       }//for i
      if( Math.abs(sum - 1) > epsilon ) error( " sum v[i] =" + sum );
      return; // OK
    }//check(v)
   // ..................................

   /** The UnParameterised Uniform Model over a K_Simplex data-space
       ({@link #Mdl Mdl} is fully parameterised).
       A Uniform Simplex Model is a plausible <I>prior</I> for
       {@link MultiState.M} (<B>note</B> that a MultiState
       over a data-space of 'n' possible Values itself has n-1 parameters
       (degrees of freedom) which lie in an (n-1)_Simplex).
       Also see {@link Direction.Uniform}.  */
   public static class Uniform extends Simplex               // Simplex.Uniform
    { private final int k; // see K()

      /** Uniform K-simplex in [0,&nbsp;1]<SUP>D</SUP>, where D=K+1. */
      public Uniform(Value K /*dimension*/)
       { super(K);  // K is the UPModel's one defn param
         this.k = ((Value.Int) K) .n();
       }//constructor

      /** The dimension (degrees of freedom) of the sub-space
          (in [0,&nbsp;1]<SUP>K+1</SUP>).
          Also see {@link Direction.Uniform#D()}.  */
      public int K() { return k; }

      /** Return a "fully parameterised" Uniform Simplex Model;
          the statistical parameter, t, is trivial. */
      public M sp2Model(double m1, double m2, Value t)
       { return new M(m1, m2, t); }

      /** The default sufficient statistics ss&nbsp;=&nbsp;stats(ds)
          of a data-set ds; ss&nbsp;=&nbsp;ds itself.
          See {@link Uniform.M#nlLH}.
          More on stats {@link UPModel#stats here}. */
      public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo,hi); }

      public Vector stats(boolean add, Value ss0, Value ss1)
       { return ((Vector) ss0).combine(add, (Slice) ss1); }

      /** The trivial Estimator that is a Uniform's. */
      public Estimator estimator(final Value t /* = triv */)
       { return Mdl.asEstimator(t); }
      // ............................

      /** The "given" fully parameterised Simplex.Uniform Model. */
      public M Mdl = new M(0, 0, triv);

      /** {@link #Mdl} should be sufficient for most purposes, but here is
          M, the class of fully parameterised Uniform Simplex Model(s). */
      public class M extends R_D.M                         // Simplex.Uniform.M
       { /** NB. msg1=0 and sp=triv. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

         /** The negative log probability <I>density</I> of datum d
             (L<SUB>1</SUB>-normalised which is {@link #check checked}(!))
             under the Uniform model is +{@link #logArea() logArea()}.  */
         public double nlPdf(Value d)                               // nlPdf(d)
          { check(d);          // dimension?, etc.
            return logArea();  // negative log pdf
          }//nlPdf(v)

         /** Assumes that {@link Uniform#stats(Vector,int,int) statistics}
             are the data-set itself. */
         public double nlLH(Value ss)
          { return sumNlPr((Vector) ss); }

         /** Put in array 'x' the components of a Vector uniformly at
             random in the K-Simplex.  Called by {@link #random() random()}.
             Use random(x) if you can reuse the array&nbsp;x. */
         public void random(double [] x)                           // random(x)
          { final int D = x.length;  double sum = 0;
            for( int i = 0; i < D; i ++ )
             { final double r =                     // r in [0, infin)
                  - Math.log(1 - RNG.nextDouble()); // see 'Exponential'
                          // ^^^^^^^ in (0, 1]
               x[i] = r;
               sum += r;
             }
            for( int i = 0; i < D; i ++ ) x[i] /= sum;
          }//random()

       }//Simplex.Uniform.M class, the Model

    }//Simplex.Uniform class, the UPModel

 }//Simplex (abstract) class, the UPModel

// (c) Lloyd Allison, 2012, Wed 28 Dec 2013. Licence to 'Simplex' is Gnu GPL.
