//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Type;
import la.la.Value;
import la.maths.Vector;
import la.maths.Vector.Slice;
import mml.MML;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of UnParameterised (Decision | Classification | Regression)-
    Tree FunctionModels.  Its problem defining parameter is
    ({@link #iType iType},&nbsp;{@link #leafUPM leafUPM}) - which
    is quite general. The estimator is {@link Tree.Est}.
    The fully parameterised FunctionModels
    are&nbsp;{@link M} (abstract), {@link Tree.Fork} (abstract),
    {@link Tree.Leaf}, {@link Tree.DFork} and {@link Tree.OFork}. */
public class Tree extends UPFunctionModel                               // Tree
 {
   /** log<SUB>e</SUB>(2), i.e., one bit. */
   public final static double log2 = Math.log( 2.0 );


   /** The {@link Type} of the {@link Param} - statistical parameter -
       of a fully parameterised {@link M} FunctionModel. */
   public final static Type.Option TreeParam =
      new Type.Option( "TreeParam",
                       new String [] {"Leaf", "DFork", "OFork"}, // ids
                       new int    [] {    1,       2,       3 }  // arities
                     );

   /** The root class for the <I>statistical parameter</I>
       of a fully parameterised {@link M Tree} FunctionModel.
       Also see {@link Param.Leaf} and {@link Param.Fork}.  */
   public abstract static class Param extends Value.Option
    { /** Returns {@link #TreeParam}. */
      public Type.Option type() { return TreeParam; }

      /** The statistical parameter of a {@link Tree.Leaf}, i.e.,
          a statistical parameter to make a {@link Tree.Leaf}
          from {@link Tree#leafUPM}. */
      public static class Leaf extends Param
       { private final Value sp;
         public Leaf(Value sp) { this.sp = sp; }
         public int n() { return 0; }
         /** Return {@link #sp sp}, providing i==0. */
         public Value elt(int i)
          { if( i == 0 ) return sp; else throw RTE( ".elt(" + i + ")" ); }
       }//Leaf class

      /** The superclass of {@link Param.DFork} and {@link Param.OFork}. */
      public static abstract class Fork extends Param
       { /** The number of the <I>Discrete</I> bounded input- (dependent-)
             column (variable) on which to split the data. */
         private final int col;
         /** The parameters of the sub-Tree Models. */
         private final Param [] subTrees;
         public Fork(int col, Param [] subTrees)
          { this.col = col; this.subTrees = subTrees; }
      // public abstract int n();  from Value.Option
         public int         col() { return col; }
         public Param []  subTs() { return subTrees; }
         public Param subT(int i) { return subTrees[i]; }
       }//Param.Fork abstract class

      /** The statistical parameter of a {@link Tree.DFork}, i.e., an
          input-column number, and a Vector of parameters for sub-Trees.
          Also see {@link Param.OFork}.  */
      public static class DFork extends Fork
       { public DFork(int col, Param [] subTrees)
          { super(col, subTrees); }
         public int n() { return 1; }
         /** Return {@link #col col} if i==0, {@link #subTs} if i==1. */
         public Value elt(int i)
          { switch( i )
             { case 0: return new Int(col());
               case 1: return Vector.values(subTs());
             }
            throw RTE( ".elt(" + i + ")" );
          }
       }//Param.DFork class

      /** The statistical parameter of a {@link Tree.OFork}, i.e., an
          input-column number, a splitting Value <NOBR>(&lt; v. &ge;)</NOBR>,
          and parameters for <I>two</I> sub-Trees.
          Also see {@link Param.DFork}.  */
      public static class OFork extends Fork
       { private final Value split;
         public OFork(int col, Value split, Param [] subTrees)
          { super(col, subTrees); this.split = split;
            if( subTrees.length != 2 )
               error( ", "+subTrees.length+" subTrees" );
          }
         public int       n() { return 2; }
         public Value split() { return split; }
         /** Return {@link #col col} if i==0, {@link #split split}
             if i==1, or {@link #subTs} if i==2. */
         public Value elt(int i)
          { switch( i )
             { case 0: return new Int(col());
               case 1: return split();
               case 2: return Vector.values(subTs());
             }
            throw RTE( ".elt(" + i + ")" );
          }
       }//Param.OFork class

    }//Param abstract class
   // .........................................................................

   /** The {@link Type} of the input datum (variable), id, in id&rarr;od. */
   public final Type iType;

   /** The UnParameterised Model of the output datum, od, that is to be
       {@link Tree.Leaf#mdl parameterised} in each {@link Tree.Leaf Leaf}
       of the fully {@link M parameterised} Tree.  In general the instances
       will have different statistical parameters.
       Note that leafUPM can be a simple {@link UPModel UPModel} --
       {@link Discretes Discretes}, {@link Continuous Continuous},
       {@link Multivariate Multivariate}, etc. -- or even a
       {@link UPFunctionModel UPFunctionModel} (regression), say. */
   public final UPModel leafUPM;

   /** Problem defining parameter dp is
       ({@link #iType iType},&nbsp;{@link #leafUPM leafUPM}).
       UPModel leafUPM is to be parameterised in each
       {@link Tree.Leaf Leaf} of a fully parameterised Tree. */
   public Tree(Value dp)
    { super(dp);
      Tuple dpT = dp.isTuple(2);
      iType   = (Type)    dpT.fst();
      leafUPM = (UPModel) dpT.snd();
    }//constructor

   /** Return a fully parameterised Tree FunctionModel. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { Param spN = (Param) sp;
      switch( spN.n() )
       { case 0: return new  Leaf(msg1, msg2, sp);
         case 1: return new DFork(msg1, msg2, sp);
         case 2: return new OFork(msg1, msg2, sp);
       }
      throw RTE( ", sp = " + sp + "." );
    }//sp2Model(,,)

   /** Statistics ss&nbsp;=&nbsp;stats(ds), of data-set ds, is ds itself. */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo,hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return an {@link Est Estimator} for a Tree FunctionModel,
       M:id&rarr;od.  Parameter ps is passed to
       {@link #leafUPM leafUPM}.estimator(ps). */
   public Est estimator(final Value ps)                        // estimator(ps)
    { return new Est(ps); }
   // .........................................................................

   /** The class of Estimators for a Tree FunctionModel.  It is public
       so that it can be extended, e.g., to allow for new kinds of
       {@link Tree.Fork} or changes to the {@link #search search}, etc.. */
   public class Est extends UPFunctionModel.Est                     // Tree.Est
    {
      /** leafEst estimates a Model of the output datum, col=1.
          Called by {@link #estLeaf estLeaf}. */
      public final Estimator leafEst;

      /** Note, ps is passed to {@link #leafUPM leafUPM}.estimator(ps)
          to make {@link #leafEst leafEst}. */
      public Est(Value ps)
       { super(ps);
         leafEst = leafUPM.estimator(ps);
       }//constructor

      /** Given statistics {@link #stats ss}&nbsp;(=&nbsp;ds) of a
          data-set, ds, estimate a fully parameterised M:id&rarr;od;
          calls {@link #search search(ss,&nbsp;1)}.
          You can call the latter directly to vary the lookahead. */
      public M ss2Model(Value ss)
       { return search((Vector) ss, /*lookahead*/ 1); }

      /** Given a data-set, ds=(id,od)<SUP>*</SUP>, estimate a fully
          parameterised M:id&rarr;od; calls {@link #recSearch recSearch}.
          If lookahead=0, potential Fork choices will be compared initially
          on the basis that each one is a Fork of Leafs. Increasing lookahead
          &gt;0 may give a better result, but will take more time.
          Called by {@link #ss2Model ss2Model(ss)}. */
      public M search(Vector ds, int lookahead)
       { if( iType instanceof Type.Atomic ) // id is univariate
            return recSearch(1, null, ds, lookahead);
      // else, id:iType is Multivariate ...
         Vector id = ds.col(0);             // the input col(s)
         int nC = id.nCols(), nF = 0;       // # of input, and forkable, cols
         boolean [] f = new boolean[nC];    // flag the fork-able cols
         for( int cN = 0; cN < nC; cN ++ )
          { f[cN] = false;
            Type colT = ((Type.Tuple.GP) iType) .elt(cN);
            if( ! ( colT instanceof Type.Atomic ) ) continue; // else...
            f[cN] = true;  nF ++ ;          // can fork on id column 'cN'
          }//for cN; now let us search...
         return recSearch(nF, f, ds, lookahead);
       }//search(ds,la)

      /** Coordinate the recursive search for a good (best?) Tree Model.
          There remain 'nF' free Fork-able input columns (as indicated
          in&nbsp;f, if f&ne;null).  Calls {@link #estDF estDF},
          {@link #estOF estOF}, and {@link #searchMV searchMV}.
          Called by {@link #search search}. */
      public M recSearch(int nF, boolean [] f, Vector ds, int lookahead)
       {
         M leaf = estLeaf(ds);  // a Leaf is always an option
         if( nF <= 0 || lookahead < 0 || ds.nElts() < 2 || ds.wts() < 2 )
            return leaf;        // I'm all partitioned out
         M tree;                // otherwise, can we do better? ...
         if( f == null )        // input datum, id, is Atomic, univariate
          { Vector id = ds.col(0);
            if( iType instanceof Type.Discrete
            && ((Type.Discrete) iType) .bounded() )
               tree = estDF(nF, null, id, -1, sel0, ds, lookahead);
            else // Ordered Fork
               tree = estOF(nF, null, id, -1, sel0, ds, lookahead);
          }
         else // id is Multivariate
            tree = searchMV(nF, f, ds, lookahead, leaf);
         return leaf.msg() <= tree.msg() ? leaf : tree;
       }//recSearch(...)

      /** Search for a column of a Multivariate input datum on which to split.
          If there are multiple candidates, try each one in turn, using
          lookahead-1, to find the likely best one.  Calls
          {@link #sMV_cN sMV_cN}. Called by {@link #searchMV searchMV}. */
      public M searchMV(int nF, boolean [] f, Vector ds, int lookahead, M leaf)
       { assert nF >= 1 : " nF = " + nF + " < 1";
         int N = ds.nElts();  assert N >= 2 : " N = " + N + " < 2";
         int bestCn = -1;            // i.e., L, no column
         double bestML = Double.MAX_VALUE/100;
         Vector id = ds.col(0);      // the input data cols
         int nC = id.nCols();        // # of input data cols
         // search for the best column, bestCn, on which to Fork...
         for( int cN = 0; cN < nC; cN ++ )
          { if( ! f[cN] ) continue;  // cannot (now) fork on cN
            // All ok for column 'cN'...
            if( nF <= 1 )            // one choice, just take it and
             { bestCn = cN; break; } // go straight to 'lookahead'
            M F = sMV_cN(nF, f, id.col(cN), cN, sel(cN), ds, lookahead-1);
            if( F.msg() < bestML )
             { bestCn = cN; bestML = F.msg(); }
          }//for cN
         if( bestCn < 0 ) return leaf; // ???
         return sMV_cN(nF, f, id.col(bestCn), bestCn, sel(bestCn),
                                                   ds, lookahead );
       }//searchMV(...)

      /** Column cN was selected by {@link #searchMV searcMV},
          now which kind of Fork? Calls {@link #estDF estDF} or
          {@link #estOF estOF}. */
      M sMV_cN(int nF, boolean [] f,
               Vector id, int cN, Sel sel,
               Vector ds, int lookahead)
       {
         final Type idT = ((Type.Tuple.GP) iType) .elt(cN); // Type of id
         if( idT instanceof Type.Discrete
         &&  ((Type.Discrete) idT). bounded() )
            return estDF(nF, f, id, cN, sel, ds, lookahead);
         else
            return estOF(nF, f, id, cN, sel, ds, lookahead);
       }//sMV_cN(...)


      /** Class of variable selector of the input datum id,
          where iod=(id,od). */
      public class Sel
       { public Value from(Value iod) { return iod.elt(0); } }

      /** Selects <I>all</I> of the input datum id, where iod=(id,od). */
      public Sel sel0 = new Sel();

      /** Selector of input column (variable) 'col' of the input datum id,
          where iod=(id,od). */
      public Sel sel(final int col)
       { return new Sel()
          { public Value from(Value iod) { return iod.elt(0).elt(col); }
       }; }


      /** Estimate a DFork on the Discrete Bounded column 'col',
          numbered 'cN', of the input (indep.) datum.
          Calls {@link #recSearch recSearch}. */
      public DFork estDF(int nF, boolean [] f,
                  Vector col, final int cN, final Sel sel,
                  Vector ds, int lookahead)
       { if( cN >= 0 ) f[cN] = false; // Multivariate, no re-use of cols
         final Type.Discrete colT = (Type.Discrete) col.eltType();
         final int N = ds.nElts();
         Vector dsSorted = ds.sorted(
            new java.util.Comparator <Value> ()
             { public int compare(Value e1, Value e2)
                { return sel.from(e1).compareTo( sel.from(e2) ); } } );
         double msg1 = log2 /*fork*/ + Math.log(nF) /*col# */,
                msg2 = 0;
         int p1 = 0, p2 = 0;
         Param [] subT = new Param[colT.upb_n()-colT.lwb_n()+1];
         for( int n = colT.lwb_n(); n <= colT.upb_n(); n ++ ) //each val
          { while( p2 < N
                && sel.from(dsSorted.elt(p2)).n() == n )
               p2 ++;
            // NB. id univariate => nF=1 => est returns a Leaf...
            M subM = recSearch(nF-1, f, dsSorted.slice(p1, p2), lookahead);
            msg1 += subM.msg1();  // NB. subM is a Leaf if ...
            msg2 += subM.msg2();  //     ... lookahead < 0
            subT[n-colT.lwb_n()] = (Param) subM.statParams();
            p1 = p2;
          }//for n
         if( cN >= 0 ) f[cN] = true; // reset
         return new DFork(msg1, msg2, new Param.DFork(cN, subT));
        }//estDF(...)


      /** Estimate an OFork on the input column 'col', numbered 'cN'.
          Specifying the median as the splitting Value would cost 1-bit,
          a quartile 3-bits, an octile 5 bits, and so on, thus effecting
          variable precision (Wallace &amp; Patrick 1993).
          Calls {@link #recSearch recSearch}.  */
      public M estOF(final int nF, final boolean [] f,
               Vector col, final int cN, final Sel sel,
               Vector ds, final int lookahead)
       { final Type colT = col.eltType();
         final int N = ds.nElts();    // NB., N >= 2
         assert N >= 2 : "N=" + N + " which is < 2.";
         final java.util.Comparator <Value> cmprtr =
            new java.util.Comparator <Value> ()
             { public int compare(Value e1, Value e2)  // NB. uses 'sel'
                { return sel.from(e1).compareTo( sel.from(e2) ); }
             };
         final Vector dsSorted = ds.sorted(cmprtr);  // sorted on column 'cN'

         /** Infer the best split point, {@link DandC#split split}, in
             [lo,hi) that minimises the two-part message length. */
         class DandC  // Did you know this was possible in Java?
          { /** The sorted data is split on the split<SUP>th</SUP>
                element into [lo,split) and [split,hi). */
            int    split;
            /** {@link #NodeCost} is the cost of just the top OFork node;
                {@link #msg} is the total for the Tree and data. */
            double NodeCost = Double.MAX_VALUE/10, // depends on split, only
                   msg      = Double.MAX_VALUE/10; // depends on lookahead

            /** Find the best split point, s, in [lo,hi). */
            public DandC(int lo, int hi, int bits)        // NB. [lo,hi)
             { if( hi < lo + 2 ) return;  // at most one elt
               int l = lo-1, h = hi, lWt = 0, hWt = 0;
               while( true )       // find median of +ve weighted data...
                { if( lWt <= hWt ) // NB. <=
                   { l++; if( dsSorted.wt(l) > 0 ) lWt++; }
                  else /* >  */
                   { h--; if( dsSorted.wt(h) > 0 ) hWt++; }
                  if( l >= h-1 ) break;
                }
               split = h;
               if( split >= hi ) return;  // poss. with weird weighting?
               if( cmprtr.compare(dsSorted.elt(split-1),
                                  dsSorted.elt(split)) < 0 )
                { // 'split' is a valid cut-point, evaluate it.
                  // if constWt() > 0, is equiv. to ...
                  // split = (lo+hi)/2; // median [0,2)->1, [0,3)->1, [0,4)->2
                  // hi > lo+1, so hi >= lo+2, so hi > split > lo
                  Vector srtd1 = dsSorted.slice(0, split),   // [0, split)
                         srtd2 = dsSorted.slice(split, N);   // [split, N)
                  // NB., m1 and/or m2 may re-test col
                  //      (tho return Leaf if LA < 0)
                  M m1 = recSearch(nF, f, srtd1, lookahead-1),  // NB. '-1'
                    m2 = recSearch(nF, f, srtd2, lookahead-1);
                  NodeCost = log2         // a Fork
                           + Math.log(nF) // col number
                           + bits / log2; // split Value (nits) ...
                           // ... median 1-bit, a quartile 3-bits, etc..
                  double msg1 = NodeCost + m1.msg1() + m2.msg1(),
                         msg2 = m1.msg2() + m2.msg2();
                  msg = msg1 + msg2;
                }//if
               // else 'split' was not valid. In any case...
               // Now Divide and Conquer, try quartiles...
               DandC qtl = new DandC(lo, split, bits+2);  // NB. bits+2
               if( qtl.msg < msg ) // left better
                { split=qtl.split; NodeCost=qtl.NodeCost; msg=qtl.msg; }
               qtl = new DandC(split, hi, bits+2);        // NB. bits+2
               if( qtl.msg < msg ) // right better
                { split=qtl.split; NodeCost=qtl.NodeCost; msg=qtl.msg; }
             }//DandC constructor
          }//DandC class

         // The search tries: median, quartiles, octiles, ...
         DandC dc = new DandC(0, N, 1 /*the median costs 1 bit*/);

         // Commit to dc.split; search lft and rt with full lookahead...
         M m1 = recSearch(nF, f, dsSorted.slice(0, dc.split), lookahead),
           m2 = recSearch(nF, f, dsSorted.slice(dc.split, N), lookahead);
         return new OFork(
         /* msg1 */ dc.NodeCost + m1.msg1() + m2.msg1(),
         /* msg2 */ m1.msg2() + m2.msg2(),
         /* s.p. */ new Param.OFork(cN, sel.from(dsSorted.elt(dc.split)),
                       new Param[]{(Param)m1.statParams(),
                                   (Param)m2.statParams()}) );
       }//estOF(...)

      /** Estimate a Leaf, i.e., Model of the output (dependent) column
          ignoring the input (indep.) column(s).  Note, for a k-way
          tree <NOBR>#leaves = (k-1) &times; #forks + 1</NOBR>. */
      public Leaf estLeaf(Vector ds)
       { Model odMdl = leafEst.apply(ds.col(1)); // of the output datum
         return new Leaf( log2/*leaf*/ + odMdl.msg1(), odMdl.msg2(),
                          new Param.Leaf(odMdl.statParams()) );
       }//estLeaf(ds)

    }//Tree.Est class
   // .........................................................................

   /** A fully {@link Param.DFork parameterised} Tree FunctionModel that
       tests a <I>Discrete</I> Bounded {@link #col() column} (variable)
       of the input datum to select a {@link #subTrees sub-Model}
       into which to descend. */
   public class DFork extends Fork                                     // DFork
    { /** The number of the input column (variable) to be
          tested; <NOBR>NB. '-1' means 'all'</NOBR> of id. */
      public final int col;

      /** The sub-Models, one for each possible Value of
          the (Discrete) input {@link #col() column}. */
      public final M [] subTrees;

      /** sp = (col, [sp<SUB>0</SUB>, ..., sp<SUB>n-1</SUB>]) where col is
          {@link #col() column} number and sp<SUB>i</SUB> is the statistical
          parameter(s) of {@link #subTrees subTrees<SUB>i</SUB>}.  */
      public DFork(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Param.DFork spF = (Param.DFork) sp;
         col      = spF.col();
         subTrees = new M[ spF.subTs().length ];
         for( int i = 0; i < subTrees.length; i ++ )
            subTrees[i] = sp2Model(0, 0, spF.subT(i));
       }//constructor

      public int col() { return col; }

      /** Use column (variable) {@link #col() col} of the input
          datum to choose one of {@link #subTrees subTrees}.
          Also see {@link Tree.OFork#condModel}.  */
      public Model condModel(Value id)
       { final Value f = field(id);
         final Type.Discrete t = (Type.Discrete) f.type();
         return subTrees[ f.n() - t.lwb_n() ].condModel(id);
       }//condModel(id)

    }//DFork class, fully parameterised FunctionModel
   // .........................................................................

   /** A fully {@link Param.OFork parameterised} Tree FunctionModel that
       compares an ordered {@link #col() field} (variable) of the input
       datum to the {@link #split split} to select a
       {@link #subTrees sub-Model} into which to descend. */
   public class OFork extends Fork                                     // OFork
    { /** The number of the input column (variable) to be
          tested; <NOBR>NB. '-1' means 'all'</NOBR> of it. */
      public final int col;

      /** Select subTree '0', or '1', as the input datum
          is '&lt;' or '&ge;' split respectively.  */
      public final Value split;

      /** The <I>two</I> sub-Models for &lt;{@link #split split}
          or &ge;split respectively. */
      public final M [] subTrees;

      /** sp = (col, split, [sp<SUB>0</SUB>, sp<SUB>1</SUB>]). */
      public OFork(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Param.OFork spF = (Param.OFork) sp;
         col      = spF.col();
         split    = spF.split();
         subTrees = new M[ 2 ];      // NB. 2
         if( spF.subTs().length != 2 )
            error( ", " + spF.subTs().length + " subModels" );
         for( int i = 0; i < subTrees.length; i ++ )
            subTrees[i] = sp2Model(0, 0, spF.subT(i));
       }//constructor

      public int col() { return col; }

      /** Use column (variable) {@link #col() col} of the input datum, &lt;
          v.&nbsp;&ge; {@link #split split}, to choose one of the two
          {@link #subTrees subTrees}.
          Also see {@link Tree.DFork#condModel}. */
      public Model condModel(Value id)
       { return subTrees[ field(id).compareTo(split) < 0 ? 0 : 1 ]
                       .condModel(id);
       }//condModel(id)

    }//OFork class, fully parameterised FunctionModel
   // .........................................................................

   /** Superclass of fully {@link Param.Fork parameterised}
       non-{@link Tree.Leaf Leaf} FunctionModels;
       see {@link Tree.DFork} and {@link Tree.OFork}. */
   public abstract class Fork extends M                               // * Fork
    {
      public Fork(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Assumes {@link #stats(Vector,int,int) statistics} 'ss'
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }//nlLH(ss)

   // public abstract Model condModel(Value id);

      /** Select column ({@link #col() col()}&ge;0), or
          all (col()&lt;0), of the input datum. */
      public Value field(Value id)
       { return col() < 0 ? id : id.elt(col()); }

      /** The number of the column of the input that is to be tested.
          <NOBR>Note, '-1' means 'all'</NOBR> of it.  */
      public abstract int col();

    }//Fork abstract class, fully parameterised FunctionModel
   // .........................................................................

   /** A fully {@link Param.Leaf parameterised} Tree FunctionModel consisting
       of a single Leaf; it contains a {@link #mdl Model} over the output
       (dependent) datum, 'od'. mdl&nbsp;is a parameteristaion of
       {@link #leafUPM}.  Also see {@link Tree.Fork}.  */
   public class Leaf extends M                                          // Leaf
    { /** 'mdl', the Model of the output datum, 'od'.
          mdl&nbsp;is {@link #leafUPM leafUPM}.apply(sp.fst()).
          Note that mdl can be a simple {@link Model} or a
          {@link FunctionModel}, say. */
      public final Model mdl;

      /** {@link #leafUPM leafUPM} {@link Param.Leaf parameterised}
          by 'sp' gives {@link #mdl mdl}. */
      public Leaf(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         mdl = leafUPM.apply(((Param.Leaf) sp) .fst());
       }//constructor

      /** Assumes {@link #stats(Vector,int,int) statistics} 'ss'
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }//nlLH(ss)

      /** Given input datum 'id', always return {@link #mdl mdl} of od. */
      public Model condModel(Value id)
       { return mdl; }  // NB. could pass id to a *FunctionModel* mdl   ?? TODO

    }//Leaf class, fully parameterised FunctionModel
   // .........................................................................

   /** The superclass of fully {@link Param parameterised} (Decision |
       Classification | Regression) -Trees; see {@link Tree.Leaf},
       {@link Tree.DFork} and {@link Tree.OFork}.
       The UnParameterised Tree is {@link Tree here}. */
   public abstract class M extends UPFunctionModel.M                  // Tree.M
    {
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
       }//constructor

   // public abstract Model condModel(id),  from FunctionModel

    }//M abstract class, fully parameterised FunctionModel
   // -------------------------------------------------------------------------

   /** main() allows Tree to be (slightly) tested, in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Tree --");

      Value.Bool  tt = Value.ttrue, ff = Value.ffalse;
      Value.Tuple ff_tt = Value.pair(ff, tt);
      Value.Int zero = Value.zero, one = Value.one;

      Value.Cts  pt1 = Value.real(0.1), pt5 = Value.real(0.5),
                 pt9 = Value.real(0.9);
      Vector odds1_9 = Vector.values(new Value[]{pt1, pt9}),   // 1:9
             odds5_5 = Vector.values(new Value[]{pt5, pt5}),   // 5:5
             odds9_1 = Vector.values(new Value[]{pt9, pt1});   // 9:1

      MultiState boolUPM =  new MultiState(ff_tt);

      // cook up some data-sets...
      Type B2 = new Type.Tuple.GP(null, new Type[]{Type.BOOL, Type.BOOL});
      Vector r0 = Vector.doubles(new double[]{0.1, 1.0,-0.1, 0.9, 1.1});
      Vector c0 = Vector.values( new Value[] {ff,  tt,  ff,  tt,  tt }),
             c1 = Vector.values( new Value[] {ff,  ff,  tt,  tt,  tt }),
             c2 = Vector.values( new Value[] {ff,  ff,  ff,  tt,  ff });
      double wt = 10;
      Vector ds0 = (c0 .zip(c1)) .zip(c2),              // ((c0, c1), c2)
             ds1 = ds0. weight(wt),
             ds2 = (r0 .zip(c1)) .zip(c2) .weight(wt);  // ((r0, r1), c2)
      Type CB = new Type.Tuple.GP(null, new Type[]{Type.CTS, Type.BOOL});

      Tree upT = new Tree( Value.pair(B2, boolUPM) );
      Value sp =
        new Param.DFork(0, new Param[]{
              new Param.Leaf(odds9_1),
              new Param.DFork(1, new Param[]{
                    new Param.Leaf(odds9_1),
                    new Param.Leaf(odds5_5) }) });

      M tree = (M) upT.apply( sp );  // a given Tree
      System.out.println( "   tree = " + tree );
      Test.exercise(tree, ds0);
      Test.exercise(upT.estimator(triv), ds1);   // ds1 : [((B, B), B)]

      Tree upT2 = new Tree( Value.pair(CB, boolUPM) );
      Test.exercise(upT2.estimator(triv), ds2);  // ds2 : [((Cts, B), B)]

      Test.exercise(
        new Tree(Value.pair(Type.CTS, boolUPM)).estimator(triv),
        r0.zip(c2).weight(wt) );                // [(Cts, B)]

      System.out.println("-- done Tree --");
    }//main

 }//Tree class, the UnParameterised FunctionModel

// (c) Lloyd Allison, Jan 2014, Thu 3 July 2014. Licence to 'Tree' is Gnu GPL.
