//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Series;
import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The (abstract) class of UnParameterised Models of Series.
    Fully parameterised Models of Series are
    {@link UPSeriesModel.M} and {@link SeriesModel}.  Also see
    {@link UPModel UPModel} and {@link UPFunctionModel UPFunctionModel}.  */
public abstract class UPSeriesModel extends UPModel          // * UPSeriesModel
 {
   /** Given problem-defining parameter(s) 'dp', construct a SeriesModel. */
   public UPSeriesModel(Value dp)
    { super(dp); }//constructor

   /** Return the fully parameterised SeriesModel with
       statistical parameters sp.  */
   public SeriesModel apply(Value sp) { return sp2Model(0, 0, sp); }

   /** Two part message lengths, msg1 and msg2, and
       statistical parameters, sp, return a fully parameterised
       {@link UPSeriesModel.M}.  */
   public abstract SeriesModel sp2Model(double msg1, double msg2, Value sp);

   /** A class of Estimator of <I>SeriesModel</I>, that uses
       information from its enclosing {@link UPSeriesModel}.  */
   public abstract class Est extends UPModel.Est         // * UPSeriesModel.Est
    { /** Parameter(s) 'ps' may, for example, control a <I>prior</I>
          on a Model's statistical parameters.  */
      public Est(Value ps)
       { super(ps); }//constructor
      /** Given a data-set, ds, return a fully parameterised SeriesModel. */
      public SeriesModel ds2Model(Value ds)                     // ds2Model(ds)
       { return ss2Model(stats((Vector) ds)); }
      /** Synonym for {@link #ds2Model ds2Model(ds)}. */
      public final SeriesModel ds2SeriesModel(Value ds)
       { return ds2Model(ds); }
      /** Given statistics <SPAN STYLE="white-space:nowrap">
          ss = {@link #stats stats(ds)}</SPAN> of a data-set, ds,
          return a fully parameterised SeriesModel.  */
      public abstract SeriesModel ss2Model(Value ss);         // * ss2Model(ss)
      /** Synonym for {@link #ss2Model ss2Model(ds)}. */
      public final SeriesModel ss2SeriesModel(Value ss)
       { return ss2Model(ss); }
    }//Est class
   // .........................................................................

   /** The (abstract) fully parameterised Series Model.
       The UnParameterised Series Model is {@link UPSeriesModel}.  */
   public abstract class M extends SeriesModel             // * UPSeriesModel.M
    { /** Two-part message length, msg1 and msg2, and statistical
          parameter(s), construct a fully parameterised Series Model. */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Use the enclosing UPSeriesModel's
          {@link UPSeriesModel#stats stats(ds,lo,hi)} to return
          statistics of elements [lo,&nbsp;hi) of data-set 'ds'.
          This is often, but not always, valid;
          more on stats {@link UPModel#stats here}.  */
      public Value stats(Vector ds, int lo, int hi)
       { return UPSeriesModel.this.stats(ds, lo, hi); }

      /** Use the enclosing UPSeriesModel's
          {@link UPSeriesModel#stats(boolean,Value,Value) stats(add,ss0,ss1)}
          to combine statisticses ss0 and ss1. This is often,
          but not always, valid.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return UPSeriesModel.this.stats(add, ss0, ss1); }

      /** Calls {@link #asGiven(double,double) asGiven(0,msg2)}. */
      public M asGiven(double msg2) { return asGiven(0, msg2); }
      /** Enables setting the first- and second-part message lengths, msg1
          and msg2, after having estimated the statistical parameter(s),
          say. Also see {@link #asGiven(double) asGiven(msg2)}.  */
      public M asGiven(double msg1, double msg2)
       { return (M) sp2Model(msg1, msg2, statParams()); }

   // public abstract Analysis analysis(Scannable sv)  *  from SeriesModel.

      /** Return a String representation of 'this' SeriesModel. */
      public String toString()
       { return "(" + UPSeriesModel.this + " " + statParams() + ")"; }
    }//UPSeriesModel.M (abstract) class, the Series Model(s)
   // -------------------------------------------------------------------------

   /** A subclass of UPSeriesModels where Length.M has an explicit model of
       {@link #lenUPM lengths} (as opposed to a terminating Value/symbol).
       Also see fully parameterised {@link M}.  */
   public static abstract class Length extends UPSeriesModel        // * Length
    {
      /** 'dp' is the problem-defining parameter(s), possibly trivial.  */
      public Length(Value dp) { super(dp); }//constructor

      /** Return the explicit, UnParameterised Model of Series lengths. */
      public abstract UPModel lenUPM();                                    // *

      /** Return {@link #lenUPM()}.stats(seqs,lo.hi) statistics of
          the lengths of seqs, a data-set of Vectors.
          This is used in {@link #stats stats(seqs)}.
          Also see {@link Length.M#lenStats}.  */
      public Value lenStats(Vector seqs, int lo, int hi)
       { int [] len = new int[seqs.nElts()];
         for(int i = 0; i < seqs.nElts(); i ++ )
          { Vector vi = (Vector) seqs.elt(i);
            len[i] = vi.nElts();
          }
         return lenUPM().stats( Vector.ints(len).weightLike(seqs), lo, hi );
       }//lenStats(seqs)
      // ......................................................................

      /** Fully parameterised Model of Series having an explicit Model
          of {@link #lenMdl() lengths} (as opposed to a terminating
          Value/symbol). Also see UnParameterised {@link Length}.  */
      public abstract class M extends UPSeriesModel.M             // * Length.M
       {
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); }//constuctor

         /** Return the explicit, fully parameterised Model of lengths. */
         public abstract Model lenMdl();                                   // *

         /** The negative log probability of datum Series sv is the
             sum of the nlPr of sv's length, and of sv's elements.  */
         public double nlPr(Value sv)
          { Analysis a = analysis((Scannable) sv);
            double eltsNlPr = a.remainingNlPr();
            return eltsNlPr + lenMdl().nlPr(new Value.Int(a.position()));
          }//nlPr(v)

         /** Return {@link #lenMdl()}.stats(seqs,lo,hi) statistics on
             the <I>lengths</I> of seqs, a data-set of Vectors.
             This is used in {@link #stats stats(seqs)}.
             Also see {@link Length#lenStats}.  */
         public Value lenStats(Vector seqs, int lo, int hi)
          { int [] len = new int[seqs.nElts()];
            for(int i = 0; i < seqs.nElts(); i ++ )
             { Vector vi = (Vector) seqs.elt(i);
               len[i] = vi.nElts();
             }
            return lenMdl().stats(Vector.ints(len).weightLike(seqs), lo, hi);
          }//lenStats(seqs)

         // stats(Vector ds, int lo, int hi)         from enclosing Length

         // stats(boolean add, Valie ss0, Valie ss1) from enclosing Length

         /** Generate a random Series, that is a Vector of a random
             length, with random contents, from the Model. It needs
             {@link #lenMdl() lenMdl()} and {@link SeriesModel.Analysis#eltM}
             to implement random().  */
         public Vector random()
          { Value [] vs = new Value [ lenMdl().random().n() ];
            Vector vv = Vector.values(vs);
            Analysis a = analysis(vv); // !
            while( a.hasSome() )
             { vs[a.position()] = a.eltM().random(); // assumes model...
               a.advance();                          // ...is left to right
             }
            return vv;
          }//random()

       }//UPSeriesModel.Length.M class, fully parameterised SeriesModel

    }//UPSeriesModel.Length class, (abstract) UPSeriesModel
  // --------------------------------------------------------------------------

   /** A very simple, UnParameterised SeriesModel that uses a given
       UnParameterised Model of length&nbsp;&ge;&nbsp;0, and a given
       UnParameterised Model of (every) element.  The fully parameterised
       SeriesModel is {@link UPSeriesModel.K.M K.M}.  */
   public static class K extends UPSeriesModel.Length        // UPSeriesModel.K
    {
      /** eltUPM, the UnParameterised Model of elements;
          also see {@link #lenUPM}.  */
      public  final UPModel eltUPM;

      /** lenUPM, the UnParameterised Model of lengths;
          also see {@link #eltUPM}.  */
      private final UPModel lenUPM;

       /** Problem defining parameter
           <SPAN STYLE="white-space:nowrap">
              dp = &lang;lenUPM, eltUPM&rang;</SPAN>,
           for lengths and elements respectively.  */
      public K(Value dp)
       { super(dp);  Value.Tuple dpT = dp.isTuple(2);
         lenUPM = (UPModel)dpT.fst();  eltUPM = (UPModel)dpT.snd();
       }//constructor

       /** Return the UnParameterised Model of Lengths,
           {@link #lenUPM lenUPM}.  */
      public UPModel lenUPM() { return lenUPM; }

      /** Statistical parameter <SPAN STYLE="white-space:nowrap">
             sp = (lenMdl's sp, eltMdl's sp)</SPAN>, return a fully
          parameterised {@link K.M K.M} Series Model.  */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Given a data-set, seqs, that is a Vector of Vectors, compute
          sufficient statistics, <SPAN STYLE="white-space:nowrap">
             ss = ({@link #lenUPM lenUPM}.stats(...),
                   {@link #eltUPM eltUPM}.stats(...))</SPAN>.
          More on stats {@link UPModel#stats here}.  */
      public Value stats(Vector seqs, int lo, int hi)
       { return pair( lenStats(seqs, lo, hi),                        // lengths
                      eltUPM.stats( seqs.slice(lo,hi).append() ) );  // elements
       }//stats(,lo,hiseqs)

      /** Uses the stats(add,.,.) of {@link #lenUPM} and {@link #eltUPM}. */
      public Value stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
         return pair( lenUPM.stats(add, ss0t.fst(), ss1t.fst()),
                      eltUPM.stats(add, ss0t.snd(), ss1t.snd()) );
       }//stats(add,ss0,ss1)

      /** <SPAN STYLE="white-space:nowrap">
             ps = (ps for lenUPM's, ps for eltUPM's)</SPAN>.  */
      public Est estimator(Value ps)
       { final Value.Tuple psT = ps.isTuple(2);
         return new Est(ps)
          { public M ss2Model(Value ss)
             { Value.Tuple ssT = ss.isTuple(2);
               Model lenMdl = lenUPM.estimator(psT.fst()).ss2Model(ssT.fst());
               Model eltMdl = eltUPM.estimator(psT.snd()).ss2Model(ssT.snd());
               return new M(lenMdl.msg1()+eltMdl.msg1(),
                            lenMdl.msg2()+eltMdl.msg2(),
                       Value.pair(lenMdl.statParams(), eltMdl.statParams()));
             }//ss2Model(ss)
       }; }//estimator(ps)
      // ......................................................................

      /** {@link UPSeriesModel.K K's} fully parameterised SeriesModel. */
      public class M extends UPSeriesModel.Length.M        // UPSeriesModel.K.M
       {
         /** Fully parameterised Model, eltMdl, of elements.
             Also see {@link #lenMdl}. */
         public  final Model eltMdl;

         /** Fully parameterised Model, lenMdl, of lengths.
             Also see {@link #eltMdl}.  */
         private final Model lenMdl;

         /** Statistical parameters, <SPAN STYLE="white-space:nowrap">
             sps = &lang;{@link #lenMdl lenMdl}'s sp, 
             {@link #eltMdl eltMdl}'s sp&rang;. </SPAN>  */
         public M(double msg1, double msg2, Value sps)
          { super(msg1, msg2, sps);  Value.Tuple spsT = sps.isTuple(2);
            lenMdl = lenUPM.apply(spsT.fst()); // Model of length
            eltMdl = eltUPM.apply(spsT.snd()); // Model of (every) elt
          }//constructor

         public final Model lenMdl() { return lenMdl; }

         /** Given a data-set, seqs, that is a Vector of Vectors, compute
             statistics, <SPAN STYLE="white-space:nowrap">
                ss = ({@link #lenMdl lenMdl}.stats(...),
                      {@link #eltMdl eltMdl}.stats(...))</SPAN>.
             More on stats {@link UPModel#stats here}.  */
         public Value stats(Vector seqs, int lo, int hi)
          { return pair( lenStats(seqs, lo, hi),                    // lengths
                         eltMdl.stats(seqs.slice(lo,hi).append())); // elements
          }//stats(,lo,hiseqs)

         /** Uses the stats(add,.,.) of {@link #lenMdl} and {@link #eltMdl}. */
         public Value stats(boolean add, Value ss0, Value ss1)
          { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
            return pair( lenMdl.stats(add, ss0t.fst(), ss1t.fst()),
                         eltMdl.stats(add, ss0t.snd(), ss1t.snd()) );
          }//stats(add,ss0,ss1)

         /** Given statistics, <SPAN STYLE="white-space:nowrap">
             ss = {@link #stats stats(ds)}</SPAN>, of a data-set,
             ds, return the negative log likelihood of ds.  */
         public double nlLH(Value ss /* = (ssL,ssE) */ )
          { Value.Tuple sst = ss.isTuple(2);
            Value ssL = sst.fst(), ssE = sst.snd();
            return eltMdl.nlLH(ssE) + lenMdl.nlLH(ssL); // elts + lengths
          }//nlLH(ss)

         /** The analysis of data Series sv returns the same Model,
             {@link #eltMdl eltMdl}, for every element in sv.  */
         public Analysis analysis(Scannable sv)
          { return new Analysis(sv)
             { public Model   eltM() { return eltMdl;  }
               public void advance() { ds.advance(); }
          }; }//analysis(ds)

       }//UPSeriesModel.K.M class, the SeriesModel

    }//UPSeriesModel.K class, the UPSeriesModel
   // -------------------------------------------------------------------------

 }//UPSeriesModel (abstract) class, the UP Series Model(s)

// (c) L.Allison, 2012, Thu 7 Dec 2017. Licence to 'UPSeriesModel' is Gnu GPL.
