//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Model {@link MML#WallaceInt0 WallaceInt0} should be enough for most
    purposes but here are the classes, UnParameterised (WallaceInt0UPM) and
    fully (trivially) {@link WallaceInt0UPM.M parameterised}&nbsp;(M). */
public class WallaceInt0UPM extends Discretes                 // WallaceInt0UPM
 {
   /** dp=triv -- the problem defining parameter is trivial. */
   public WallaceInt0UPM(Value dp)
    { super(dp);  if( dp != triv ) error( ", dp=" + dp); }//constructor

   /** Return <NOBR>ss = ds,</NOBR> that is the data-set itself.
       <NOBR>ss = stats(ds)</NOBR> are sufficient statistics of ds.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo,hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return {@link MML#WallaceInt0 WallaceInt0} essentially,
       with msg2 set.  Note that msg1=0 and sp=triv. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }
   /** Return the trivial Estimator of {@link MML#WallaceInt0 WallaceInt0}. */
   public Estimator estimator(Value t)
    { return Mdl.asEstimator(t); }
   public String toString() { return "WallaceInt0upm"; }
   // .....................

   /** Note, {@link MML#WallaceInt0 WallaceInt0} is Mdl. */
   public final M Mdl = new M(0, 0, triv);

   /** Model {@link MML#WallaceInt0 WallaceInt0} should be enough for
       most purposes but here are the classes,
       fully (trivially) parameterised&nbsp;(M) and
       UnParameterised ({@link WallaceInt0UPM WallaceInt0UPM}). */
   public class M extends Discretes.M                    // WallaceInt0UPM.M
    { /** Requires msg1=0, sp=triv. */
      public M(double msg1, double msg2, Value sp)
       { super(0, msg2, Value.triv); zeroTriv(msg1, sp); }//constructor

      /** The negative log probability of int n&ge;0 depends on
          {@link la.maths.Maths#cummulativeCatalans cummulativeCatalans(i)}.
          The first few code lengths are 1, 3, 5, 5, 7, ... bits.  */
      public double nlPr_n(int n)
       { assert n >= 0 : this + ".nlPr(" + n + ")";
         int i;
         for( i = 0; cummulativeCatalans(i) <= n; i ++ ) /* skip */ ;
         // we now have the first i s.t. cummulativeCatalans(i) > n
         return (2*i + 1) * log2; // nits
       }//nlPr_n(n)

      /** Assumes that {@link WallaceInt0UPM#stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); }

      /** random_n() is not supported. */
      public int random_n()
       { throw UOE( ".random_n() is not yet(?) supported" ); }

      public String toString()
       { return this == WallaceInt0 ? "WallaceInt0" : super.toString(); }

    }//class WallaceInt0UPM.M, the Model

 }//class WallaceInt0UPM, the UPModel

// (c) L. Allison 2012, Fri 25 Nov 2016. Licence to 'WallaceInt0UPM' is Gnu GPL.
